// -*- arduino -*-
// Meissner program to either blink a LED or activate a neopixel.
//
// Copyright (C) 2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Blink.h>

// Cosntructor
Meissner_Blink::Meissner_Blink (void)
: led			(1),
  neopixel_led		(-1),
  dotstar_data		(-1),
  dotstar_clock		(-1),
  do_trace		(false),
  do_led		(true),
  do_neopixel		(true),
  do_dotstar		(true),
  neopixel_power	(-1),
  neopixel_enable	(-1),
  neopixel		(Adafruit_NeoPixel ()),
  dotstar		(Adafruit_DotStar (1))
{
}

// Destructor
Meissner_Blink::~Meissner_Blink (void)
{
}

// Setup pins, etc.
void
Meissner_Blink::begin (int  _led,
		       int  _neopixel_led,
		       int  _neopixel_power,
		       int  _neopixel_enable,
		       bool _do_trace,
		       int  _dotstar_data,
		       int  _dotstar_clock)
{
  led		  = _led;
  neopixel_led	  = _neopixel_led;
  neopixel_power  = _neopixel_power;
  neopixel_enable = _neopixel_enable;
  dotstar_data    = _dotstar_data;
  dotstar_clock   = _dotstar_clock;
  do_trace	  = (_do_trace && (USE_SERIAL_OUTPUT > 0));
  do_led	  = (_led >= 0);
  do_neopixel	  = (_neopixel_led >= 0);
  do_dotstar	  = (_dotstar_data >= 0 && _dotstar_clock >= 0);

  // Enable single LED, and turn off the led
  if (led >= 0 && do_led)
    {
        pinMode (led, OUTPUT);
	digitalWrite (led, LOW);
    }

  // Enable single neopixel, and turn it off
  if (do_neopixel)
    {
      if (neopixel_power >= 0)
	{
	  pinMode (neopixel_power, OUTPUT);
	  digitalWrite (neopixel_power, neopixel_enable);
	}

      neopixel.setPin (neopixel_led);
      neopixel.updateLength (1);
      neopixel.updateType (NEO_GRB + NEO_KHZ800);
      neopixel.begin ();

      neopixel.setPixelColor (0, 0, 0, 0);
      neopixel.show ();
    }

  // Endable single dotstar, and turn it off
  if (do_dotstar)
    {
      dotstar.updatePins (dotstar_data, dotstar_clock);
      dotstar.begin ();

      dotstar.setPixelColor (0, 0, 0, 0);
      dotstar.show ();
    }

  // Trace if desired
#if USE_SERIAL_OUTPUT
  if (do_trace)
    {
      Serial.begin (9600);

      // wait up to 3 seconds for the Serial device to become available
      while (!Serial && (millis () <= 3000))
	;

      Serial.print ("Meissner blink");
      if (do_led)
	{
	  Serial.print (", single led = ");
	  Serial.print (led);
	}
      else
	{
	  Serial.print (", no single led");
	}

      if (do_neopixel)
	{
	  Serial.print (", neopixel led = ");
	  Serial.print (neopixel_led);
	  if (neopixel_power >= 0)
	    {
	      Serial.print (", neopixel_power = ");
	      Serial.print (neopixel_power);
	      Serial.print (", neopixel_enable = ");
	      Serial.print (neopixel_enable);
	    }
	}
      else
	{
	  Serial.print (", no neopixel");
	}


      if (do_dotstar)
	{
	  Serial.print (", dotstar data = ");
	  Serial.print (dotstar_data);
	  Serial.print (", dotstar clock = ");
	  Serial.print (dotstar_clock);
	}
      else
	{
	  Serial.print (", no dotstar");
	}

      Serial.println ("");
    }
#endif	/* USE_SERIAL_OUTPUT.  */
}


// Blink the pixel
// Argument 1: Level to blink (0 == off, non-zero on, possibly multiple colors)
// Argument 2: If true blink neopixel/dotstar even if we have a single LED, if false just print single led.
void
Meissner_Blink::blink (int level, bool do_multi)
{
  int single	= LOW;
  int red	= 0;
  int green	= 0;
  int blue	= 0;

  if (do_led)
    {
      single = (level == 0) ? LOW : HIGH;
      digitalWrite (led, single);
    }

  // Neopixel & dotstar
  if (do_neopixel || do_dotstar)
    {
      int value;

      if (level == 0)
	red = green = blue = 0;

      else
	{
	  value = level & 0x7;
	  if (value == 0)
	    value = 1;

	  red   = (value & 0x1) ? 10 : 0;
	  green = (value & 0x2) ? 10 : 0;
	  blue  = (value & 0x4) ? 10 : 0;;
	}

      // Neopixel
      if (do_neopixel && (do_multi || !do_led))
	{
	  neopixel.setPixelColor (0, red, green, blue);
	  neopixel.show ();
	}

      // Dotstar
      if (do_dotstar && (do_multi || !do_led))
	{
	  dotstar.setPixelColor (0, red, green, blue);
	  dotstar.show ();
	}
    }

#if USE_SERIAL_OUTPUT
  if (do_trace)
    {
      Serial.print ("Blink (");
      Serial.print (level);
      Serial.print (", ");
      Serial.print (do_multi ? "true" : "false");
      Serial.print (")");

      if (do_led)
	{
	  Serial.print (", single pixel (");
	  Serial.print (single);
	  Serial.print (")");
	}

      if (do_neopixel && (do_multi || !do_led))
	{
	  Serial.print (", neopixel (0, ");
	  Serial.print (red);
	  Serial.print (", ");
	  Serial.print (green);
	  Serial.print (", ");
	  Serial.print (blue);
	  Serial.print (")");
	}

      if (do_dotstar && (do_multi || !do_led))
	{
	  Serial.print (", dotstar (0, ");
	  Serial.print (red);
	  Serial.print (", ");
	  Serial.print (green);
	  Serial.print (", ");
	  Serial.print (blue);
	  Serial.print (")");
	}

      Serial.println ("");
    }
#endif /* USE_SERIAL_OUTPUT.  */

  return;
}



// Enable or disable tracing.  Return the previous tracing state.
bool
Meissner_Blink::enable_trace (bool enable)
{
  bool old_state = do_trace;
  do_trace       = (enable && (USE_SERIAL_OUTPUT > 0));

#if USE_SERIAL_OUTPUT
  if (do_trace || old_state)
    {
      Serial.print ("enable_trace(");
      Serial.print (do_trace ? "true" : "false");
      Serial.println (")");
    }
#endif

  return old_state;
}


// Enable or disable blinking a single LED.  Return the previous tracing state.
bool
Meissner_Blink::enable_led (bool enable)
{
  bool old_state = do_led;
  do_led         = (enable && (led >= 0));

#if USE_SERIAL_OUTPUT
  if (do_trace)
    {
      Serial.print ("enable_led(");
      Serial.print (do_led ? "true" : "false");
      Serial.println (")");
    }
#endif

  return old_state;
}


// Enable or disable blinking a neopixel.  Return the previous tracing state.
bool
Meissner_Blink::enable_neopixel (bool enable)
{
  bool old_state = do_neopixel;
  do_neopixel    = (enable && (neopixel_led >= 0));

#if USE_SERIAL_OUTPUT
  if (do_trace)
    {
      Serial.print ("enable_neopixel(");
      Serial.print (do_neopixel ? "true" : "false");
      Serial.println (")");
    }
#endif

  return old_state;
}


// Enable or disable blinking a dotstar.  Return the previous tracing state.
bool
Meissner_Blink::enable_dotstar (bool enable)
{
  bool old_state = do_dotstar;
  do_dotstar     = (enable && (dotstar_data >= 0) && (dotstar_clock >= 0));

#if USE_SERIAL_OUTPUT
  if (do_trace)
    {
      Serial.print ("enable_dotstar(");
      Serial.print (do_dotstar ? "true" : "false");
      Serial.println (")");
    }
#endif

  return old_state;
}

// Enable or disable all LEDs.
bool
Meissner_Blink::enable_all_leds (bool enable)
{
  return (  enable_led      (enable)
	  | enable_neopixel (enable)
	  | enable_dotstar  (enable));
}


// HISTORY
// $Log: Meissner_Blink.cpp,v $
// Revision 1.6  2022/07/24 20:29:15  michaelmeissner
// Add 2nd argument to blink.
//
// Revision 1.5  2022/07/24 18:07:40  michaelmeissner
// Move enable functions from .h file to .cpp; Wrap tracing inside of #if USE_SERIAL_OUTPUT.
//
