// -*- arduino -*-
// Meissner program to either blink a LED or activate a neopixel.
//
// Copyright (C) 2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_BLINK_H
#define MEISSNER_BLINK_H		1

#include <stddef.h>
#include <stdlib.h>
#include <inttypes.h>

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Adafruit_DotStar.h>

#ifndef PIN_INTERNAL_NEOPIXEL_POWER
#define PIN_INTERNAL_NEOPIXEL_POWER	-1
#endif

#if (   defined (PIN_INTERNAL_NEOPIXEL)					\
     && defined (PIN_INTERNAL_NEOPIXEL_NUM)				\
     && PIN_INTERNAL_NEOPIXEL     >= 0					\
     && PIN_INTERNAL_NEOPIXEL_NUM == 1)

#define PIN_BLINK_NEOPIXEL_DATA		PIN_INTERNAL_NEOPIXEL
#define PIN_BLINK_NEOPIXEL_POWER	PIN_INTERNAL_NEOPIXEL_POWER

#else
#define PIN_BLINK_NEOPIXEL_DATA		-1
#define PIN_BLINK_NEOPIXEL_POWER	-1
#endif

#if (   defined (PIN_INTERNAL_DOTSTAR_DATA)				\
     && defined (PIN_INTERNAL_DOTSTAR_CLOCK)				\
     && defined (PIN_INTERNAL_DOTSTAR_NUM)				\
     && (PIN_INTERNAL_DOTSTAR_DATA  >= 0)				\
     && (PIN_INTERNAL_DOTSTAR_CLOCK >= 0)				\
     && (PIN_INTERNAL_DOTSTAR_NUM   == 1))

#define PIN_BLINK_DOTSTAR_DATA	PIN_INTERNAL_DOTSTAR_DATA
#define PIN_BLINK_DOTSTAR_CLOCK	PIN_INTERNAL_DOTSTAR_CLOCK

#else
#define PIN_BLINK_DOTSTAR_DATA	-1
#define PIN_BLINK_DOTSTAR_CLOCK	-1
#endif

#ifndef USE_SERIAL_OUTPUT
#define USE_SERIAL_OUTPUT		0
#endif

class Meissner_Blink {
 private:
  int			led;			// Single LED to use
  int			neopixel_led;		// Single Neopixel to use
  int			dotstar_data;		// Single Dotstar to use (data)
  int			dotstar_clock;		// Single Dotstar to use (clock)
  bool			do_trace;		// Whether to write to the serial device
  bool			do_led;			// Whether to enable or disable blinking a single LED
  bool			do_neopixel;		// Whether to enable or disable blinking a neopixel
  bool			do_dotstar;		// Whether to enable or disable blinking a dotstar
  int			neopixel_power;		// Pin to set to enable neopixel power
  int			neopixel_enable;	// Value to set the neopixel power pin
  Adafruit_NeoPixel	neopixel;		// Neopixel datastructure
  Adafruit_DotStar	dotstar;		// Dotstar datastructure

 public:
  Meissner_Blink (void);
  ~Meissner_Blink (void);

  // Setup either a single LED, or a single neopixel LED, and/or a single dotstar LED
  void begin (int  _led			= PIN_LED,				// Single LED (such as 13)
	      int  _neopixel_led	= PIN_BLINK_NEOPIXEL_DATA,		// Neopixel LED
	      int  _neopixel_power	= PIN_BLINK_NEOPIXEL_POWER,		// Pin to set to enable neopixel power
	      int  _neopixel_enable	= HIGH,					// Value to set the neopixel power pin
	      bool _do_trace		= false,				// Whether to trace to the serial monitor
	      int  _dotstar_data	= PIN_BLINK_DOTSTAR_DATA,		// Dotstar data LED
	      int  _dotstar_clock	= PIN_BLINK_DOTSTAR_CLOCK);		// Dotstar clock LED

  // Either turn off blink (0), or do a blink.  For neopixel/dotstar, use 1-7 to do different colors
  void blink (int  level,							// 0 for off, non-0 for on
	      bool do_multi = true);						// true to use both single and dotstar/neopixel

  // Enable tracing and LEDs.
  bool enable_trace    (bool enable);
  bool enable_led      (bool enable);
  bool enable_neopixel (bool enable);
  bool enable_dotstar  (bool enable);
  bool enable_all_leds (bool enable);
};


/*
 * Default pin to use for blinking.  If we have an internal dotstar or
 * neopixel, use that and don't use the normal single LED pin.
 */

static inline int
default_blink_pin (void)
{
#if defined(PIN_BLINK_NEOPIXEL_DATA)
  if (PIN_BLINK_NEOPIXEL_DATA >= 0)
    return PIN_BLINK_NEOPIXEL_DATA;
#endif

#if defined(PIN_BLINK_DOTSTAR_DATA)
  if (PIN_BLINK_DOTSTAR_DATA >= 0)
    return PIN_BLINK_DOTSTAR_DATA;
#endif

  return PIN_LED;
}

#endif  /* MEISSNER_BLINK_H */


// HISTORY
// $Log: Meissner_Blink.h,v $
// Revision 1.10  2022/07/25 01:05:10  michaelmeissner
// Add default_blink_pin.
//
// Revision 1.9  2022/07/24 20:29:02  michaelmeissner
// Add 2nd argument to blink.
//
// Revision 1.8  2022/07/24 18:08:35  michaelmeissner
// Update comment.
//
// Revision 1.7  2022/07/24 18:08:03  michaelmeissner
// Move enable functions from .h file to .cpp.
//
// Revision 1.6  2022/07/24 17:55:10  michaelmeissner
// Don't update led/neopixel/dotstar unless we have the leds; Add tracing to enable leds; Add enable_all_leds.
//
// Revision 1.5  2022/07/24 17:44:42  michaelmeissner
// Fix spacing.
//
// Revision 1.4  2022/07/24 17:43:49  michaelmeissner
// Update comment.
//
// Revision 1.3  2022/07/24 17:40:23  michaelmeissner
// Enable using dotstar.
//
// Revision 1.2  2022/07/24 17:19:57  michaelmeissner
// Lay the groundwork for adding dotstar support.
//
// Revision 1.1  2022/05/15 22:01:45  michaelmeissner
// Initial version.
//
