// -*- arduino -*-
// Meissner EEPROM configuration file
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_CONFIG_EEPROM_H
#define MEISSNER_CONFIG_EEPROM_H	1

#if defined(PROCESSOR_TEENSY_ARM)
#include <avr/eeprom.h>
#endif

// Reserved areas in EEPROM
// Once a field is allocated EEPROM space, never change the location
// Leave EEPROM 0..255 available to the program

#if defined(PROCESSOR_TEENSY_ARM)
// Note, the Teensy Propshield NXPMotionSense library uses some EEPROM
const int16_t EEPROM_NXPMOTION_START	= 76;
const int16_t EEPROM_NXPMOTION_END	= 76 + 51;
#endif

// Record the max EEPROM available for each processor
#if defined(PROCESSOR_TEENSY_LC)
const int16_t EEPROM_MAX		= 128;
const int16_t EEPROM_FIRST_RESERVED	= 32;

#elif defined(E2END)
const int16_t EEPROM_MAX		= E2END + 1;
const int16_t EEPROM_FIRST_RESERVED	= EEPROM_MAX - 256;

// In theory using E2END should give all of these values, but keep it around
// just in case it gets changed.
#elif defined(PROCESSOR_TEENSY_3_5) || defined(PROCESSOR_TEENSY_3_6)
const int16_t EEPROM_MAX		= 4096;
const int16_t EEPROM_FIRST_RESERVED	= 4096 - 256;

#elif defined(PROCESSOR_TEENSY_4_X)
const int16_t EEPROM_MAX		= 1080;
const int16_t EEPROM_FIRST_RESERVED	= 1080 - 256;

#elif defined(PROCESSOR_TEENSY_3_X)
const int16_t EEPROM_MAX		= 2048;
const int16_t EEPROM_FIRST_RESERVED	= 2048 - 256;

#elif defined(PROCESSOR_ATMEGA328P) || defined(PROCESSOR_ATMEGA32U4)
const int16_t EEPROM_MAX		= 1024;
const int16_t EEPROM_FIRST_RESERVED	= 1024 - 256;

#elif defined(__SAM3X8E__)
// Note, the Due does not have EEPROM, but the DigiX does, since I own DigiX, and not Due's I use that
const int16_t EEPROM_MAX		= 4096;
const int16_t EEPROM_FIRST_RESERVED	= 4096 - 256;

#elif defined (PROCESSOR_ATTINY85)
const int16_t EEPROM_MAX		= 512;
const int16_t EEPROM_FIRST_RESERVED	= 256;

#else
const int16_t EEPROM_MAX		= -1;
const int16_t EEPROM_FIRST_RESERVED	= -1;
#endif

// Saved elapsed time
struct elapsed_time {
  uint16_t	num_weeks;
  uint8_t	num_days;
  uint8_t	num_hours;
  uint8_t	num_minutes;
  uint8_t	num_seconds;
  uint16_t	num_micro;
};

enum eeprom_fields {
  // Current version/length of the eeprom information
  EEPROM_CUR_VERSION			= EEPROM_FIRST_RESERVED,
  EEPROM_CUR_LENGTH,

  // 14 overrides to default pin locations
  EEPROM_PIN_LED,					// pin to use for led, default to PIN_LED
  EEPROM_PIN_RANDOM,					// pin to use for seeing random, default to PIN_RANDOM
  EEPROM_PIN_NEOPIXEL,					// pin to use for neopixels, default to PIN_NEOPIXEL
  EEPROM_PIN_UNUSED,					// unused pin for tindie OLED display
  EEPROM_PIN_RESERVED,					// save some more space

  // unsigned 32-bit integer to hold 32 dip-switches
  EEPROM_DIP_SWITCH			= EEPROM_FIRST_RESERVED + 16,

  // Unsigned 16-bit integer to hold the random state between runs
  EEPROM_RANDOM				= EEPROM_DIP_SWITCH + 4,

  // unsigned long to hold elapsed time
  EEPROM_ELAPSED			= EEPROM_RANDOM + 4,

  // last byte
  EEPROM_LAST_RESERVED			= EEPROM_ELAPSED + sizeof (unsigned long)
};

// Return value if the pin is not set (eeprom default value)
const uint8_t NO_PIN			= 0xff;

// Change version # if we ever add/subtract fields in the middle
// Version 2:	changed to switch to using put/get/update
// Version 3:	store elapsed time as unsigned long

const uint8_t EEPROM_VERSION		= 3;
const uint8_t EEPROM_LENGTH		= ((uint8_t)(((uint16_t)EEPROM_LAST_RESERVED) - (uint16_t)EEPROM_FIRST_RESERVED));

#endif	/* MEISSNER_CONFIG_EEPROM_H  */


// HISTORY
// $Log: Meissner_Config_Eeprom.h,v $
// Revision 1.8  2022/01/18 17:20:20  michaelmeissner
// Update copyright year.
//
// Revision 1.7  2020/01/28 04:55:27  michaelmeissner
// Update copyright year.
//
// Revision 1.6  2019/08/18 14:39:52  michaelmeissner
// On Teensy include <avr/eeprom.h> and use the EEPROM size indicated.
//
// Revision 1.5  2019/08/17 14:50:55  michaelmeissner
// Update comment.
//
// Revision 1.4  2019/08/17 14:47:19  michaelmeissner
// Add support for Teensy 4.0
//
// Revision 1.3  2019/08/16 04:56:32  michaelmeissner
// Use PROCESSOR_TEENSY_ARM, not PROCESSOR_TEENSY_3_X.
//
// Revision 1.2  2019/01/14 02:13:26  michaelmeissner
// Update copyright.
//
// Revision 1.1  2018/12/27 06:22:48  michaelmeissner
// Initial version.
//

