// Adapted by Michael Meissner from defragster, post #421
// https://forum.pjrc.com/threads/21680-New-I2C-library-for-Teensy3?p=116089&viewfull=1#post116089

#include <stddef.h>
#include <stdlib.h>
#include <inttypes.h>
#include <Meissner_Config.h>

#if PROCESSOR_TEENSY_3_X
#include <TeensyMAC.h>
#endif

#if PROCESSOR_TEENSY_4_X
#include <T4_PowerButton.h>
extern void tempmon_init (void);
extern float tempmonGetTemp (void);

static bool tempmon_init_done = false;

float
GetTemp_C (void)
{
  if (!tempmon_init_done)
    {
      tempmon_init ();
      tempmon_init_done = true;
    }

  return tempmonGetTemp ();
}

float
GetTemp_F (void)
{
  return ((GetTemp_C () / 5.0f) * 9.0f) + 32.0f;
}
#endif

#if USE_SERIAL_OUTPUT && !NO_SERIAL
#define COLUMN_SIZE	28

// Print a header
static void
print_hdr (const char *msg)
{
  size_t len = strlen (msg);

  Serial.print (msg);
  while (len++ < COLUMN_SIZE)
    Serial.print (" ");

  Serial.print (" = ");
}

// Print a number
static void
print_num (const char *msg, int32_t num)
{
  print_hdr (msg);
  Serial.println (num);
}

// Print a pin number if it exists
static void
print_pin (const char *msg, int pin)
{
  if (pin >= 0)
    {
      print_hdr (msg);
      if (pin <= 9)
	Serial.print (" ");

      Serial.println (pin);
    }
}

// Print a string
static void
print_str (const char *msg, const char *str)
{
  print_hdr (msg);
  Serial.println (str);
}

// Convert a number to a string
static char *
number_to_string (uint32_t num)
{
  static char buffer[32];
  char *p = &buffer[31];
  int num_digits = 0;

  *p-- = '\0';
  do
    {
      *p-- = '0' + (num % 10);
      num /= 10;
      if (num > 0 && ++num_digits >= 3)
	{
	  *p-- = ',';
	  num_digits = 0;
	}
    }
  while (num > 0);

  return p+1;
}

// Print a speed value in Hz
static void
print_hz (void)
{
  size_t max_len = 0;

  struct hz_value {
    const char *msg;
    uint32_t    hz;
  };

  static struct hz_value hz_values[] =
    {
#if F_CPU
     { "F_CPU", F_CPU },
#endif

#if F_PLL
     { "F_PLL", F_PLL },
#endif

#if F_BUS
     { "F_BUS", F_BUS },
#endif

#if F_MEM
     { "F_MEM", F_MEM },
#endif
  };

  for (size_t i = 0; i < sizeof (hz_values) / sizeof (struct hz_value); i++)
    {
      size_t len = strlen (number_to_string (hz_values[i].hz));
      if (len > max_len)
	max_len = len;
    }

  for (size_t i = 0; i < sizeof (hz_values) / sizeof (struct hz_value); i++)
    {
      char *string = number_to_string (hz_values[i].hz);
      size_t len   = strlen (string);

      print_hdr (hz_values[i].msg);
      while (len++ < max_len)
	{
	  Serial.print (" ");
	}

      Serial.println (string);
    }
}

#if PROCESSOR_TEENSY_3_OR_4_X
static void
print_teensy (void)
{
  print_num ("NVIC_NUM_INTERRUPTS", NVIC_NUM_INTERRUPTS);
  print_num ("DMA_NUM_CHANNELS",    DMA_NUM_CHANNELS);
  print_num ("CORE_NUM_TOTAL_PINS", CORE_NUM_TOTAL_PINS);
  print_num ("CORE_NUM_DIGITAL",    CORE_NUM_DIGITAL);
  print_num ("CORE_NUM_INTERRUPT",  CORE_NUM_INTERRUPT);
}
#endif

#if PROCESSOR_TEENSY_3_X
static void
print_hex (const char *msg, uint32_t num)
{
  print_hdr (msg);
  Serial.print ("0x");
  Serial.println (num, HEX);
}

static void
print_teensy_3_x ()
{
  uint64_t serial = teensySerial ();
  uint64_t mac    = teensyMAC ();
  ssize_t  i;
  uint8_t  hex_byte;

#if PIN_TEMP_SENSOR > 0
  print_num ("Temperature sensor", analogRead (PIN_TEMP_SENSOR));
#endif

  print_hex ("Serial number", serial);

  print_hdr ("Ethernet MAC address");
  for (i = 40; i >= 0; i -= 8)
    {
      if (i != 40)
	Serial.print (":");
      hex_byte = (mac >> i);
      Serial.print ((hex_byte > 4) & 0xf, HEX);
      Serial.print (hex_byte & 0xf, HEX);
    }
  Serial.println ("");
}
#endif	/* PROCESSOR_TEENSY_3_X  */


// Print Teensy 4.x information
#if PROCESSOR_TEENSY_4_X
static void
print_teensy_4_x ()
{
  float temp_c = GetTemp_C ();
  float temp_f = GetTemp_F ();

  if (temp_c && temp_f)
    {
      Serial.printf ("%-*s = %gF (%gC)\n", COLUMN_SIZE, "Temperature", temp_f, temp_c);
    }

#ifdef ARDUINO_TEENSY41
  {
    extern long _extram_start	__attribute__((weak));
    extern long _extram_end	__attribute__((weak));

    if (&_extram_start != NULL && &_extram_end != NULL)
      {
	Serial.printf ("%-*s = 0x%lx\n", COLUMN_SIZE, "Extram start", _extram_start);
	Serial.printf ("%-*s = 0x%lx\n", COLUMN_SIZE, "Extram end",   _extram_end);
      }
  }
#endif

  Serial.println ("");

#if 0
  flexRamInfo ();
#endif
}
#endif	/* Teensy 4.x.  */
#endif	/* Use serial  */

void
CPUspecs (bool print_config,
	   int  pin_led,
	   int  pin_neopixel,
	   int  pin_neopixel_single,
	   int  pin_button,
	   int  pin_random,
	   int  pin_unused,
	   int  pin_dac,
	   int  pin_dac1,
	   int  pin_temp,
	   int  pin_light,
	   int  pin_dotstar_data,
	   int  pin_dotstar_clock)
{
#if USE_SERIAL_OUTPUT && !NO_SERIAL
  // wait up to 3 seconds for the Serial device to become available
  while (!Serial && (millis () <= 3000))
    ;

  print_str ("CPU",                PROCESSOR_NAME);
  print_str ("Date/time compiled", __DATE__ " at " __TIME__);

#if ARDUINO
  print_num ("ARDUINO", ARDUINO);
#endif

  print_hz ();

  if (print_config)
    {
#if PROCESSOR_TEENSY_3_OR_4_X
      print_teensy ();
#endif

      if (pin_button >= 0)
	print_pin ("PIN_BUTTON",			pin_button);

      if (pin_led >= 0)
	print_pin ("PIN_LED",				pin_led);

      if (pin_neopixel >= 0)
	print_pin ("PIN_NEOPIXEL",			pin_neopixel);

      if (pin_neopixel_single >= 0)
	print_pin ("PIN_INTERNAL_NEOPIXEL",		pin_neopixel_single);

      if (pin_dotstar_data >= 0 && pin_dotstar_clock >= 0)
	{
	  print_pin ("PIN_INTERNAL_DOTSTAR_DATA",	pin_dotstar_data);
	  print_pin ("PIN_INTERNAL_DOTSTAR_CLOCK",	pin_dotstar_clock);
	}

      if (pin_random >= 0)
	print_pin ("PIN_RANDOM",			pin_random);

      if (pin_unused >= 0)
	print_pin ("PIN_UNUSED",			pin_unused);

      if (pin_dac >= 0)
	print_pin ("PIN_DAC",				pin_dac);

      if (pin_dac1 >= 0)
	print_pin ("PIN_DAC1",				pin_dac1);

      if (pin_temp >= 0)
	print_pin ("PIN_TEMP_SENSOR",			pin_temp);

      if (pin_light >= 0)
	print_pin ("PIN_LIGHT_SENSOR",			pin_light);

      print_num ("USE_EEPROM",				USE_EEPROM);
      print_num ("USE_EEPROM_LIB",			USE_EEPROM_LIB);

#if USE_EEPROM
      print_num ("EEPROM_MAX",				EEPROM_MAX);
      print_num ("EEPROM_FIRST_RESERVED",		EEPROM_FIRST_RESERVED);
      print_num ("EEPROM_NXPMOTION_START",		EEPROM_NXPMOTION_START);
      print_num ("EEPROM_NXPMOTION_END",		EEPROM_NXPMOTION_END);
#endif

#if PROCESSOR_TEENSY_4_X
      print_teensy_4_x ();
#endif
      
#if PROCESSOR_TEENSY_3_X
      print_teensy_3_x ();
#endif
    }
#endif	/* USE_SERIAL_OUTPUT */
}


// HISTORY
// $Log: Meissner_Cpu_Info.cpp,v $
// Revision 1.41  2022/08/08 05:07:27  michaelmeissner
// Add support for printing the internal dotstar data/clock on certain Adafruit boards.
//
// Revision 1.40  2022/01/11 01:06:44  michaelmeissner
// Comment flexRamInfo for now.
//
// Revision 1.39  2021/01/03 17:31:11  michaelmeissner
// On Teensy 4.1, print start/end of external ram.
//
// Revision 1.38  2020/06/07 03:37:09  michaelmeissner
// On Teensy 3.x, print the value of the internal temperature sensor.
//
// Revision 1.37  2020/06/07 03:29:20  michaelmeissner
// Use T4_PowerButton to print T4.x information.
//
// Revision 1.36  2020/05/14 03:49:35  michaelmeissner
// If we built this on Teensy 4.1 and use at least 1.52 beta 5, print out the psram size.
//
// Revision 1.35  2020/05/12 00:01:46  michaelmeissner
// Print DTCM/ITCM memory map.
//
// Revision 1.34  2020/05/11 23:53:11  michaelmeissner
// Print out DTCM for Teensy 4.x.
//
// Revision 1.33  2020/05/11 23:10:50  michaelmeissner
// Print Teensy 4.0/4.1 memory information.  Re-add printing some information from Teensy 3.x.
//
// Revision 1.32  2019/11/26 15:00:25  michaelmeissner
// Don't print temps. if it is 0.0
//
// Revision 1.31  2019/11/26 14:55:39  michaelmeissner
// Add GetTemp_{C,F} on Teensy 4.0.
//
// Revision 1.30  2019/11/13 04:34:13  michaelmeissner
// Add temp. sensor for Teensy 4.0.
//
// Revision 1.29  2019/07/11 04:33:16  michaelmeissner
// Don't display ethernet # on Teensy 4 for now #2
//
// Revision 1.28  2019/07/11 04:32:03  michaelmeissner
// Don't display ethernet # on Teensy 4 for now.
//
// Revision 1.27  2019/02/20 03:45:07  michaelmeissner
// Disable getting Teensy serial # and ethernet # for now.
//
// Revision 1.26  2019/01/11 03:17:17  michaelmeissner
// Dac2 -> Dac1.
//
// Revision 1.25  2019/01/06 14:35:32  michaelmeissner
// Add pin_neopixel_single, pin_dac2
//
// Revision 1.24  2018/12/27 14:33:05  michaelmeissner
// Pass more pins as arguments; Print HZ values with commas; Wait for serial to be available.
//
// Revision 1.23  2018/12/27 07:42:17  michaelmeissner
// Spacing.
//
// Revision 1.22  2018/12/27 07:41:36  michaelmeissner
// Move ARDUINO higher than F_<xxx>.
//
// Revision 1.21  2018/12/27 07:41:10  michaelmeissner
// Move date/time compiled higher.
//
// Revision 1.20  2018/12/27 07:29:40  michaelmeissner
// Mod last change to be safe in 32-bit systems.
//
// Revision 1.19  2018/12/27 07:28:21  michaelmeissner
// Use 64-bit types for printing if F_CPU is too large.
//
// Revision 1.18  2018/12/27 07:22:38  michaelmeissner
// void static => static void
//
// Revision 1.17  2018/12/27 07:21:32  michaelmeissner
// Spacing; Move print_hex into Teensy only section.
//
// Revision 1.16  2018/12/27 07:19:22  michaelmeissner
// Move Teensy specific prints to separate function.
//
// Revision 1.15  2018/12/27 07:15:51  michaelmeissner
// Rework the print out
//
// Revision 1.14  2018/12/27 06:51:31  michaelmeissner
// Don't bother printing i2c pin defs.
//
// Revision 1.13  2018/12/27 06:49:57  michaelmeissner
// Pass various pin defaults from the caller.
//
// Revision 1.12  2018/12/27 06:36:51  michaelmeissner
// Add support for new Meissner_Config.h settings.
//
// Revision 1.11  2016/10/22 22:18:24  michaelmeissner
// Add option to suppress printing PIN_<xxx>
//
// Revision 1.10  2016/10/20 13:35:41  michaelmeissner
// PIN_DAT1 -> PIN_DAC1.
//
// Revision 1.9  2016/10/20 13:27:23  michaelmeissner
// Print the DAC prints correctly on the 3.5/3.6.
//
// Revision 1.8  2016/10/10 02:26:04  michaelmeissner
// Fix alignment of Serial.print's
//
// Revision 1.7  2016/10/08 04:32:36  michaelmeissner
// Print Teensy EEPROM values.
//
// Revision 1.6  2016/10/08 04:23:58  michaelmeissner
// Print some meissner config values
//
// Revision 1.5  2016/09/24 05:25:45  michaelmeissner
// Fix printing mac address to print on byte boundaries.
//
// Revision 1.4  2016/09/24 03:44:44  michaelmeissner
// Improve printing the MAC address.
//
// Revision 1.3  2016/09/24 03:29:41  michaelmeissner
// Print the serial number and MAC address as well.
//
// Revision 1.2  2016/09/22 12:26:14  michaelmeissner
// Print date/time compiled.
//
// Revision 1.1  2016/09/22 12:19:26  michaelmeissner
// Initial version (with meissner hacks)
//
