// -*- arduino -*-
// Meissner delay library
// based on http://www.gammon.com.au/blink
//
// Copyright (C) 2014-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_DELAY_H
#define MEISSNER_DELAY_H	1

#include <stddef.h>
#include <stdlib.h>
#include <inttypes.h>

#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h"

#else
#include "WProgram.h"
#include "WConstants.h"
#endif

class Delay {
 private:
  unsigned long	init_time;			// time when begin was called
  unsigned long	interval;			// interval in milliseconds to return true
  bool		repeat;				// whether to repeat the timer
  bool		immediate;			// immediate return true on the next call

 public:
  // constructor, destructor
  Delay (void) : init_time (0UL), interval (0UL), repeat (false), immediate (false) {}
  ~Delay (void) {}

  // register that delay_done will return true in the future
  void
  begin (unsigned long amount, bool r = false, bool imm = false)
  {
    init_time = millis ();
    interval  = amount;
    repeat    = r;
    immediate = imm;
  }

  // clear any potential delay
  void
  clear (void)
  {
    init_time = 0UL;
    interval  = 0UL;
    repeat    = false;
    immediate = false;
  }

  // Mark that the next call to delay_done will return true
  // If we are passed true, do not reset the timers.
  void
  run_immediate (bool no_reset = false)
  {
    if (!no_reset)
      {
	init_time = 0UL;
	interval  = 0UL;
	repeat    = false;
      }

    immediate = true;
  }

  // register a time in the future
  void
  run_delay (unsigned long amount)
  {
    init_time = millis ();
    interval  = amount;
    repeat    = false;
    immediate = false;
  }

  // register a time in the future with repeat
  void
  run_delay_repeat (unsigned long amount)
  {
    init_time = millis ();
    interval  = amount;
    repeat    = true;
    immediate = false;
  }

  // Return true if the delay is done
  bool delay_done (void);
};
#endif	/* MEISSNER_DELAY_H */


// HISTORY
// $Log: Meissner_Delay.h,v $
// Revision 1.12  2022/02/13 00:58:49  michaelmeissner
// Update copyright year.
//
// Revision 1.11  2020/01/28 05:05:44  michaelmeissner
// Update copyright year.
//
// Revision 1.10  2019/01/14 02:16:10  michaelmeissner
// Update copyright.
//
// Revision 1.9  2018/03/18 21:48:03  michaelmeissner
// Add optional argument to run_immedate to suppress reseting init_time, interval, or repeat.
//
// Revision 1.8  2016/04/03 22:43:31  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.7  2016/03/21 13:58:56  michaelmeissner
// Update copyright year.
//
// Revision 1.6  2015/01/12 02:15:13  michaelmeissner
// Move delay_done code from .h file to .cpp file.
//
// Revision 1.5  2015/01/11 14:46:48  michaelmeissner
// Add run_immediate, run_delay, and run_delay_repeat.
//
// Revision 1.4  2015/01/10 03:13:13  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.3  2015/01/01 23:35:08  michaelmeissner
// Add ability to run immediately.
//
// Revision 1.2  2015/01/01 19:15:32  michaelmeissner
// Update copyright year
//
// Revision 1.1  2014/12/31 13:54:50  michaelmeissner
// Initial version.
//
