// -*- arduino -*-
// Meissner EEPROM files
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>

#if USE_EEPROM_LIB
#include <Meissner_Eeprom.h>
#include <EEPROM.h>

Meissner_Eeprom MRM_eeprom;


// Initialize the EEPROM fields on the first call
// If the EEPROM area was a different version, just wipe it all out

void
Meissner_Eeprom::initialize (void)
{
  uint8_t e_version = EEPROM.read (EEPROM_CUR_VERSION);
  uint8_t e_length  = EEPROM.read (EEPROM_CUR_LENGTH);

  // Previous version or different size?
  if (e_version != EEPROM_VERSION || e_length != EEPROM_LENGTH)
    {
      char bytes[EEPROM_LENGTH];

      memset (bytes, 0xff, EEPROM_LENGTH);
      bytes[ (int)EEPROM_CUR_VERSION - (int)EEPROM_FIRST_RESERVED ] = EEPROM_VERSION;
      bytes[ (int)EEPROM_CUR_LENGTH  - (int)EEPROM_FIRST_RESERVED ] = EEPROM_LENGTH;
      EEPROM.put (EEPROM_FIRST_RESERVED, bytes);
    }

  first_time = false;
}


// EEPROM functions for keeping track of the default pins
// across random power cycles

// get saved pin number

uint8_t
Meissner_Eeprom::get_pin (int16_t pin_num, uint8_t def)
{
  uint8_t value;

  if (pin_num < 0 || pin_num >= EEPROM_MAX)
    return def;

  if (first_time)
    initialize ();

  value = EEPROM.read (pin_num);
  if (value == NO_PIN)
    {
      if (def == NO_PIN)
	{
	  switch (pin_num)
	    {
	    case EEPROM_PIN_LED:	def = PIN_LED;		break;
	    case EEPROM_PIN_RANDOM:	def = PIN_RANDOM;	break;
	    case EEPROM_PIN_NEOPIXEL:	def = PIN_NEOPIXEL;	break;
	    case EEPROM_PIN_UNUSED:	def = PIN_UNUSED;	break;
	    default:						break;
	    }

	  value = def;
	}
    }

  return value;
}

// set saved pin

void
Meissner_Eeprom::set_pin (int16_t pin_num, uint8_t value)
{
  if (pin_num < 0 || pin_num >= EEPROM_MAX)
    return;

  if (first_time)
    initialize ();

  EEPROM.update (pin_num, value);
  return;
}


// EEPROM functions for keeping track of dip switches
// across random power cycles

// Get the current dip switches.  If all dip switches are 0xff, assume the
// EEPROM has not been set, and we should pretend the switches are not used

// Get dip switches as a 32-bit unsigned integer

uint32_t
Meissner_Eeprom::get_dip_switch (void)
{
  uint32_t value;

  if (first_time)
    initialize ();

  EEPROM.get (EEPROM_DIP_SWITCH, value);
  return (value == 0xffffffffU) ? 0 : value;
}

// Set the dip switches

void
Meissner_Eeprom::set_dip_switch (uint32_t value)
{
  if (first_time)
    initialize ();

  EEPROM.put (EEPROM_DIP_SWITCH, value);
}

// Return one dip switch

bool
Meissner_Eeprom::get_one_dip_switch (int dipsw)
{
  uint32_t value;

  if (dipsw < 0 || dipsw >= 32)
    return false;

  if (first_time)
    initialize ();

  value = get_dip_switch ();
  return (value & (((uint32_t)1) << dipsw)) != 0;
}

// Set one dip switch
void
Meissner_Eeprom::set_one_dip_switch (int dipsw, bool state)
{
  uint32_t value, orig_value;

  if (dipsw < 0 || dipsw >= 32)
    return;

  if (first_time)
    initialize ();

  orig_value = value = get_dip_switch ();
  if (state)
    value |= ((uint32_t)1) << dipsw;
  else
    value &= ~(((uint32_t)1) << dipsw);

  if (orig_value != value)
    set_dip_switch (value);
}


// EEPROM functions for keeping track of unsigned 16-bit values across random
// power cycles

// get 16-bit unsigned value.

uint16_t
Meissner_Eeprom::get_uint16 (int16_t addr)
{
  uint16_t ret;

  if (addr < 0 || addr >= EEPROM_MAX-1)
    return 0;

  if (first_time)
    initialize ();

  EEPROM.get (addr, ret);
  return ret;
}

// set 16-bit unsigned value

void
Meissner_Eeprom::set_uint16 (int16_t addr, uint16_t value)
{
  if (addr < 0 || addr >= EEPROM_MAX-1)
    return;

  if (first_time)
    initialize ();

  EEPROM.put (addr, value);
}


// EEPROM functions for keeping track of unsigned 32-bit values across random
// power cycles

// get 32-bit unsigned value.

uint32_t
Meissner_Eeprom::get_uint32 (int16_t addr)
{
  uint32_t ret;

  if (addr < 0 || addr >= EEPROM_MAX-3)
    return 0;

  if (first_time)
    initialize ();

  EEPROM.get (addr, ret);
  return ret;
}

// set 32-bit unsigned value

void
Meissner_Eeprom::set_uint32 (int16_t addr, uint32_t value)
{
  if (addr < 0 || addr >= EEPROM_MAX-3)
    return;

  if (first_time)
    initialize ();

  EEPROM.put (addr, value);
}


// EEPROM functions to keep track of elapsed time across random power cycles

// Clear elapsed time

void
Meissner_Eeprom::clear_elapsed (void)
{
  if (first_time)
    initialize ();

  for (size_t i = 0; i < sizeof (struct elapsed_time); i++)
    {
      elapsed.bytes[i] = 0;
    }

  elapsed_millis = 0UL;
  EEPROM.put (EEPROM_ELAPSED, elapsed_millis);
}

// Bump elapsed time

void
Meissner_Eeprom::bump_elapsed (bool update_eeprom,
			       bool only_eeprom,
			       bool skip_eeprom)
{
  unsigned long now_millis;
  unsigned long delta;

  if (first_time)
    initialize ();

  // On first call, read previous time from eeprom
  if (!read_elapsed)
    {
      read_elapsed = true;

      if (skip_eeprom)
	elapsed_eeprom = 0UL;

      else
	EEPROM.get (EEPROM_ELAPSED, elapsed_eeprom);
    }

  // If we only want the elapsed time from eeprom, return now
  if (only_eeprom)
    return;

  now_millis      = millis ();
  delta           = now_millis - prev_millis;
  prev_millis     = now_millis;
  elapsed_millis += delta;

  // Update eeprom unless we just wanted to update the in core time
  if (update_eeprom)
    {
      saved_elapsed_eeprom = true;
      EEPROM.put (EEPROM_ELAPSED, elapsed_millis);
    }
}

// Return current elapsed time
// Assume that bump_elapsed has been called

struct elapsed_time *
Meissner_Eeprom::get_elapsed (bool eeprom_p)
{
  unsigned long total;
  unsigned long micros;
  unsigned long seconds;
  unsigned long minutes;
  unsigned long hours;
  unsigned long days;
  unsigned long weeks;

  total = (eeprom_p) ? elapsed_eeprom : elapsed_millis;

  micros  = total   % 1000UL;
  seconds = total   / 1000UL;

  minutes = seconds / 60UL;
  seconds = seconds % 60UL;

  hours   = minutes / 60UL;
  minutes = minutes % 60UL;

  days    = hours   / 24UL;
  hours   = hours   % 24UL;

  weeks   = days    / 7UL;
  days    = days    % 7UL;

  elapsed.e.num_micro   = micros;
  elapsed.e.num_seconds = seconds;
  elapsed.e.num_minutes = minutes;
  elapsed.e.num_hours   = hours;
  elapsed.e.num_days    = days;
  elapsed.e.num_weeks   = weeks;

  return &elapsed.e;
}

// Return current elapsed time formatted as a text string

char *
Meissner_Eeprom::format_elapsed (bool eeprom_p)
{
  char *tb = elapsed_buffer;
  struct elapsed_time *ep;
  int hours, minutes, seconds;

  if (eeprom_p && !saved_elapsed_eeprom)
    {
      strcpy (tb, "<none>");
      return tb;
    }

  ep = get_elapsed (eeprom_p);

  // weeks
  tb[0] = ((ep->num_weeks > 9)
	   ? '*'
	   : '0' + ep->num_weeks);

  // days
  tb[1] = '/';
  tb[2] = '0' + (ep->num_days % 10);

  // hours
  hours = (ep->num_hours % 24);
  tb[3] = ' ';
  tb[4] = '0' + (hours / 10);
  tb[5] = '0' + (hours % 10);

  // minutes
  minutes = (ep->num_minutes % 60);
  tb[6]   = ':';
  tb[7]   = '0' + (minutes / 10);
  tb[8]   = '0' + (minutes % 10);

  // seconds
  seconds = (ep->num_seconds % 60);
  tb[ 9]  = ':';
  tb[10]  = '0' + (seconds / 10);
  tb[11]  = '0' + (seconds % 10);

  // optional '*' to indicate we are updating the eeprom
  if (saved_elapsed_eeprom)
    {
      tb[12] = '*';
      tb[13] = '\0';
    }
  else
    tb[12] = '\0';

  return tb;
}

#endif /* USE_EEPOM_LIB */


// HISTORY
// $Log: Meissner_Eeprom.cpp,v $
// Revision 1.27  2022/02/13 01:04:23  michaelmeissner
// Update copyright year.
//
// Revision 1.26  2020/01/28 05:06:08  michaelmeissner
// Update copyright year.
//
// Revision 1.25  2019/01/14 02:15:43  michaelmeissner
// Update copyright.
//
// Revision 1.24  2017/11/13 04:35:17  michaelmeissner
// Spacing.
//
// Revision 1.23  2017/11/13 04:34:17  michaelmeissner
// Rewrite elapsed time interface.
//
// Revision 1.22  2017/11/12 14:25:46  michaelmeissner
// Add format_elapsed to return elapsed time as a printable string.
//
// Revision 1.21  2016/10/03 14:44:39  michaelmeissner
// Simplify initialize to reset the array if the version/length are different & use EEPROM put.
//
// Revision 1.20  2016/10/03 03:58:29  michaelmeissner
// Fix limit testing for get/set_uint32
//
// Revision 1.19  2016/10/03 03:24:24  michaelmeissner
// Add get/set_uint32; Use EEPROM.put/get to write elapsed time structure all at once.
//
// Revision 1.18  2016/10/03 02:40:55  michaelmeissner
// Use new EEPROM update, get, put functions; Delete a lot of unused functions.
//
// Revision 1.17  2016/10/03 02:09:02  michaelmeissner
// Use EEPROM.update instead of EEPROM.write.
//
// Revision 1.16  2016/04/03 22:44:12  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.15  2016/03/21 13:59:49  michaelmeissner
// Update copyright year.
//
// Revision 1.14  2015/11/25 13:01:15  michaelmeissner
// Don't provide Eeprom library at all on some systems.
//
// Revision 1.13  2015/02/01 16:33:31  michaelmeissner
// Fix typo.
//
// Revision 1.12  2015/01/10 03:12:40  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.11  2015/01/01 19:16:22  michaelmeissner
// Update copyright year
//
// Revision 1.10  2014/11/18 18:40:13  michaelmeissner
// Update copyright
//
// Revision 1.9  2014-05-14 15:53:09  michaelmeissner
// whitespace
//
// Revision 1.8  2014-05-14 15:49:20  michaelmeissner
// Rewrite elapsed time processing to prevent overflows.
//
// Revision 1.7  2014-04-03 05:44:29  michaelmeissner
// Add argument to skip eeprom time altogether.
//
// Revision 1.6  2014-04-03 04:32:13  michaelmeissner
// Add additional argument to elapsed time functions to not incorporate local time.
//
// Revision 1.5  2013-11-09 05:08:32  michaelmeissner
// Add ability to bump elapsed time without updating the EEPROM.
//
// Revision 1.4  2013-11-09 04:47:35  michaelmeissner
// Fix calculation of hours/days in elapsed time.
//
// Revision 1.3  2013-11-08 13:59:12  michaelmeissner
// Flesh out elapsed time support
//
// Revision 1.2  2013-11-08 04:48:57  michaelmeissner
// Convert to class form.
//
// Revision 1.1  2013-11-07 15:29:25  michaelmeissner
// Initial version.
//
// Revision 1.1  2013-10-11 13:07:58  michaelmeissner
// Initial version.
//
