// -*- arduino -*-
// Meissner EEPROM support file
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_EEPROM_H
#define MEISSNER_EEPROM_H	1

#include <stddef.h>
#include <stdlib.h>
#include <inttypes.h>

#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h"

#else
#include "WProgram.h"
#include "WConstants.h"
#endif

#if USE_EEPROM_LIB
#include <Meissner_Config.h>
#include <EEPROM.h>

class Meissner_Eeprom {
  bool first_time;						// initialization flag
  bool read_elapsed;						// we have read the elased time
  bool saved_elapsed_eeprom;					// we have saved the elapsed time in EEPROM

  void initialize (void);					// internal function to possibly zap the eeprom information

  union {
    struct elapsed_time e;					// elapsed time fields in normal form
    uint8_t bytes[sizeof (struct elapsed_time)];		// byte form of elapsed time
  } elapsed;

  unsigned long elapsed_millis;					// total elapsed time in milliseconds
  unsigned long elapsed_eeprom;					// previous elapsed time from EEPROM
  unsigned long prev_millis;					// last time we bumped the elapsed time

  char elapsed_buffer[16];					// buffer for formatting elapsed time

public:
  Meissner_Eeprom (void)
  {
    first_time = true;
    read_elapsed = false;
    saved_elapsed_eeprom = false;
    prev_millis = elapsed_millis = 0UL;

    for (size_t i = 0; i < sizeof (struct elapsed_time); i++)
      elapsed.bytes[i] = 0;

    return;
  }

  ~Meissner_Eeprom (void)
  {
    return;
  }

  // get/set default pins
  uint8_t get_pin (int16_t pin_num, uint8_t def = NO_PIN);
  void    set_pin (int16_t pin_num, uint8_t value);

  // get/set 32 dip switches
  uint32_t get_dip_switch (void);
  void	   set_dip_switch (uint32_t switches);
  bool	   get_one_dip_switch (int switch_num);
  void	   set_one_dip_switch (int switch_num, bool value);

  // get/set saved unsigned 16-bit values
  uint16_t get_uint16 (int16_t addr);
  void     set_uint16 (int16_t addr, uint16_t value);

  // get/set saved unsigned 32-bit values
  uint32_t get_uint32 (int16_t addr);
  void     set_uint32 (int16_t addr, uint32_t value);

  // set, bump, return, and format elapsed time
  void clear_elapsed (void);
  void bump_elapsed (bool update_eeprom = true,
		     bool only_eeprom   = false,
		     bool skip_eeprom	= false);

  struct elapsed_time	*get_elapsed    (bool eeprom_p);
  char			*format_elapsed (bool eeprom_p);
};

extern Meissner_Eeprom MRM_eeprom;

#endif	/* USE_EEPROM_LIB */
#endif	/* MEISSNER_EEPROM_H */


// HISTORY
// $Log: Meissner_Eeprom.h,v $
// Revision 1.20  2022/02/13 00:59:12  michaelmeissner
// Update copyright year.
//
// Revision 1.19  2020/01/28 05:06:01  michaelmeissner
// Update copyright year.
//
// Revision 1.18  2019/01/14 02:15:29  michaelmeissner
// Update copyright.
//
// Revision 1.17  2017/11/13 04:34:01  michaelmeissner
// Rewrite elapsed time interface.
//
// Revision 1.16  2017/11/12 14:26:06  michaelmeissner
// Add format_elapsed to return elapsed time as a printable string.
//
// Revision 1.15  2016/10/03 03:11:56  michaelmeissner
// Add get/set_uin32
//
// Revision 1.14  2016/10/03 02:41:11  michaelmeissner
// Use new EEPROM update, get, put functions; Delete a lot of unused functions.
//
// Revision 1.13  2016/04/03 22:43:53  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.12  2016/03/21 13:59:34  michaelmeissner
// Update copyright year.
//
// Revision 1.11  2015/11/25 13:00:29  michaelmeissner
// Don't provide Eeprom library at all on some systems.
//
// Revision 1.10  2015/01/10 03:12:15  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.9  2015/01/01 19:16:07  michaelmeissner
// Update copyright year
//
// Revision 1.8  2014/11/18 18:39:56  michaelmeissner
// Update copyright
//
// Revision 1.7  2014-04-03 05:42:32  michaelmeissner
// Add field for whether we've read the elapsed time.
//
// Revision 1.6  2014-04-03 05:36:18  michaelmeissner
// Add argument to skip eeprom time altogether.
//
// Revision 1.5  2014-04-03 04:32:28  michaelmeissner
// Add additional argument to elapsed time functions to not incorporate local time.
//
// Revision 1.4  2013-11-09 05:08:53  michaelmeissner
// Add ability to bump elapsed time without updating the EEPROM.
//
// Revision 1.3  2013-11-08 13:59:28  michaelmeissner
// Flesh out elapsed time support
//
// Revision 1.2  2013-11-08 04:48:44  michaelmeissner
// Convert to class form.
//
// Revision 1.1  2013-11-07 15:29:10  michaelmeissner
// Initial version.
//

