// Arduino Morse Library
// Written by Erik Linder SM0RVV and Mark VandeWettering K6HX
//
// Released 2011 under GPLv3
//
// Version 0.2

// Code modified by Michael Meissner
// Code changes Copyright (C) 2012-2022 Michael Meissner
//
// Meissner_Morse is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 3, or (at your option) any later
// version.

#ifndef MEISSNER_MORSE_H
#define MEISSNER_MORSE_H 1

#include <Meissner_Config.h>

#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h"

#else
#include "WProgram.h"
#include "WConstants.h"
#endif

#ifndef MORSE_DEFAULT_DOT_DELAY
#define MORSE_DEFAULT_DOT_DELAY		50
#endif

#ifndef MORSE_DEFAULT_DASH_DELAY
#define MORSE_DEFAULT_DASH_DELAY	(3 * MORSE_DEFAULT_DOT_DELAY)
#endif

#ifndef MORSE_DEFAULT_PWM_RATE
#define MORSE_DEFAULT_PWM_RATE		128
#endif

#ifndef MORSE_DEFAULT_DAT_TONE
#define MORSE_DEFAULT_DAT_TONE		1500
#endif

// Inner sub-class for doing morse code
class Morse_inner {
 private:
  unsigned long dot_delay;		/* delay for dots.  */
  unsigned long dash_delay;		/* delay for dashes.  */

 public:
  Morse_inner (unsigned long _dot_delay,				// delay for dots
	       unsigned long _dash_delay)				// delay for dashes
  {
    dot_delay  = (_dot_delay)  ? _dot_delay  : MORSE_DEFAULT_DOT_DELAY;
    dash_delay = (_dash_delay) ? _dash_delay : MORSE_DEFAULT_DASH_DELAY;
  }

  ~Morse_inner (void)
  {
  }

  void inner_begin (unsigned long _dot_delay,				// delay for dots
		    unsigned long _dash_delay);				// delay for dashes

  // Functions to print text
  void sendmsg (const char *str);
  void sendmsg (char *str);
  void send (char c);

  // Inner functions to turn the led/buzzer on/off
  virtual void on  (void)	= 0;
  virtual void off (void)	= 0;
};


// Morse class for flashing a LED on/off
class Morse_led : public Morse_inner {
 private:
  int led_pin;

 public:
  Morse_led (int	   _led,					// led to flash on/off
	     unsigned long _dot_delay	= 0,				// delay for dots
	     unsigned long _dash_delay	= 0)				// delay for dashes

    : Morse_inner (_dot_delay, _dash_delay)

  {
    led_pin = _led;
  }

  ~Morse_led (void)
  {
  }

  // Initialization
  void begin (int	    _led	= -1,				// led to flash on/off
	      unsigned long _dot_delay	= 0,				// delay for dots
	      unsigned long _dash_delay	= 0);				// delay for dashes

  void on (void);
  void off (void);
};


// Morse class for using analogWrite to set a particular tone, optionally including a digital LED as well
class Morse_analog : public Morse_inner {
 private:
  int pwm_pin;
  int led_pin;
  uint16_t pwm_rate;

 public:
  Morse_analog (int		_pwm,
		uint16_t	_rate		= 0,
		int		_led		= -1,
		unsigned long	_dot_delay	= 0,
		unsigned long	_dash_delay	= 0) : Morse_inner (_dot_delay, _dash_delay)
  {
    pwm_pin  = _pwm;
    pwm_rate = (_rate) ? _rate : MORSE_DEFAULT_PWM_RATE;
    led_pin  = _led;
  }

  ~Morse_analog (void)
  {
  }

  // Initialization
  void begin (int	    _pwm	= -1,			// PWM pin to write to
	      uint16_t	    _rate	= 0,			// Rate to flash pin on/off
	      int	    _led	= -1,			// Secondary pin to flash
	      unsigned long _dot_delay	= 0,			// Delay for dots
	      unsigned long _dash_delay	= 0);			// Delay for dashes

  // Turn leds/buzzer, etc. on/off
  void on (void);
  void off (void);

  // Adjust the current pwm rate
  void set_pwm_rate (uint16_t _rate)
  {
    pwm_rate = _rate;
  }

  // Return the current pwm rate
  uint16_t get_pwm_rate (void)
  {
    return pwm_rate;
  }
};


// Morse class for using analogWrite to set a particular tone on the Teensy Prop shield, optionally including a digital LED as well

class Morse_propshield : public Morse_inner {
 private:
  int		pwm_pin;
  int		led_pin;
  int		enable_dat_pin;
  int		enable_neopixel_pin;
  int		enable_flash_pin;
  uint16_t	tone;
  uint8_t	save_dat;
  uint8_t	save_neopixel;
  uint8_t	save_flash;

 public:
  Morse_propshield (int		  _pwm,					// Pin connected to the DAT
		    uint16_t	  _tone		= 0,			// Tone to use
		    int		  _led		= -1,			// Pin (13) blink at the same time as doing the DAT
		    int		  _dat		= -1,			// Pin to enable speaker
		    int		  _flash	= -1,			// Pin for reading/writing flash memory
		    int		  _neopixel	= -1,			// Pin to enable 11/13 as outputs
		    unsigned long _dot_delay	= 0,			// Delay used for dots
		    unsigned long _dash_delay	= 0)			// Delay used for dashes

    : Morse_inner (_dot_delay, _dash_delay)

  {
    pwm_pin		= _pwm;
    tone		= (_tone) ? _tone : MORSE_DEFAULT_DAT_TONE;
    led_pin		= _led;
    enable_dat_pin	= _dat;
    enable_flash_pin	= _flash;
    enable_neopixel_pin	= _neopixel;
  }

  ~Morse_propshield (void)
  {
  }

  // Initialization
  void begin (int		_pwm		= -1,			// Pin connected to the DAT
	      uint16_t		_tone		= 0,			// Tone to use
	      int		_led		= -1,			// Pin (13) blink at the same time as doing the DAT
	      int		_dat		= -1,			// Pin to enable speaker
	      int		_flash		= -1,			// Pin for reading/writing flash memory
	      int		_neopixel	= -1,			// Pin to enable 11/13 as outputs
	      unsigned long	_dot_delay	= 0,			// Delay used for dots
	      unsigned long	_dash_delay	= 0);			// Delay used for dashes

  // Turn DAT on/off
  void on (void);
  void off (void);

  // Adjust the current tone to use
  void set_tone (uint16_t _tone)
  {
    tone = _tone;
  }

  // Return the current pwm rate
  uint16_t get_tone (void)
  {
    return tone;
  }
};

#endif	/* MEISSNER_MORSE_H  */


// HISTORY
// $Log: Meissner_Morse.h,v $
// Revision 1.4  2022/02/13 00:59:33  michaelmeissner
// Update copyright year.
//
// Revision 1.3  2019/01/14 04:37:33  michaelmeissner
// Update copyright.
//
// Revision 1.2  2016/04/04 15:04:14  michaelmeissner
// Reset prop shield enable pins to their prior state.
//
// Revision 1.1  2016/04/04 03:32:01  michaelmeissner
// Initial version.
//
