// -*- arduino -*-
// Meissner neopixel (ws2812) wrapper library
//
// Copyright (C) 2014-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Neopixel.h>


// Should we use Serial output?  If we do want to use it, insure that it is set
// up for use.

bool
use_serial_output (void)
{
#if USE_SERIAL_OUTPUT
  static bool first_time = true;

  if (first_time)
    {
      first_time = false;

      Serial.begin (9600);
      while (!Serial && millis () < 3000UL)
	;
    }

  return true;

#else
  return false;
#endif
}


// Abort with a message

void
abort_with_message (const char *msg)
{
  (void) use_serial_output ();

  while (1)
    {
      Serial.println (msg);
      delay (1000L);
    }
}


// Neopixel_single code

// Initialize neopixel_single with the real neopixel object
void
Neopixel_single::single_init (Adafruit_NeoPixel *np)
{
  if (!np && !neopixel)
    abort_with_message ("Error, Neopixel_single with no neopixel.");

  else if (np)
    neopixel = np;

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    {
      Serial.print ("Neopixel_single::single_init (0x");
      Serial.print ((unsigned long)neopixel, HEX);
      Serial.println (")");
    }
#endif
}

// Show all of the neopixel LEDs
void
Neopixel_single::show (void)
{
  if (neopixel)
    neopixel->show ();
  else
    abort_with_message ("Neopixel_single::show, no neopixel");
}

// Set a pixel to RGB or RGBW
void
Neopixel_single::setPixelColor (Neopixel_index_t num,
				Neopixel_power_t red,
				Neopixel_power_t green,
				Neopixel_power_t blue)
{
  if (neopixel)
    neopixel->setPixelColor (num, pack_rgb (red, green, blue));
  else
    abort_with_message ("Neopixel_single::setPixelColor, no neopixel");
}

void
Neopixel_single::setPixelColor (Neopixel_index_t num,
				Neopixel_power_t red,
				Neopixel_power_t green,
				Neopixel_power_t blue,
				Neopixel_power_t white)
{
  if (neopixel)
    neopixel->setPixelColor (num, pack_rgbw (red, green, blue, white));
  else
    abort_with_message ("Neopixel_single::setPixelColor, no neopixel");
}

void
Neopixel_single::setPixelColor (Neopixel_index_t num,
				Neopixel_rgb_t   color)
{
  if (neopixel)
    neopixel->setPixelColor (num, color);
  else
    abort_with_message ("Neopixel_single::setPixelColor, no neopixel");
}

// Return total number of pixels
Neopixel_index_t
Neopixel_single::max_pixels (void)
{
  Neopixel_index_t ret = 0;

  if (neopixel)
    ret = neopixel->numPixels ();
  else
    abort_with_message ("Neopixel_single::max_pixels, no neopixel");

  return ret;
}


// Neopixel_pair code

// Initialize neopixel_pair with the pair of real neopixel objects
void
Neopixel_pair::pair_init (Adafruit_NeoPixel *np1,
			  Adafruit_NeoPixel *np2)
{
  if (np1)
    neopixel1 = np1;

  if (np2)
    neopixel2 = np2;

  if (!neopixel1 || !neopixel2)
    abort_with_message ("Error, Neopixel_pair with no neopixels.");
  else
    pixels_in_first = neopixel1->numPixels ();


#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    {
      Serial.print ("Neopixel_pair::pair_init (");
      Serial.print (pixels_in_first);
      Serial.print (", 0x");
      Serial.print ((unsigned long)neopixel1, HEX);
      Serial.print (", 0x");
      Serial.print ((unsigned long)neopixel2, HEX);
      Serial.println (")");
    }
#endif

}

// Show all of the neopixel LEDs
void
Neopixel_pair::show (void)
{
  if (neopixel1 && neopixel2 && pixels_in_first)
    {
      if (neopixel1_set)
	neopixel1->show ();

      if (neopixel2_set)
	neopixel2->show ();
    }
  else
    abort_with_message ("Neopixel_pair::show, no paired neopixels");

  neopixel1_set = neopixel2_set = false;
}

// Set a pixel to RGB or RGBW
void
Neopixel_pair::setPixelColor (Neopixel_index_t num,
			      Neopixel_power_t red,
			      Neopixel_power_t green,
			      Neopixel_power_t blue)
{
  Neopixel_rgb_t color = pack_rgb (red, green, blue);

  if (neopixel1 && neopixel2 && pixels_in_first)
    {
      if (num < pixels_in_first)
	{
	  neopixel1->setPixelColor (num, color);
	  neopixel1_set = true;
	}
      else
	{
	  neopixel2->setPixelColor (num - pixels_in_first, color);
	  neopixel2_set = true;
	}
    }
  else
    abort_with_message ("Neopixel_pair::setPixelColor, no paired neopixels");
}

void
Neopixel_pair::setPixelColor (Neopixel_index_t num,
			      Neopixel_power_t red,
			      Neopixel_power_t green,
			      Neopixel_power_t blue,
			      Neopixel_power_t white)
{
  Neopixel_rgb_t color = pack_rgbw (red, green, blue, white);

  if (neopixel1 && neopixel2 && pixels_in_first)
    {
      if (num < pixels_in_first)
	{
	  neopixel1->setPixelColor (num, color);
	  neopixel1_set = true;
	}
      else
	{
	  neopixel2->setPixelColor (num - pixels_in_first, color);
	  neopixel2_set = true;
	}
    }
  else
    abort_with_message ("Neopixel_pair::setPixelColor, no paired neopixels");
}

void
Neopixel_pair::setPixelColor (Neopixel_index_t num,
			      Neopixel_rgb_t   color)
{
  if (neopixel1 && neopixel2 && pixels_in_first)
    {
      if (num < pixels_in_first)
	{
	  neopixel1->setPixelColor (num, color);
	  neopixel1_set = true;
	}
      else
	{
	  neopixel2->setPixelColor (num - pixels_in_first, color);
	  neopixel2_set = true;
	}
    }
  else
    abort_with_message ("Neopixel_pair::setPixelColor, no paired neopixels");
}

// Return total number of pixels
Neopixel_index_t
Neopixel_pair::max_pixels (void)
{
  Neopixel_index_t ret = 0;

  if (neopixel1 && neopixel2)
    ret = neopixel1->numPixels () + neopixel2->numPixels ();
  else
    abort_with_message ("Neopixel_pair::max_pixels, no paired neopixels");

  return ret;
}


#if PROCESSOR_TEENSY_ARM

// Ws2812serial_single code

// Initialize ws2812_single with the real Ws2812serial object
void
Ws2812serial_single::single_init (WS2812Serial *np)
{
  if (!np && !ws2812serial)
    abort_with_message ("Error, Ws2812serial_single with no Ws2812serial.");

  else if (np)
    ws2812serial = np;

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    {
      Serial.print ("Ws2812serial_single::single_init (0x");
      Serial.print ((unsigned long)ws2812serial, HEX);
      Serial.println (")");
    }
#endif
}

// Show all of the neopixel LEDs
void
Ws2812serial_single::show (void)
{
  if (ws2812serial)
    ws2812serial->show ();
  else
    abort_with_message ("Ws2812serial_single::show, no WS2812Serial");
}

// Set a pixel to RGB or RGBW
void
Ws2812serial_single::setPixelColor (Neopixel_index_t num,
				    Neopixel_power_t red,
				    Neopixel_power_t green,
				    Neopixel_power_t blue)
{
  if (ws2812serial)
    ws2812serial->setPixelColor (num, pack_rgb (red, green, blue));
  else
    abort_with_message ("Ws2812serial_single::setPixelColor, no WS2812Serial");
}

void
Ws2812serial_single::setPixelColor (Neopixel_index_t num,
				    Neopixel_power_t red,
				    Neopixel_power_t green,
				    Neopixel_power_t blue,
				    Neopixel_power_t white)
{
  if (ws2812serial)
    ws2812serial->setPixelColor (num, pack_rgbw (red, green, blue, white));
  else
    abort_with_message ("Ws2812serial_single::setPixelColor, no WS2812Serial");
}

void
Ws2812serial_single::setPixelColor (Neopixel_index_t num,
				    Neopixel_rgb_t   color)
{
  if (ws2812serial)
    ws2812serial->setPixelColor (num, color);
  else
    abort_with_message ("Ws2812serial_single::setPixelColor, no WS2812Serial");
}

// Return total number of pixels
Neopixel_index_t
Ws2812serial_single::max_pixels (void)
{
  Neopixel_index_t ret = 0;

  if (ws2812serial)
    ret = ws2812serial->numPixels ();
  else
    abort_with_message ("Ws2812serial_single::max_pixels, no WS2812Serial");

  return ret;
}


// Ws2812serial_pair code

// Initialize neopixel_pair with the pair of real neopixel objects
void
Ws2812serial_pair::pair_init (WS2812Serial *np1,
			      WS2812Serial *np2)
{
  if (np1)
    ws2812serial1 = np1;

  if (np2)
    ws2812serial2 = np2;

  if (!ws2812serial1 || !ws2812serial2)
    abort_with_message ("Error, Ws2812serial_pair with no neopixels.");
  else
    pixels_in_first = (ws2812serial1) ? ws2812serial1->numPixels () : 0;

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    {
      Serial.print ("Ws2812serial_pair::pair_init (");
      Serial.print (pixels_in_first);
      Serial.print (", 0x");
      Serial.print ((unsigned long)ws2812serial1, HEX);
      Serial.print (", 0x");
      Serial.print ((unsigned long)ws2812serial2, HEX);
      Serial.println (")");
    }
#endif

}

// Show all of the neopixel LEDs
void
Ws2812serial_pair::show (void)
{
  if (!ws2812serial1 && !ws2812serial2)
    abort_with_message ("Ws2812serial_pair::show, no paired WS2812Serials");

  else
    {
      if (ws2812serial1 && ws2812serial1_set)
	ws2812serial1->show ();

      if (ws2812serial2 && ws2812serial2_set)
	ws2812serial2->show ();

      ws2812serial1_set = ws2812serial2_set = false;
    }
}

// Set a pixel to RGB or RGBW
void
Ws2812serial_pair::setPixelColor (Neopixel_index_t num,
				  Neopixel_power_t red,
				  Neopixel_power_t green,
				  Neopixel_power_t blue)
{
  Neopixel_rgb_t color = pack_rgb (red, green, blue);

  if (ws2812serial1 && num < pixels_in_first)
    {
      ws2812serial1->setPixelColor (num, color);
      ws2812serial1_set = true;
    }

  else if (ws2812serial2)
    {
      ws2812serial2->setPixelColor (num - pixels_in_first, color);
      ws2812serial2_set = true;
    }

  else
    abort_with_message ("Ws2812serial_pair::setPixelColor, no paired WS2812Serials");
}

void
Ws2812serial_pair::setPixelColor (Neopixel_index_t num,
				  Neopixel_power_t red,
				  Neopixel_power_t green,
				  Neopixel_power_t blue,
				  Neopixel_power_t white)
{
  Neopixel_rgb_t color = pack_rgbw (red, green, blue, white);

  if (ws2812serial1 && num < pixels_in_first)
    {
      ws2812serial1->setPixelColor (num, color);
      ws2812serial1_set = true;
    }

  else if (ws2812serial2)
    {
      ws2812serial2->setPixelColor (num - pixels_in_first, color);
      ws2812serial2_set = true;
    }

  else
    abort_with_message ("Ws2812serial_pair::setPixelColor, no paired WS2812Serials");
}

void
Ws2812serial_pair::setPixelColor (Neopixel_index_t num,
				  Neopixel_rgb_t   color)
{
  if (ws2812serial1 && num < pixels_in_first)
    {
      ws2812serial1->setPixelColor (num, color);
      ws2812serial1_set = true;
    }

  else if (ws2812serial2)
    {
      ws2812serial2->setPixelColor (num - pixels_in_first, color);
      ws2812serial2_set = true;
    }

  else
    abort_with_message ("Ws2812serial_pair::setPixelColor, no paired WS2812Serials");
}

// Return total number of pixels
Neopixel_index_t
Ws2812serial_pair::max_pixels (void)
{
  Neopixel_index_t ret = 0;

  if (!ws2812serial1 && !ws2812serial2)
    abort_with_message ("Ws2812serial_pair::max_pixels, no paired neopixels");

  else
    {
      if (ws2812serial1)
	ret += ws2812serial1->numPixels ();

      if (ws2812serial2)
	ret += ws2812serial2->numPixels ();
    }

  return ret;
}

#endif	/* PROCESSOR_TEENSY_ARM  */

// constructor for the main Neopixel objects
Neopixel_object::Neopixel_object (Neopixel_index_t num,		// number of logical pixels
				  Neopixel_index_t max,		// total pixels
				  Neopixel_pin_t   pin,		// pin neopixel is on
				  Neopixel_pin_t   led,		// pin led is on
				  Neopixel_type_t  t,		// type
				  bool		   mirror,	// whether this is 2 objects that either have
								// the same addressing or reverse indexing
								// for doing mirrored rings
				  bool		   white_led,	// whether the neopixel is RGBW instead of RGB
				  bool		   spi_reserve,	// do SPI.beginTransaction to prevent SPI bus conflicts (prop shield)
				  Neopixel_pin_t   enable,	// pin to set high during neopixel show to enable LEDs
				  Neopixel_pin_t   pin2,	// pin the second neopixel is on
				  unsigned long    spi_delay)	// delay after doing SPI.beginTransaction

  :  lowlevel		((Neopixel_lowlevel *)0),
     type		(t),
     num_pixels		(num),
     total_pixels	(max),
     second_pixel	(pin2 >= 0 ? num : 0xffff),
     cur_pixel		(0),
     pin_neopixel	(pin),
     pin_neopixel2	(pin2),
     pin_led		(((led == pin) || (led == pin2)) ? -1 : led),
     pin_enable		(enable),
     led_state		(false),
     do_blink		(led >= 0 && led != pin),
     have_mirror	(mirror),
     have_white_leds	(white_led),
     do_spi_reserve	(spi_reserve),
     do_spi_delay	(spi_delay),
     blink_level	(1)
{
  if (pin_enable >= 0)
    pinMode (pin_enable, OUTPUT);
}

// destructor
Neopixel_object::~Neopixel_object ()
{
}


// Common initialization function between begin with one neopixel pattern,
// and a split ring setup where there are two separate neopixels.

void
Neopixel_object::begin_common (Neopixel_index_t cur)
{
  cur_pixel = cur;
  led_state = false;

  if (pin_led >= 0)
    {
      // Enable blinking on microprocessors that have a single neopixel LED.
      if (pin_led == PIN_NEOPIXEL_SINGLE)
	blink_data.begin (-1, pin_led);			// single neopixel pin
      else
	blink_data.begin (pin_led);			// single pin
    }

  // If we don't have a single LED to blink, see if we have an internal neopixel or dotstar.
  else
    {
      if ((PIN_BLINK_NEOPIXEL_DATA >= 0)
	  || (PIN_BLINK_DOTSTAR_DATA >= 0 && PIN_BLINK_DOTSTAR_CLOCK >= 0))
	{
	  blink_data.begin ();
	  do_blink = true;
	}
    }

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    {
      Serial.println ("==========");
      Serial.println ("Neopixel_object: begin");
      Serial.print   ("lowlevel       = 0x");	Serial.println ((size_t)lowlevel, HEX);
      Serial.print   ("cur            = ");	Serial.println (cur);
      Serial.print   ("num_pixels     = ");	Serial.println (num_pixels);
      Serial.print   ("total_pixels   = ");	Serial.println (total_pixels);
      Serial.print   ("second_pixel   = ");	Serial.println (second_pixel);
      Serial.print   ("pin_neopixel   = ");	Serial.println (pin_neopixel);
      Serial.print   ("pin_neopixel2  = ");	Serial.println (pin_neopixel2);
      Serial.print   ("pin_led        = ");	Serial.println (pin_led);
      Serial.print   ("pin_enable     = ");	Serial.println (pin_enable);
      Serial.print   ("have_mirror    = ");	Serial.println (have_mirror     ? "true" : "false");
      Serial.print   ("white_leds     = ");	Serial.println (have_white_leds ? "true" : "false");
      Serial.print   ("do_spi_reserve = ");	Serial.println (do_spi_reserve  ? "true" : "false");
      Serial.print   ("do_spi_delay   = ");	Serial.println (do_spi_delay);
      Serial.println ("==========");
    }
#endif
}

// Initialization with an existing neopixel descriptor

void
Neopixel_object::begin (Adafruit_NeoPixel *np, Neopixel_index_t cur, bool do_begin)
{
  lowlevel = &single;
  single.single_init (np);
  if (do_begin)
    np->begin ();

  begin_common (cur);
}

// Initialization with two existing neopixel descriptors

void
Neopixel_object::begin (Adafruit_NeoPixel *np, Adafruit_NeoPixel *np2, Neopixel_index_t cur, bool do_begin)
{
  lowlevel = &pair;
  pair.pair_init (np, np2);

  if (do_begin)
    {
      np->begin ();
      np2->begin ();
    }

  begin_common (cur);
}

#if PROCESSOR_TEENSY_ARM
// Initialization with an existing WS2812Serial descriptor

void
Neopixel_object::begin (WS2812Serial *np, Neopixel_index_t cur, bool do_begin)
{
  lowlevel = &ws2812serial_single;
  ws2812serial_single.single_init (np);
  if (do_begin)
    np->begin ();

  begin_common (cur);
}

// Initialization with two existing neopixel descriptors

void
Neopixel_object::begin (WS2812Serial *np, WS2812Serial *np2, Neopixel_index_t cur, bool do_begin)
{
  lowlevel = &ws2812serial_pair;
  ws2812serial_pair.pair_init (np, np2);

  if (do_begin)
    {
      np->begin ();
      np2->begin ();
    }

  begin_common (cur);
}
#endif	/* PROCESSOR_TEENSY_ARM  */


#if defined (SPI_HAS_TRANSACTION) && PROCESSOR_TEENSY_ARM

// Return true if the pin is one of the special SPI0 pins for ouptut
static inline bool
is_spi0_pin (Neopixel_pin_t pin)
{
  if (pin < 0)
    return false;

  switch (pin)
    {
    default:
      break;

#if PROCESSOR_TEENSY_3_5 || PROCESSOR_TEENSY_3_6
    case 27:			// 3.5/3.6 alternate SCLK0
    case 28:			// 3.5/3.6 alternate MOSI0
#endif
    case  7:			// alternate MOSI0
    case 11:			// normal MOSI0
    case 13:			// normal SCLK
    case 14:			// alternate SCLK
      return true;
    }

  return false;
}

#endif

// display the pixels and possibly blink the LED

void
Neopixel_object::show (void)
{
  if (lowlevel)
    {
#if defined (SPI_HAS_TRANSACTION) && PROCESSOR_TEENSY_ARM
      bool pin_spi_p  = do_spi_reserve && is_spi0_pin (pin_neopixel);
      bool pin2_spi_p = do_spi_reserve && is_spi0_pin (pin_neopixel2);

      if (do_spi_reserve)
	{
	  // beginTransaction prevents SPI bus conflicts
	  // We need to reset 11 & 13 to being digital pins
	  // https://forum.pjrc.com/threads/46640-Neopixels-with-SPI-transactions-on-prop-shield-working-with-ST7735-SSD1351-displays
	  SPI.beginTransaction (SPISettings(24000000, MSBFIRST, SPI_MODE0));

	  if (pin_spi_p)
	    pinMode (pin_neopixel, OUTPUT);

	  if (pin2_spi_p)
	    pinMode (pin_neopixel2, OUTPUT);

	  // Doing uncanny eyes, we need a slight delay after doing the beingTransaction so that the first LED doesn't glitch
	  if (do_spi_delay)
	    delay (do_spi_delay);
	}
#endif

      if (pin_enable >= 0)
	digitalWrite (pin_enable, HIGH);

      lowlevel->show ();

      if (do_blink)
	blink ();

      if (pin_enable >= 0)
	digitalWrite (pin_enable, LOW);

#if defined (SPI_HAS_TRANSACTION) && PROCESSOR_TEENSY_ARM
      if (do_spi_reserve)
	{
	  // allow other libs to use SPI again
	  SPI.endTransaction ();

#if PROCESSOR_TEENSY_3_X || PROCESSOR_TEENSY_LC
	  if (pin_spi_p)
	    {
	      volatile uint32_t *reg = portConfigRegister (pin_neopixel);
	      *reg = PORT_PCR_MUX (2);
	    }

	  if (pin2_spi_p)
	    {
	      volatile uint32_t *reg = portConfigRegister (pin_neopixel2);
	      *reg = PORT_PCR_MUX (2);
	    }

#elif PROCESSOR_TEENSY_4_x
	  if (pin_spi_p || pin_spi2_p)
	    Serial.println ("Need to add Teensy 4.0 transaction support");
#endif

	}
#endif
    }
  else
    abort_with_message ("Neopixel_object::show, no lowlevel neopixel");
}

void
Neopixel_object::show (bool explicit_blink)
{
  bool save_blink = do_blink;

  do_blink = explicit_blink;
  show ();
  do_blink = save_blink;
}


// Blink the led to show progress

void
Neopixel_object::blink (void)
{
  if (do_blink)
    {
      led_state = !led_state;
      blink_data.blink (led_state ? blink_level : 0);
    }
}


// Set all of the neopixels to the same color

void
Neopixel_object::fill (Neopixel_rgb_t color)
{
  if (lowlevel)
    {
      for (Neopixel_index_t i = 0; i < total_pixels; i++)
	lowlevel->setPixelColor (i, color);
    }
  else
    abort_with_message ("Neopixel_object::fill, no lowlevel neopixel");
}

void
Neopixel_object::fill_rgb (Neopixel_power_t red,
			   Neopixel_power_t green,
			   Neopixel_power_t blue)
{
  if (lowlevel)
    {
      Neopixel_rgb_t color = pack_rgb (red, green, blue);

      for (Neopixel_index_t i = 0; i < total_pixels; i++)
	lowlevel->setPixelColor (i, color);
    }
  else
    abort_with_message ("Neopixel_object::fill_rgb, no lowlevel neopixel");
}

void
Neopixel_object::fill_white (Neopixel_power_t white)
{
  if (lowlevel)
    {
      Neopixel_rgb_t color = white_color (white);

      for (Neopixel_index_t i = 0; i < total_pixels; i++)
	lowlevel->setPixelColor (i, color);
    }
  else
    abort_with_message ("Neopixel_object::fill_white, no lowlevel neopixel");
}


// Flash the LEDs as part of the bootup sequence

void
Neopixel_object::flash_leds (Neopixel_power_t power_level, unsigned long timeout)
{
  // first do all red
#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    Serial.println ("neopixel flash to red");
#endif

  fill_rgb (power_level, 0, 0);
  show ();

  // next do all green
#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    Serial.println ("neopixel flash to green");
#endif

  delay (timeout);
  fill_rgb (0, power_level, 0);
  show ();

  // next do all blue
#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    Serial.println ("neopixel flash to blue");
#endif

  delay (timeout);
  fill_rgb (0, 0, power_level);
  show ();

  // if we have RGBW leds, flash the white light
  if (have_white_leds)
    {
#ifdef NEOPIXEL_DEBUG
      if (use_serial_output ())
	Serial.println ("neopixel flash to white");
#endif

      delay (timeout);
      fill_white (power_level);
      show ();
    }

  // indicate the top pixel (and side pixels if this is a ring)
#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    Serial.println ("neopixel flash first pixel blue, rest to red");
#endif

  delay (timeout);
  fill_rgb (power_level, 0, 0);
  setPixelColor (0, pack_rgb (0, 0, power_level));
  show ();

  // reset neopixel
#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    Serial.println ("neopixel reset");
#endif

  delay (timeout);
  fill_rgb (0, 0, 0);
  show ();

  return;
}

// Same as above, but give the user explict control over the blinking
void
Neopixel_object::flash_leds (Neopixel_power_t power_level,
			     unsigned long    timeout,
			     bool             explicit_blink)
{
  bool save_blink = do_blink;

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output () && explicit_blink != save_blink)
    {
      Serial.print ("flash_leds, override blink to");
      Serial.println (explicit_blink ? "on" : "off");
    }
#endif

  do_blink = explicit_blink;
  flash_leds (power_level, timeout);
  do_blink = save_blink;

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output () && explicit_blink != save_blink)
    {
      Serial.print ("flash_leds, restore blink to");
      Serial.println (save_blink ? "on" : "off");
    }
#endif

  return;
}


// Debug functions

void
Neopixel_object::debug_color_32bit (const char      *type_name,
				    Neopixel_index_t num,
				    Neopixel_rgb_t   color,
				    Neopixel_mask_t  mask)
{
#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    {
      Serial.print (type_name);
      Serial.print (" SetPixelColor (");
      if (num < 10)
	Serial.print ("  ");
      else if (num < 100)
	Serial.print (" ");
      Serial.print (num);
      Serial.print (", 0x");

      for (int i = 28; i >= 0; i -= 4)
	Serial.print ((color >> i) & 0xf, HEX);

      Serial.print ("), mask: ");
      if (mask == NEOPIXEL_MASK_NONE)
	Serial.print ("none");

      else
	{
	  const char *comma = "";

	  if ((mask & NEOPIXEL_MASK_MIRROR_REVERSE) != 0)
	    {
	      Serial.print (comma);
	      Serial.print ("mirror reverse");
	      comma = ", ";
	      mask &= ~NEOPIXEL_MASK_MIRROR_REVERSE;
	    }

	  if ((mask & NEOPIXEL_MASK_MIRROR_NO) != 0)
	    {
	      Serial.print (comma);
	      Serial.print ("mirror no");
	      comma = ", ";
	      mask &= ~NEOPIXEL_MASK_MIRROR_NO;
	    }

	  if ((mask & NEOPIXEL_MASK_MIRROR_ONLY) != 0)
	    {
	      Serial.print (comma);
	      Serial.print ("mirror only");
	      comma = ", ";
	      mask &= ~NEOPIXEL_MASK_MIRROR_ONLY;
	    }

	  if ((mask & NEOPIXEL_MASK_MIRROR_RANDOM) != 0)
	    {
	      Serial.print (comma);
	      Serial.print ("mirror random");
	      comma = ", ";
	      mask &= ~NEOPIXEL_MASK_MIRROR_RANDOM;
	    }

	  if (mask)
	    {
	      Serial.print (comma);
	      Serial.print ("other bits = 0x");
	      Serial.print (mask, HEX);
	    }
	}

      Serial.println ("");
    }
#endif

  return;
}


/*
 * Single ring element.
 */

// set the color for a given pixel

void
Neopixel_ring::set_pixel_color_32bit_mask (Neopixel_index_t num,
					   Neopixel_rgb_t   color,
					   Neopixel_mask_t  mask)
{
  set_pixel_color_32bit (num, color);

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    debug_color_32bit ("ring:", num, color, mask);
#endif
}


/*
 * Ring elements with 2 rings, the second ring mirrors the first (but at times
 * we can reverse the direction of the second ring).
 */

// set the color for a given pixel

void
Neopixel_ring_mirror::set_pixel_color_32bit_mask (Neopixel_index_t num,
						  Neopixel_rgb_t   color,
						  Neopixel_mask_t  mask)
{
  if (num >= num_pixels)
    {
#ifdef NEOPIXEL_DEBUG
      if (use_serial_output ())
	debug_color_32bit ("ring bad:   ", num, color, mask);
#endif
      return;
    }

  if ((mask & NEOPIXEL_MASK_MIRROR_ONLY) == 0)
    {
      set_pixel_color_32bit (num, color);

#ifdef NEOPIXEL_DEBUG
      if (use_serial_output ())
	debug_color_32bit ("ring main:  ", num, color, mask);
#endif
    }

  if ((mask & NEOPIXEL_MASK_MIRROR_NO) == 0)
    {
      Neopixel_index_t num2;

      if ((mask & NEOPIXEL_MASK_MIRROR_REVERSE) == 0)
	num2 = num;

      // If we have nested rings, we must reverse the pixels within the nesting ring
      else
	{
	  num2 = num_pixels - 1 - num;

	  for (int i = 0; i < num_nested_rings; i++)
	    {
	      Neopixel_index_t start = nested_rings[i].start;
	      Neopixel_index_t stop  = nested_rings[i].stop;

	      if (num >= start && num <= stop)
		{
		  Neopixel_index_t num_pixels_ring = stop - start + 1;
		  num2 = start + (num_pixels_ring - 1 - (num - start));
		  break;
		}
	    }
	}

      num2 += num_pixels;
      set_pixel_color_32bit (num2, color);

#ifdef NEOPIXEL_DEBUG
      if (use_serial_output ())
	debug_color_32bit ("ring mirror:", num2, color, mask);
#endif
    }
}

// Add nested rings

void
Neopixel_ring_mirror::add_nested_ring (Neopixel_index_t start,
				       Neopixel_index_t stop)
{
  if (num_nested_rings >= MAX_NESTED_RINGS - 1)
    {
#ifdef NEOPIXEL_DEBUG
      if (use_serial_output ())
	Serial.println ("Too many nested rings");
#endif
      return;
    }

  nested_rings[num_nested_rings].start = start;
  nested_rings[num_nested_rings].stop  = stop;

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    Serial.printf ("Add nested ring #%d, start = %d, stop = %d\n", num_nested_rings + 1, start, stop);
#endif

  num_nested_rings++;
  return;
}


/*
 * Stick elements.
 */

// set the color for a given pixel
void
Neopixel_stick::set_pixel_color_32bit_mask (Neopixel_index_t num,
					    Neopixel_rgb_t   color,
					    Neopixel_mask_t  mask)
{
  Neopixel_index_t num2 = stick_start + num;

  set_pixel_color_32bit (num2, color);

#ifdef NEOPIXEL_DEBUG
  if (use_serial_output ())
    debug_color_32bit ("stick:", num2, color, mask);
#endif
}


/*
 * Container class for the various Neopixel actions.
 */

Neopixel_action::Neopixel_action (Neopixel_iter_t	min_iter,
				  Neopixel_iter_t	max_iter,
				  Neopixel_power_t	bright,
				  unsigned long		dsmall,
				  unsigned long		dlarge)

  :	timeout		(Delay ()),
	neo_obj	 	((Neopixel_object *)0),
	delay_small	(dsmall),
	delay_large	(dlarge),
	iterations	(0),
	min_iterations	(min_iter),
	max_iterations	(max_iter),
	num_pixels	(0),
	brightness	(bright)
{
}

Neopixel_action::~Neopixel_action (void)
{
}


/*
 * HISTORY
 * $Log: Meissner_Neopixel.cpp,v $
 * Revision 1.79  2022/09/01 16:24:07  michaelmeissner
 * If debugging, trace explicitly overriding blink mode in flash_leds.
 *
 * Revision 1.78  2022/08/29 02:27:43  michaelmeissner
 * Fix spelling of Ws2812serial* types; Add begin() support for WS2812Serial lights.
 *
 * Revision 1.77  2022/08/28 21:25:10  michaelmeissner
 * Initial implementation to add WS2812Serial support.
 *
 * Revision 1.76  2022/07/24 22:52:13  michaelmeissner
 * Remove test for having blink pin.
 *
 * Revision 1.75  2022/05/16 00:41:56  michaelmeissner
 * Use Meissner_Blink for blinking the LED.
 *
 * Revision 1.74  2022/04/18 03:58:38  michaelmeissner
 * Only do SPI transactions on Teeny.
 *
 * Revision 1.73  2022/02/28 00:36:59  michaelmeissner
 * Add use_serial_output.
 *
 * Revision 1.72  2022/02/13 01:06:28  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.71  2020/01/28 04:17:38  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.70  2019/09/03 16:08:33  michaelmeissner
 * Add print statement for Teensy 4 to say we need to update the support.
 *
 * Revision 1.69  2019/08/16 04:52:04  michaelmeissner
 * Don't do PORT_PCR_MUX stuff on Teensy 4.0
 *
 * Revision 1.68  2019/07/16 14:37:32  michaelmeissner
 * Add nested ring support to mirror rings.
 *
 * Revision 1.67  2019/01/14 04:38:41  michaelmeissner
 * Update copyright.
 *
 * Revision 1.66  2017/10/11 15:33:51  michaelmeissner
 * Make fill functions deal with mirrored rings.
 *
 * Revision 1.65  2017/10/09 05:41:21  michaelmeissner
 * Add option to do a short delay after doing a beginTransaction.
 *
 * Revision 1.64  2017/10/06 05:51:11  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.63  2017/10/06 05:24:12  michaelmeissner
 * Restructure last code slightly.
 *
 * Revision 1.62  2017/10/06 05:20:06  michaelmeissner
 * Use code from KurtE to allow using the prop_shield level shifters, and reseting SPI mode.
 *
 * Revision 1.61  2016/11/06 13:51:43  michaelmeissner
 * Delete special handling in ring_mirror for mask_none
 *
 * Revision 1.60  2016/11/05 05:37:40  michaelmeissner
 * Allow ring_mirror num to be >= num_pixels if mirror mask == none.
 *
 * Revision 1.59  2016/11/05 05:24:22  michaelmeissner
 * Move pack functions to *.h; Improve debug output; Change setPixelColor in action to set_pixel_color_32bit_mask.
 *
 * Revision 1.58  2016/11/05 04:42:01  michaelmeissner
 * Fix typo.
 *
 * Revision 1.57  2016/11/05 04:41:11  michaelmeissner
 * Fix typo in ring mirror 32-bit color
 *
 * Revision 1.56  2016/11/05 04:40:09  michaelmeissner
 * Convert set_pixel_color_rgb, set_pixel_color_rgbw to set_pixel_color_32bit.
 *
 * Revision 1.55  2016/11/05 04:36:54  michaelmeissner
 * Move lowlevel_abort out of Neopixel_lowlevel to abort_with_message & call it more often.
 *
 * Revision 1.54  2016/11/05 01:14:15  michaelmeissner
 * Delete 1s delay in debug output #2.
 *
 * Revision 1.53  2016/11/05 01:12:12  michaelmeissner
 * Delete 1s delay in debug output.
 *
 * Revision 1.52  2016/11/05 01:11:52  michaelmeissner
 * Shorten debug statement
 *
 * Revision 1.51  2016/11/05 01:10:34  michaelmeissner
 * Set pin_led = -1 if it is the same as the neopixel pins; Print pin_neopixel2 in debug; Move some statements before debug.
 *
 * Revision 1.50  2016/11/04 15:09:33  michaelmeissner
 * Rename neopixel object -> neo_obj.
 *
 * Revision 1.49  2016/11/04 14:50:11  michaelmeissner
 * Rewrite code to avoid warnings.
 *
 * Revision 1.48  2016/11/04 14:47:35  michaelmeissner
 * Spacing.
 *
 * Revision 1.47  2016/11/04 14:45:52  michaelmeissner
 * Delete test for a variable not yet set.
 *
 * Revision 1.46  2016/11/04 14:42:40  michaelmeissner
 * Delete nfirst argument from pair_init.
 *
 * Revision 1.45  2016/11/04 14:33:48  michaelmeissner
 * Add lowlevel_abort
 *
 * Revision 1.44  2016/11/04 14:20:43  michaelmeissner
 * Move all of neopixel lowlevel support into Neopixel_lowlevel
 *
 * Revision 1.43  2016/11/04 14:15:27  michaelmeissner
 * Add max_pixels to the Neopixel_lowlevel support.
 *
 * Revision 1.42  2016/11/04 13:05:20  michaelmeissner
 * Rework neopixel_object begin.
 *
 * Revision 1.41  2016/11/04 06:05:55  michaelmeissner
 * Add support for Neopixel_pair; Use Neopixel_{single,pair} exlusively; Add more debugging prints.
 *
 * Revision 1.40  2016/10/31 14:02:11  michaelmeissner
 * For now, set neopixel as well as lowlevel.
 *
 * Revision 1.39  2016/10/31 01:59:23  michaelmeissner
 * Add lowlevel support to abstract the neopixels
 *
 * Revision 1.38  2016/10/24 13:16:09  michaelmeissner
 * Eliminate ring_mirror variable.
 *
 * Revision 1.37  2016/10/24 02:16:40  michaelmeissner
 * If split neopixel, call show for both parts; Print more debug information.
 *
 * Revision 1.36  2016/10/24 01:28:31  michaelmeissner
 * Add initial support for split rings (2 separate neopixel structures)
 *
 * Revision 1.35  2016/08/09 13:19:11  michaelmeissner
 * If we have an enable pin, explicitly set it to OUTPUT mode in constructor.
 *
 * Revision 1.34  2016/08/09 13:18:07  michaelmeissner
 * Add support for prop_shield to disable SPI bus during neopixels and enable the LEDs.
 *
 * Revision 1.33  2016/03/21 14:00:29  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.32  2016/03/18 15:45:45  michaelmeissner
 * Add initial support for RGBW neopixels with a 4th white LED.
 *
 * Revision 1.31  2015/02/15 19:14:21  michaelmeissner
 * Initial version
 *
 * Revision 1.30  2015/02/15 16:28:11  michaelmeissner
 * Add a space.
 *
 * Revision 1.29  2015/02/15 16:27:41  michaelmeissner
 * Move Neopixel_action constructor for .h to .cpp.
 *
 * Revision 1.28  2015/02/15 16:23:25  michaelmeissner
 * Reorder constructor so -Wreorder is happy #2.
 *
 * Revision 1.27  2015/02/15 16:16:04  michaelmeissner
 * Reorder constructor so -Wreorder is happy.
 *
 * Revision 1.26  2015/01/13 05:02:10  michaelmeissner
 * Delete support for allocating Adafruit_NeoPixel dynamically.
 *
 * Revision 1.25  2015/01/12 02:40:55  michaelmeissner
 * Add support for saying whether this is a mirrored object.
 *
 * Revision 1.24  2015/01/11 01:51:40  michaelmeissner
 * Fix thinko in mirror rings.
 *
 * Revision 1.23  2015/01/10 03:08:59  michaelmeissner
 * Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
 *
 * Revision 1.22  2015/01/09 16:24:45  michaelmeissner
 * Explicitly set led_status in beginning Neopixel_object.
 *
 * Revision 1.21  2015/01/09 13:09:57  michaelmeissner
 * Fix spacing issue #2.
 *
 * Revision 1.20  2015/01/09 13:08:56  michaelmeissner
 * Fix spacing issue.
 *
 * Revision 1.19  2015/01/09 13:07:48  michaelmeissner
 * Add support on mirrored rings to only do color on one right or another.
 *
 * Revision 1.18  2015/01/09 04:26:43  michaelmeissner
 * Make blinking a standard part of the Neopixel_object.
 *
 * Revision 1.17  2015/01/07 05:21:07  michaelmeissner
 * Move Neopixel_{ring,ring_mirror,stick} setPixelColor from .h to .cpp.
 *
 * Revision 1.16  2015/01/06 12:40:20  michaelmeissner
 * Don't overload use of 'start'.
 *
 * Revision 1.15  2015/01/05 03:00:31  michaelmeissner
 * Add types for colors.
 *
 * Revision 1.14  2015/01/05 02:44:43  michaelmeissner
 * Add types for holding iterations and power level
 *
 * Revision 1.13  2015/01/05 02:30:14  michaelmeissner
 * Make delay types unsigned long.
 *
 * Revision 1.12  2015/01/03 20:32:27  michaelmeissner
 * Move begin functions to Meissner_Neopixel.cpp; Also only wait 3 seconds for the Serial monitor if debugging.
 *
 * Revision 1.11  2015/01/03 16:21:38  michaelmeissner
 * New function flash_leds to turn all leds on in bootup sequence.
 *
 * Revision 1.10  2015/01/02 19:25:59  michaelmeissner
 * Add fill_rgb function.
 *
 * Revision 1.9  2015/01/02 17:08:20  michaelmeissner
 * Add string for debugging.
 *
 * Revision 1.8  2015/01/02 16:59:42  michaelmeissner
 * Add debug functions.
 *
 * Revision 1.7  2015/01/02 13:56:21  michaelmeissner
 * Move show, blink into Meissner_Neopixel.cpp; Add Fill function.
 *
 * Revision 1.6  2015/01/01 19:13:49  michaelmeissner
 * Update copyright year
 *
 * Revision 1.5  2015/01/01 18:58:14  michaelmeissner
 * Add comments.
 *
 * Revision 1.4  2014/12/29 13:58:35  michaelmeissner
 * Rework to pass pin_neopixel, pin_led arguments.
 *
 * Revision 1.3  2014/12/28 19:02:52  michaelmeissner
 * Add support for dynamically allocating the Adafruit_NeoPixel type.
 *
 * Revision 1.2  2014/12/28 17:45:38  michaelmeissner
 * Add constructor/destructor
 *
 * Revision 1.1  2014/12/28 15:24:01  michaelmeissner
 * Initial version.
 *
 */
