// -*- arduino -*-
// Meissner neopixel (ws2812) wrapper library
//
// Copyright (C) 2014-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_NEOPIXEL_H
#define MEISSNER_NEOPIXEL_H	1

#include <stddef.h>
#include <stdlib.h>
#include <inttypes.h>

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Delay.h>
#include <Meissner_Blink.h>

#if PROCESSOR_TEENSY_ARM
#include <SPI.h>
#include <WS2812Serial.h>
#endif

//#if !defined(NEOPIXEL_DEBUG) && USE_SERIAL_OUTPUT
//#define NEOPIXEL_DEBUG	1				// if defined, turn on debugging
//#endif

/*
 * Basic types.
 */

typedef uint16_t	Neopixel_index_t;			// Basic type to hold the neopixel index
typedef int8_t		Neopixel_pin_t;				// Type to hold pin # (must be signed)
typedef uint16_t	Neopixel_mask_t;			// mask for options to pass on
typedef uint16_t	Neopixel_iter_t;			// Type to hold iterations
typedef uint8_t		Neopixel_power_t;			// Type to power (0..255) level
typedef uint32_t	Neopixel_rgb_t;				// Type to represent red/green/blue combined colors

#ifndef DEFAULT_BRIGHTNESS					// default brightness to use
#define DEFAULT_BRIGHTNESS		30
#endif

// What type of neopixel we are dealing with
typedef enum {
  NEOPIXEL_TYPE_UNKNOWN,					// unknown variant.
  NEOPIXEL_TYPE_RING,						// LED ring.
  NEOPIXEL_TYPE_STICK,						// LED stick.
  NEOPIXEL_TYPE_MATRIX,						// LED matrix.
  NEOPIXEL_TYPE_MAX						// possibly used to size arrays
} Neopixel_type_t;

const Neopixel_mask_t NEOPIXEL_MASK_NONE	    = 0x0000;	// No special options

// Mask patterns used by the neopixel objects
const Neopixel_mask_t NEOPIXEL_MASK_MIRROR_REVERSE  = 0x0001;	// On mirrored rings, reverse the 2nd ring's index
const Neopixel_mask_t NEOPIXEL_MASK_MIRROR_NO	    = 0x0002;	// On mirrored rings, only write to the main ring
const Neopixel_mask_t NEOPIXEL_MASK_MIRROR_ONLY	    = 0x0004;	// On mirrored rings, only write to the mirrored ring

// Mask patterns used by the action functions
const Neopixel_mask_t NEOPIXEL_MASK_MIRROR_RANDOM   = 0x8000;	// Randomly turn on NEOPIXEL_MASK_MIRROR_REVERSE

// Abort if something went wrong
void abort_with_message (const char *);

// Use serial output?
bool use_serial_output (void);


// Convert separate red, green, and blue to 32-bit color

inline Neopixel_rgb_t
pack_rgb (Neopixel_power_t red,
	  Neopixel_power_t green,
	  Neopixel_power_t blue)
{
  return ((Neopixel_rgb_t)red << 16) | ((Neopixel_rgb_t)green <<  8) | blue;
}

// Convert separate red, green, blue, and white to 32-bit color

inline Neopixel_rgb_t
pack_rgbw (Neopixel_power_t red,
	   Neopixel_power_t green,
	   Neopixel_power_t blue,
	   Neopixel_power_t white)
{
  return ((Neopixel_rgb_t)white << 24) | ((Neopixel_rgb_t)red << 16) | ((Neopixel_rgb_t)green <<  8) | blue;
}

// Return white color, depending on whether we have RGB or RGBW neopixels
inline Neopixel_rgb_t
pack_white (Neopixel_power_t white,
	    bool	     have_rgbw_p = false)
{
  if (have_rgbw_p)
    return ((Neopixel_rgb_t)white) << 24;

  else
    {
      Neopixel_rgb_t level = (white + 2) / 3;
      return pack_rgb (level, level, level);
    }
}


/*
 * Base class for the underlying low level neopixel support.  This is the class that abstracts
 * the neopixel interface.  Normally, it might contain a single neopixel element, but you
 * might have split neopixel objects where two (or more) Neopixel objects make up part of
 * the logical structure.
 */

class Neopixel_lowlevel {
public:
  virtual void show (void)				= 0;

  virtual void setPixelColor (Neopixel_index_t num,
			      Neopixel_power_t red,
			      Neopixel_power_t green,
			      Neopixel_power_t blue)	= 0;

  virtual void setPixelColor (Neopixel_index_t num,
			      Neopixel_power_t red,
			      Neopixel_power_t green,
			      Neopixel_power_t blue,
			      Neopixel_power_t white)	= 0;

  virtual void setPixelColor (Neopixel_index_t num,
			      Neopixel_rgb_t   color)	= 0;

  virtual Neopixel_index_t max_pixels (void)		= 0;

  virtual bool busy (void)				= 0;

  Neopixel_lowlevel  (void) {}
  ~Neopixel_lowlevel (void) {}
};


/*
 * Class for a single neopixel object.
 */

class Neopixel_single : public Neopixel_lowlevel {
private:
  Adafruit_NeoPixel	*neopixel;				// actual neopixel data type

public:
  Neopixel_single (Adafruit_NeoPixel *np = NULL)
    : Neopixel_lowlevel ()
  {
    neopixel = np;
  }

  ~Neopixel_single  (void)
  {
  }

  // Initialize neopixel_single with the real neopixel object
  void single_init (Adafruit_NeoPixel *np);

  // Show all of the neopixel LEDs
  void show (void);

  // Set a pixel to RGB or RGBW
  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue,
		      Neopixel_power_t white);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_rgb_t   color);

  Neopixel_index_t max_pixels (void);

  // Adafruit neopixels are always blocking
  bool busy (void)
  {
    return false;
  }
};


/*
 * Class for a pair of neopixel objects.
 */

class Neopixel_pair : public Neopixel_lowlevel {
private:
  Adafruit_NeoPixel	*neopixel1;				// first  neopixel
  Adafruit_NeoPixel	*neopixel2;				// second neopixel
  Neopixel_index_t	 pixels_in_first;			// # pixels in the first neopixel
  bool			 neopixel1_set;				// whether pixels were set in neopixel1 or ws2812_serial1
  bool			 neopixel2_set;				// whether pixels were set in neopixel2 or ws2812_serial2

public:
  Neopixel_pair (Neopixel_index_t   nfirst	= 0,
		 Adafruit_NeoPixel *np1		= NULL,
		 Adafruit_NeoPixel *np2		= NULL)
    : Neopixel_lowlevel ()
  {
    pixels_in_first = nfirst;
    neopixel1	    = np1;
    neopixel2	    = np2;
    neopixel1_set   = false;
    neopixel2_set   = false;
  }

  ~Neopixel_pair  (void)
  {
  }

  // Initialize neopixel_pair with the real neopixel object
  void pair_init (Adafruit_NeoPixel *np1,
		  Adafruit_NeoPixel *np2);

  // Show all of the neopixel LEDs
  void show (void);

  // Set a pixel to RGB or RGBW
  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue,
		      Neopixel_power_t white);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_rgb_t   color);

  Neopixel_index_t max_pixels (void);

  // Adafruit neopixels are always blocking
  bool busy (void)
  {
    return false;
  }

};


#if PROCESSOR_TEENSY_ARM

/*
 * Class for a single WS2812Serial object.
 */

class Ws2812serial_single : public Neopixel_lowlevel {
private:
  WS2812Serial		*ws2812serial;				// actual WS2812Serial data type

public:
  Ws2812serial_single (WS2812Serial *np = NULL)
    : Neopixel_lowlevel ()
  {
    ws2812serial = np;
  }

  ~Ws2812serial_single  (void)
  {
  }

  // Initialize Ws2812serial_single with the real WS2812Serial object
  void single_init (WS2812Serial *np);

  // Show all of the neopixel LEDs
  void show (void);

  // Set a pixel to RGB or RGBW
  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue,
		      Neopixel_power_t white);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_rgb_t   color);

  Neopixel_index_t max_pixels (void);

  // WS2812Serial doesn't actually implement busy
  bool busy (void)
  {
    return false;
  }
};


/*
 * Class for a pair of ws2812serial objects.
 */

class Ws2812serial_pair : public Neopixel_lowlevel {
private:
  WS2812Serial		*ws2812serial1;				// first  WS2812Serial
  WS2812Serial		*ws2812serial2;				// second WS2812Serial
  Neopixel_index_t	 pixels_in_first;			// # pixels in the first neopixel
  bool			 ws2812serial1_set;			// whether pixels were set in ws2812serial1
  bool			 ws2812serial2_set;			// whether pixels were set in ws2812serial2

public:
  Ws2812serial_pair (Neopixel_index_t    nfirst	= 0,
		     WS2812Serial	*np1	= NULL,
		     WS2812Serial	*np2	= NULL)
    : Neopixel_lowlevel ()
  {
    pixels_in_first	= nfirst;
    ws2812serial1	= np1;
    ws2812serial2	= np2;
    ws2812serial1_set	= false;
    ws2812serial2_set	= false;
  }

  ~Ws2812serial_pair  (void)
  {
  }

  // Initialize Ws2812serial_pair with the real neopixel object
  void pair_init (WS2812Serial *np1,
		  WS2812Serial *np2);

  // Show all of the neopixel LEDs
  void show (void);

  // Set a pixel to RGB or RGBW
  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_power_t red,
		      Neopixel_power_t green,
		      Neopixel_power_t blue,
		      Neopixel_power_t white);

  void setPixelColor (Neopixel_index_t num,
		      Neopixel_rgb_t   color);

  Neopixel_index_t max_pixels (void);

  // WS2812Serial doesn't actually implement busy
  bool busy (void)
  {
    return false;
  }

};
#endif


/*
 * Base class for neopixel objects.
 */

class Neopixel_object {
 protected:
  Neopixel_lowlevel	*lowlevel;				// pointer to low level neopixel support
  Neopixel_single	 single;				// default Neopixel_inner to use if only a single neopixel object
  Neopixel_pair		 pair;					// default Neopixel_inner to use if only a pair neopixel object

#if PROCESSOR_TEENSY_ARM
  Ws2812serial_single	 ws2812serial_single;			// default Neopixel_inner to use if only a single WS2812Serial object
  Ws2812serial_pair	 ws2812serial_pair;			// default Neopixel_inner to use if only a pair WS2812Serial object
#endif /* PROCESSOR_TEENSY_ARM  */

  const Neopixel_type_t	 type;					// what type of neopixel this is.
  const Neopixel_index_t num_pixels;				// number of neopixels that are logically present.
  const Neopixel_index_t total_pixels;				// total number of neopixels (doubled for mirror rings).
  const Neopixel_index_t second_pixel;				// for split rings, where the second pixel begins
  Neopixel_index_t	 cur_pixel;				// current pixel we are on.
  const Neopixel_pin_t	 pin_neopixel;				// pin the neopixel is on
  const Neopixel_pin_t	 pin_neopixel2;				// pin the split neopixel ring is on
  const Neopixel_pin_t	 pin_led;				// pin the LED is on for flashing
  Neopixel_pin_t	 pin_enable;				// pin to set high during neopixel show to enable LEDs
  bool			 led_state;				// whether the LED is on/off
  bool			 do_blink;				// whether to blink the LED or not
  bool			 have_mirror;				// whether do rings in opposite direction
  bool			 have_white_leds;			// whether the neopixels have white LEDs
  bool			 do_spi_reserve;			// do SPI.beginTransaction to prevent SPI bus conflicts (prop shield)
  unsigned long		 do_spi_delay;				// If non-0, delay added after SPI.beginTransaction
  int			 blink_level;				// Level to use for blinking
  Meissner_Blink	 blink_data;				// Blink structure

  // Set a pixel element to a color
  inline void set_pixel_color_32bit (Neopixel_index_t	num,
				     Neopixel_rgb_t	color)
  {
    if (num >= total_pixels)
      return;

    if (lowlevel)
      lowlevel->setPixelColor (num, color);

    else
      abort_with_message ("Error, set_pixel_color_32bit, no lowlevel neopixel set");
  }

  // debug function to trace setPixelColor
  void debug_color_32bit (const char *,
			  Neopixel_index_t,
			  Neopixel_rgb_t,
			  Neopixel_mask_t);

  // common function to merge the two different begin functions
  void begin_common (Neopixel_index_t cur);

 public:
  // constructor, destructor
  Neopixel_object (Neopixel_index_t num,			// number of logical pixels
		   Neopixel_index_t max,			// total pixels
		   Neopixel_pin_t   pin,			// pin neopixel is on
		   Neopixel_pin_t   led,			// pin led is on
		   Neopixel_type_t  t,				// type
		   bool		    mirror	= false,	// whether this is a mirrored object
		   bool		    white_led	= false,	// whether the neopixels are RGBW instead of RGB
		   bool		    spi_reserve	= false,	// do SPI.beginTransaction to prevent SPI bus conflicts (prop shield)
		   Neopixel_pin_t   enable	= -1,		// pin to set high during neopixel show to enable LEDs
		   Neopixel_pin_t   pin2	= -1,		// pin the secondary neopixel is on
		   unsigned long    spi_delay  = 0);		// delay after doing SPI.beginTransaction

  ~Neopixel_object (void);

  // Initialization, using an existing neopixel descriptor
  void begin (Adafruit_NeoPixel *np, Neopixel_index_t cur, bool do_begin = true);

  // Initialization, using two existing neopixel descriptors
  void begin (Adafruit_NeoPixel *np, Adafruit_NeoPixel *np2, Neopixel_index_t cur, bool do_begin);

#if PROCESSOR_TEENSY_ARM
  // Initialization, using an existing WS2812Serial descriptor
  void begin (WS2812Serial *np, Neopixel_index_t cur, bool do_begin = true);

  // Initialization, using two existing WS2812Serial descriptors
  void begin (WS2812Serial *np, WS2812Serial *np2, Neopixel_index_t cur, bool do_begin = true);
#endif	/* PROCESSOR_TEENSY_ARM  */

  // return # of pixels.
  Neopixel_index_t max_pixels (void)
  {
    if (lowlevel)
      return lowlevel->max_pixels ();

    else
      return 0;
  }

  // return current element.
  Neopixel_index_t current (void)
  {
    return cur_pixel;
  }

  // set current element.
  void set_current (Neopixel_index_t c)
  {
    cur_pixel = c;
  }

  // Return the next or previous pixels
  Neopixel_index_t next (bool advance = true)
  {
    Neopixel_index_t ret = (cur_pixel >= num_pixels-1) ? 0 : cur_pixel+1;

    if (advance)
      cur_pixel = ret;

    return ret;
  }

  Neopixel_index_t previous (bool advance = true)
  {
    Neopixel_index_t ret = (cur_pixel == 0 || cur_pixel >= num_pixels) ? num_pixels-1 : cur_pixel-1;

    if (advance)
      cur_pixel = ret;

    return ret;
  }

  // Advance to either the next or previous element
  void advance (bool forward = true)
  {
    if (forward)
      cur_pixel = (cur_pixel == num_pixels-1) ? 0 : cur_pixel+1;
    else
      cur_pixel = (cur_pixel == 0) ? num_pixels-1 : cur_pixel-1;
  }

  // set the color for a given pixel
  virtual void set_pixel_color_32bit_mask (Neopixel_index_t num,
					   Neopixel_rgb_t   color,
					   Neopixel_mask_t  mask)	= 0;

  void setPixelColor (Neopixel_index_t	num,
		      Neopixel_power_t	red,
		      Neopixel_power_t	green,
		      Neopixel_power_t	blue)
  {
    set_pixel_color_32bit_mask (num,
				pack_rgb (red, green, blue),
				NEOPIXEL_MASK_NONE);
  }

  void setPixelColor (Neopixel_index_t	num,
		      Neopixel_power_t	red,
		      Neopixel_power_t	green,
		      Neopixel_power_t	blue,
		      Neopixel_power_t	white)
  {
    set_pixel_color_32bit_mask (num,
				pack_rgbw (red, green, blue, white),
				NEOPIXEL_MASK_NONE);
  }


  void setPixelColor (Neopixel_index_t	num,
		      Neopixel_rgb_t	color)
  {
    set_pixel_color_32bit_mask (num,
				color,
				NEOPIXEL_MASK_NONE);
  }

  // Return white color, depending on whether we have RGB or RGBW neopixels
  inline Neopixel_rgb_t
  white_color (Neopixel_power_t white)
  {
    return pack_white (white, have_white_leds);
  }

  // Set a given pixel to white.  If we don't have RGBW leds, set R/G/W leds
  void
  setPixelWhite (Neopixel_index_t num,
		 Neopixel_power_t white)
  {
    set_pixel_color_32bit_mask (num,
				white_color (white),
				NEOPIXEL_MASK_NONE);
  }

  // Return pin LED is on
  Neopixel_pin_t led_pin (void)
  {
    return pin_led;
  }

  // Set whether we blink or not
  void set_blink (bool b = true)
  {
    do_blink = b;
  }

  // Set the blink level (color for neopixel blink)
  void set_blink_level (int level = 1)
  {
    blink_level = (level ? level : 1);
  }

  // Return if we are set up for blinking.
  bool get_blink (void)
  {
    return do_blink;
  }

  // Return the current blink level
  int get_blink_level (void)
  {
    return blink_level;
  }

  // display the pixels
  void show (void);
  void show (bool);

  // Blink the led to show progress
  void blink (void);

  // Set all neopixels to the same color
  void fill (Neopixel_rgb_t color = 0);

  void fill_rgb (Neopixel_power_t red   = 0,
		 Neopixel_power_t green = 0,
		 Neopixel_power_t blue  = 0);

  void fill_white (Neopixel_power_t white = 0);

  // Flash the LEDs as part of the bootup sequence
  void flash_leds (Neopixel_power_t power_level = 10,
		   unsigned long    timeout     = 1000UL);

  void flash_leds (Neopixel_power_t power_level,
		   unsigned long    timeout,
		   bool		    explicit_blink);

  // Return true if this is a mirrored object
  bool is_mirror (void)
  {
    return have_mirror;
  }

  // Return true if this is a RGBW neopixel with white leds
  bool is_rgbw (void)
  {
    return have_white_leds;
  }

  // Return spi reserve status
  bool is_spi_reserve (void)
  {
    return do_spi_reserve;
  }

  // Set spi reserve status, and optionally do a small delay after doing a SPI.beginTransaction.
  void set_spi_reserve (bool reserve,
			unsigned long spi_delay = 0)
  {
    do_spi_reserve = reserve;
    do_spi_delay = spi_delay;
  }

  // Return pin we need to set high to enable LEDs is on
  Neopixel_pin_t enable_pin (void)
  {
    return pin_enable;
  }

  // Set a given pin to true before doing neopixel show (i.e. pin 7 if using the prop shield)
  void set_enable_pin (Neopixel_pin_t pin)
  {
    pin_enable = pin;
    if (pin >= 0)
      pinMode (pin, OUTPUT);
  }

  // Return busy status
  bool busy (void)
  {
    return (lowlevel && lowlevel->busy());
  }
};


/*
 * Single ring element.
 */

class Neopixel_ring : public Neopixel_object {
 public:
  // constructor
  Neopixel_ring (Neopixel_index_t num,
		 Neopixel_pin_t   pin		= PIN_NEOPIXEL,
		 Neopixel_pin_t   led		= (PIN_LED >= 0
						   ? PIN_LED
						   : PIN_NEOPIXEL_SINGLE),
		 Neopixel_index_t start		= 0,
		 bool		  white_led	= false,
		 bool		  spi_reserve	= false,
		 Neopixel_pin_t   enable	= -1,
		 Neopixel_pin_t   pin2		= -1)

  : Neopixel_object (num, num, pin, led, NEOPIXEL_TYPE_RING, false, white_led, spi_reserve, enable, pin2)
  {
  }

  // destructor
  ~Neopixel_ring (void)
  {
  }

  // set the color for a given pixel
  void set_pixel_color_32bit_mask (Neopixel_index_t num,
				   Neopixel_rgb_t   color,
				   Neopixel_mask_t  mask);
};


/*
 * Ring elements with 2 rings, the second ring mirrors the first (but at times
 * we can reverse the direction of the second ring).
 */

const int MAX_NESTED_RINGS = 3;

class Neopixel_ring_mirror : public Neopixel_object {
 private:
  // Allow for rings that are nested like Adafruits triple nested ring
  int num_nested_rings;
  struct {
    Neopixel_index_t start;		// start of a nested ring
    Neopixel_index_t stop;		// end of a nested ring
  } nested_rings[MAX_NESTED_RINGS];

 public:
  // Double ring constructor where second ring mirrors the first
  Neopixel_ring_mirror (Neopixel_index_t num,
			Neopixel_pin_t   pin		= PIN_NEOPIXEL,
			Neopixel_pin_t   led		= (PIN_LED >= 0
							   ? PIN_LED
							   : PIN_NEOPIXEL_SINGLE),
			bool		 white_led	= false,
			bool		 spi_reserve	= false,
			Neopixel_pin_t   enable		= -1,
			Neopixel_pin_t   pin2		= -1)

  : Neopixel_object (num, 2*num, pin, led, NEOPIXEL_TYPE_RING, true, white_led, spi_reserve, enable, pin2),
    num_nested_rings (0)
  {
    for (int i = 0; i < MAX_NESTED_RINGS; i++)
      nested_rings[i].start = nested_rings[i].stop = 0;
  }

  // destructor
  ~Neopixel_ring_mirror (void)
  {
  }

  // set the color for a given pixel
  void set_pixel_color_32bit_mask (Neopixel_index_t num,
				   Neopixel_rgb_t   color,
				   Neopixel_mask_t  mask);

  // Add nested rings
  void add_nested_ring (Neopixel_index_t start, Neopixel_index_t stop);
};


/*
 * Stick elements.
 */

class Neopixel_stick : public Neopixel_object {
 private:
  Neopixel_index_t stick_start;

 public:
  // constructor
  Neopixel_stick (Neopixel_index_t num,
		  Neopixel_pin_t   pin		= PIN_NEOPIXEL,
		  Neopixel_pin_t   led		= (PIN_LED >= 0
						   ? PIN_LED
						   : PIN_NEOPIXEL_SINGLE),
		  Neopixel_index_t start	= 0,
		  bool		   white_led	= false,
		  bool		   spi_reserve	= false,
		  Neopixel_pin_t   enable	= -1,
		  Neopixel_pin_t   pin2		= -1)

    : Neopixel_object (num, num, pin, led, NEOPIXEL_TYPE_STICK, false, white_led, spi_reserve, enable, pin2),
      stick_start (start)
  {
  }

  // destructor
  ~Neopixel_stick (void)
  {
  }

  // set the color for a given pixel
  void set_pixel_color_32bit_mask (Neopixel_index_t num,
				   Neopixel_rgb_t   color,
				   Neopixel_mask_t  mask);
};


/*
 * Container class for the various Neopixel actions.
 */

class Neopixel_action {
 protected:
  Delay		   timeout;					// timeout structure
  Neopixel_object *neo_obj;					// neopixels to use
  unsigned long	   delay_small;					// small delay (between elements in a pattern)
  unsigned long	   delay_large;					// large delay (between patterns)
  Neopixel_iter_t  iterations;					// current number of iterations to use before returning false
  Neopixel_iter_t  min_iterations;				// minimum number of iterations to use before returning false
  Neopixel_iter_t  max_iterations;				// maximum number of iterations to use before returning false
  Neopixel_index_t num_pixels;					// number of neopixels that are logically present.
  Neopixel_power_t brightness;					// max brightness to use

 public:
  // constructor/destructor
  Neopixel_action (Neopixel_iter_t	min_iter	= 1,
		   Neopixel_iter_t	max_iter	= 1,
		   Neopixel_power_t	bright		= DEFAULT_BRIGHTNESS,
		   unsigned long	dsmall		= 125,
		   unsigned long	dlarge		= 3000);

  ~Neopixel_action (void);

  // Function to set up the actions, possibly setting iterations or brightness levels
  virtual void start (Neopixel_iter_t  iter   = 0,
		      Neopixel_power_t bright = DEFAULT_BRIGHTNESS) = 0;

  // Function called from the main loop function to possibly do the next action.
  // Return false, when all of the current actions are finished.
  // If do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.
  virtual bool loop (bool *do_show_p = NULL) = 0;

  // Return a string giving the name of the action for debugging
  virtual const char *name (void) = 0;

  // Return the number of logical pixels
  Neopixel_index_t max_pixels (void)
  {
    if (num_pixels == 0 && neo_obj)
      num_pixels = neo_obj->max_pixels ();

    return num_pixels;
  }

  // Function to set the brightness level
  void set_brightness (Neopixel_power_t bright)
  {
    brightness = bright;
  }

  // Function to return the current brightness level
  Neopixel_power_t get_brightness (void)
  {
    return brightness;
  }
};

#endif	/* MEISSNER_NEOPIXEL_H */


/*
 * HISTORY
 * $Log: Meissner_Neopixel.h,v $
 * Revision 1.97  2022/08/29 02:40:09  michaelmeissner
 * Properly use #ifdef PROCESSOR_TEENSY_ARM for non-Teensy systems.
 *
 * Revision 1.96  2022/08/29 02:25:56  michaelmeissner
 * WS2812Serial really doesn't implement a busy() function.
 *
 * Revision 1.95  2022/08/28 21:24:57  michaelmeissner
 * Initial implementation to add WS2812Serial support.
 *
 * Revision 1.94  2022/07/24 22:39:40  michaelmeissner
 * Add get_blink and get_blink_level.
 *
 * Revision 1.93  2022/05/16 02:01:12  michaelmeissner
 * Make code slightly clearer.
 *
 * Revision 1.92  2022/05/16 00:42:40  michaelmeissner
 * Insure blink_level is not 0.
 *
 * Revision 1.91  2022/05/16 00:41:41  michaelmeissner
 * Use Meissner_Blink for blinking the LED.
 *
 * Revision 1.90  2022/02/28 00:36:32  michaelmeissner
 * Add use_serial_output.
 *
 * Revision 1.89  2022/02/27 05:00:01  michaelmeissner
 * Set the default brightness.
 *
 * Revision 1.88  2022/02/26 06:47:46  michaelmeissner
 * Fix spacing.
 *
 * Revision 1.87  2022/02/13 01:01:15  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.86  2020/01/28 04:08:33  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.85  2020/01/17 13:08:07  michaelmeissner
 * Add {set,get}_brightness functions.
 *
 * Revision 1.84  2020/09/03 16:05:18  michaelmeissner
 * PROCESSOR_TEENSY_3_X => PROCESSOR_TEENSY_ARM.
 *
 * Revision 1.83  2019/07/16 14:37:47  michaelmeissner
 * Add nested ring support to mirror rings.
 *
 * Revision 1.82  2019/01/14 04:38:27  michaelmeissner
 * Update copyright.
 *
 * Revision 1.81  2018/12/28 00:35:11  michaelmeissner
 * Turn off setting neopixel by default.
 *
 * Revision 1.80  2018/12/28 00:31:39  michaelmeissner
 * Turn on neopixel tracing by default.
 *
 * Revision 1.79  2017/10/09 05:42:39  michaelmeissner
 * Add option to do a short delay after doing a beginTransaction.
 *
 * Revision 1.78  2017/10/06 05:50:54  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.77  2016/11/12 03:26:30  michaelmeissner
 * Change error message #2.
 *
 * Revision 1.76  2016/11/12 03:25:53  michaelmeissner
 * Change error message.
 *
 * Revision 1.75  2016/11/11 16:46:46  michaelmeissner
 * Make pack_{rgb,rgbw,white} shifts all 32-bit
 *
 * Revision 1.74  2016/11/11 15:15:13  michaelmeissner
 * Add pack_white inline to create the appropriate white color based on RGB/RGBW.
 *
 * Revision 1.73  2016/11/10 15:35:03  michaelmeissner
 * Fix next/previous in case current pixel is out of bounds.
 *
 * Revision 1.72  2016/11/05 15:37:33  michaelmeissner
 * Use consistant spacing
 *
 * Revision 1.71  2016/11/05 15:36:39  michaelmeissner
 * Remove mask last argument from SetPixelColor calls.
 *
 * Revision 1.70  2016/11/05 05:38:48  michaelmeissner
 * Move pack functions here from .cpp; Collapse separate red/green/blue functions into passing 32-bit color.
 *
 * Revision 1.69  2016/11/05 04:39:56  michaelmeissner
 * Convert set_pixel_color_rgb, set_pixel_color_rgbw to set_pixel_color_32bit.
 *
 * Revision 1.68  2016/11/05 04:36:40  michaelmeissner
 * Move lowlevel_abort out of Neopixel_lowlevel to abort_with_message & call it more often.
 *
 * Revision 1.67  2016/11/05 01:09:31  michaelmeissner
 * Delete neopixel_pin accessor function.
 *
 * Revision 1.66  2016/11/04 15:10:01  michaelmeissner
 * Rename neopixel object -> neo_obj.
 *
 * Revision 1.65  2016/11/04 14:43:00  michaelmeissner
 * Delete nfirst argument from pair_init.
 *
 * Revision 1.64  2016/11/04 14:33:34  michaelmeissner
 * Add lowlevel_abort
 *
 * Revision 1.63  2016/11/04 14:20:31  michaelmeissner
 * Move all of neopixel lowlevel support into Neopixel_lowlevel
 *
 * Revision 1.62  2016/11/04 14:15:16  michaelmeissner
 * Add max_pixels to the Neopixel_lowlevel support.
 *
 * Revision 1.61  2016/11/04 13:04:59  michaelmeissner
 * Rework neopixel_object begin.
 *
 * Revision 1.60  2016/10/31 01:59:10  michaelmeissner
 * Add lowlevel support to abstract the neopixels
 *
 * Revision 1.59  2016/10/24 13:16:18  michaelmeissner
 * Eliminate ring_mirror variable.
 *
 * Revision 1.58  2016/10/24 13:05:51  michaelmeissner
 * Eliminate 2nd ring_mirror constructor
 *
 * Revision 1.57  2016/10/24 12:53:54  michaelmeissner
 * Fix pixel number for split rings
 *
 * Revision 1.56  2016/10/24 02:15:52  michaelmeissner
 * Add new Neopixel_object arguments to the super classes.
 *
 * Revision 1.55  2016/10/24 01:28:46  michaelmeissner
 * Add initial support for split rings (2 separate neopixel structures)
 *
 * Revision 1.54  2016/08/09 13:41:49  michaelmeissner
 * Fix spelling of set_enable_pin.
 *
 * Revision 1.53  2016/08/09 13:39:27  michaelmeissner
 * If we have an enable pin, set pin mode.
 *
 * Revision 1.52  2016/08/09 13:17:49  michaelmeissner
 * Add support for prop_shield to disable SPI bus during neopixels and enable the LEDs.
 *
 * Revision 1.51  2016/04/03 22:46:23  michaelmeissner
 * Make (C) in copyright be upper case.
 *
 * Revision 1.50  2016/03/21 14:00:14  michaelmeissner
 * Update copyright year.
 *
 * Revision 1.49  2016/03/18 15:45:59  michaelmeissner
 * Add initial support for RGBW neopixels with a 4th white LED.
 *
 * Revision 1.48  2015/02/15 16:27:25  michaelmeissner
 * Move Neopixel_action constructor for .h to .cpp.
 *
 * Revision 1.47  2015/02/15 16:12:00  michaelmeissner
 * Reorder constructor so -Wreorder is happy.
 *
 * Revision 1.46  2015/01/13 05:00:10  michaelmeissner
 * Delete support for allocating Adafruit_NeoPixel dynamically.
 *
 * Revision 1.45  2015/01/12 05:33:23  michaelmeissner
 * Add virtual function to return the action name.
 *
 * Revision 1.44  2015/01/12 02:40:39  michaelmeissner
 * Add support for saying whether this is a mirrored object.
 *
 * Revision 1.43  2015/01/12 02:18:45  michaelmeissner
 * Make virtual functions in the Neopixel_object base case pure virtual functions.
 *
 * Revision 1.42  2015/01/10 22:41:23  michaelmeissner
 * Add optional argument to loop to return if show should be called.
 *
 * Revision 1.41  2015/01/10 03:08:27  michaelmeissner
 * Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
 *
 * Revision 1.40  2015/01/09 12:58:06  michaelmeissner
 * Fix typo; Add NEOPIXEL_MASK_MIRROR_{NO,ONLY}.
 *
 * Revision 1.39  2015/01/09 12:45:00  michaelmeissner
 * Add num_pixels field to Neopixel_action.
 *
 * Revision 1.38  2015/01/09 04:26:57  michaelmeissner
 * Make blinking a standard part of the Neopixel_object.
 *
 * Revision 1.37  2015/01/07 05:20:53  michaelmeissner
 * Move Neopixel_{ring,ring_mirror,stick} setPixelColor from .h to .cpp.
 *
 * Revision 1.36  2015/01/06 13:42:42  michaelmeissner
 * Add mask bit for randomly setting mirror reverse.
 *
 * Revision 1.35  2015/01/06 13:17:29  michaelmeissner
 * Correct off by one error in mirrored rings.
 *
 * Revision 1.34  2015/01/05 04:11:42  michaelmeissner
 * Fix typo #2
 *
 * Revision 1.33  2015/01/05 03:07:09  michaelmeissner
 * Fix typo.
 *
 * Revision 1.32  2015/01/05 02:58:59  michaelmeissner
 * Add types for colors.
 *
 * Revision 1.31  2015/01/05 02:43:55  michaelmeissner
 * Add types for holding iterations and power level
 *
 * Revision 1.30  2015/01/05 02:29:24  michaelmeissner
 * Make delay types unsigned long.
 *
 * Revision 1.29  2015/01/05 00:43:43  michaelmeissner
 * Add Neopixel_action base class.
 *
 * Revision 1.28  2015/01/03 20:31:44  michaelmeissner
 * Move begin functions to Meissner_Neopixel.cpp
 *
 * Revision 1.27  2015/01/03 18:50:56  michaelmeissner
 * Open serial line, and print intro debug messages if NEOPIXEL_DEBUG.
 *
 * Revision 1.26  2015/01/03 16:21:52  michaelmeissner
 * New function flash_leds to turn all leds on in bootup sequence.
 *
 * Revision 1.25  2015/01/02 19:55:35  michaelmeissner
 * Undo setting brightness, since latest Adafruit libraries now properly set it.
 *
 * Revision 1.24  2015/01/02 18:20:32  michaelmeissner
 * Add fill_rgb function.
 *
 * Revision 1.23  2015/01/02 18:09:03  michaelmeissner
 * Go back to setting brightness to 255.
 *
 * Revision 1.22  2015/01/02 18:05:12  michaelmeissner
 * Update a comment.
 *
 * Revision 1.21  2015/01/02 18:03:39  michaelmeissner
 * Set neopixel brightness to 0.
 *
 * Revision 1.20  2015/01/02 17:52:05  michaelmeissner
 * Set neopixel brightness to high
 *
 * Revision 1.19  2015/01/02 17:08:02  michaelmeissner
 * Add string for debugging.
 *
 * Revision 1.18  2015/01/02 17:00:08  michaelmeissner
 * Flesh out debug functions.
 *
 * Revision 1.17  2015/01/02 16:44:50  michaelmeissner
 * Add more debugging.
 *
 * Revision 1.16  2015/01/02 16:40:22  michaelmeissner
 * Add neopixel_pin/led_pin functions; Add debugging for neopixel_ring.
 *
 * Revision 1.15  2015/01/02 13:56:03  michaelmeissner
 * Move show, blink into Meissner_Neopixel.cpp; Add Fill function.
 *
 * Revision 1.14  2015/01/01 19:13:32  michaelmeissner
 * Update copyright year
 *
 * Revision 1.13  2015/01/01 19:03:29  michaelmeissner
 * Move test for calling low level begin lower.
 *
 * Revision 1.12  2015/01/01 18:59:23  michaelmeissner
 * Don't reallocate neopixel structure if it is already present.
 *
 * Revision 1.11  2014/12/29 15:27:13  michaelmeissner
 * Change reverse option on setPixelColor to mask, allowing for other options.
 *
 * Revision 1.10  2014/12/29 14:22:46  michaelmeissner
 * Add blink support.
 *
 * Revision 1.9  2014/12/29 13:58:11  michaelmeissner
 * Rework to pass pin_neopixel, pin_led arguments.
 *
 * Revision 1.8  2014/12/28 19:29:31  michaelmeissner
 * Set cur_pixel in begin if not passed.
 *
 * Revision 1.7  2014/12/28 19:23:57  michaelmeissner
 * Add helper function to return the base adafruit neopixel descriptor
 *
 * Revision 1.6  2014/12/28 19:22:12  michaelmeissner
 * Deal with new returnning NULL
 *
 * Revision 1.5  2014/12/28 19:21:19  michaelmeissner
 * Begin call now has argument for setting current element.
 *
 * Revision 1.4  2014/12/28 19:17:32  michaelmeissner
 * Add support for total pixels being different from logical number of pixels.
 *
 * Revision 1.3  2014/12/28 19:07:55  michaelmeissner
 * Add support for dynamically allocating the Adafruit_NeoPixel type; Check for neopixel being null.
 *
 * Revision 1.2  2014/12/28 17:45:56  michaelmeissner
 * Flesh out initial implementation.
 *
 * Revision 1.1  2014/12/28 15:23:28  michaelmeissner
 * Initial version.
 *
 */
