// -*- arduino -*-
// Simple chase pattern, fill a ring with a background color, find a random
// pixel and set it to the foreground color.  Have the pixel go around the ring
// for a few times, setting the pixel to be the foreground color and reseting
// the previous spot to the background color.  After a few times, no longer
// reset the previous spot until the whole ring is filled with the foreground
// color.
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Chase.h>
#include <Meissner_Delay.h>

// Possible color choices
const size_t NUM_COLORS	= 7;

static Neopixel_rgb_t colors[NUM_COLORS];


// Constructor, destructor
Neopixel_chase::Neopixel_chase (Neopixel_iter_t	 max_iter,
				Neopixel_iter_t	 min_iter,
				Neopixel_power_t bright,
				unsigned long	 delay_inner,
				unsigned long	 delay_whole)

  : Neopixel_action (max_iter, min_iter, bright, delay_inner, delay_whole),
    mask		(NEOPIXEL_MASK_MIRROR_RANDOM),
    first_pixel		(0),
    state		(CHASE_STATE_INITIAL),
    background		(0),
    foreground		(0),
    max_single_count	(CHASE_MAX_SINGLE_COUNT),
    min_single_count	(CHASE_MIN_SINGLE_COUNT),
    single_cycle	(0),
    current_cycle	(0)
{
}

Neopixel_chase::~Neopixel_chase (void)
{
}


// Setup for chase.  This is expected to be called only once.

void
Neopixel_chase::begin (Neopixel_object *neo,
		       Neopixel_power_t	chase_brightness,
		       Neopixel_mask_t	chase_mask,
		       Neopixel_iter_t	chase_min_iterations,
		       Neopixel_iter_t	chase_max_iterations,
		       Neopixel_iter_t	chase_min_single_count,
		       Neopixel_iter_t	chase_max_single_count,
		       unsigned long	chase_delay_inner,
		       unsigned long	chase_delay_whole)
{
  neo_obj = neo;
  mask = chase_mask;

  if (chase_brightness)
    brightness = chase_brightness;

  if (chase_min_iterations)
    min_iterations = chase_min_iterations;

  if (chase_max_iterations)
    max_iterations = chase_max_iterations;

  if (chase_min_single_count)
    min_single_count = chase_min_single_count;

  if (chase_max_single_count)
    max_single_count = chase_max_single_count;

  if (chase_delay_inner)
    delay_small = chase_delay_inner;

  if (chase_delay_whole)
    delay_large = chase_delay_whole;

  state		= CHASE_STATE_INITIAL;
  current_cycle = 0;
  iterations	= random (min_iterations, max_iterations+1);

  if (neo_obj)
    num_pixels = neo_obj->max_pixels ();

#ifdef CHASE_DEBUG
#ifndef NEOPIXEL_DEBUG
    Serial.begin (9600);
    delay (1000);

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;
#endif

    Serial.println ("==========");
    Serial.println ("Chase begin");
#endif
}


// Setup for the chase.  This is expected to be called each time we switch to doing the chase
// sequence to reset the information, and start the timer.

void
Neopixel_chase::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  state	     = CHASE_STATE_INITIAL;
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);

  if (bright)
    brightness = bright;

  // setup the colors now
  colors[0] = pack_rgb (brightness,	0,		0);			// all red
  colors[1] = pack_rgb (0,		brightness,	0);			// all green
  colors[2] = pack_rgb (0,		0,		brightness);		// all blue
  colors[3] = pack_rgb (brightness/2,	brightness/2,	0);			// red+green
  colors[4] = pack_rgb (brightness/2,	0,		brightness/2);		// red+blue
  colors[5] = pack_rgb (0,		brightness/2,	brightness/2);		// green+blue
  colors[6] = pack_rgb (brightness/3,	brightness/3,	brightness/3);		// white

  // start the timer to go off imediately on the first call to loop
  timeout.run_immediate ();

#ifdef CHASE_DEBUG
    Serial.println ("==========");
    Serial.print   ("Chase start, iterations = ");
    Serial.print   (iterations);
    Serial.print   (", brightness = ");
    Serial.println (brightness);
#endif
}


// Simple chase test
// If do_show_p is not a NULL pointer, the function will not call show, but
// instead set *do_show_p to true.  It is assumed that the user has set
// *do_show_p  to false before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool
Neopixel_chase::loop (bool *do_show_p)
{
  size_t f, b;
  Neopixel_index_t pixel;
  bool ret = true;

  if (!neo_obj)
    return false;

  if (neo_obj->busy ())
    return true;

  if (timeout.delay_done ())
    {
      // set up next timeout
      timeout.run_delay (delay_small);

      switch (state)
	{
	  // Initial state, set up defaults
	case CHASE_STATE_INITIAL:
	  current_cycle  = 0;
	  single_cycle	 = random (min_single_count, max_single_count+1);
	  first_pixel	 = random (0, num_pixels);
	  state		 = CHASE_STATE_SINGLE_PIXEL;

	  do
	    {
	      f = random (0, NUM_COLORS);
	      b = random (0, NUM_COLORS);
	    }
	  while (f == b);

	  foreground = colors[f];
	  background = colors[b];

	  neo_obj->set_current (first_pixel);
	  neo_obj->fill (background);
	  neo_obj->set_pixel_color_32bit_mask (first_pixel, foreground, mask);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

	  if ((mask & NEOPIXEL_MASK_MIRROR_RANDOM) != 0)
	    {
	      if (random (0, 2) == 0)
		mask &= ~NEOPIXEL_MASK_MIRROR_REVERSE;
	      else
		mask |= NEOPIXEL_MASK_MIRROR_REVERSE;
	    }

#ifdef CHASE_DEBUG
	  Serial.println ("==========");
	  Serial.print   ("Chase setup, pixels = ");
	  Serial.println (num_pixels);
	  Serial.print   ("delay small = ");
	  Serial.print   (delay_small);
	  Serial.print   (", delay large = ");
	  Serial.print   (delay_large);
	  Serial.print   (", led pin = ");
	  Serial.println (neo_obj->led_pin ());
	  Serial.print   ("foreground = 0x");
	  Serial.print   (f);
	  Serial.print   (" (");
	  Serial.print   (foreground, HEX);
	  Serial.print   (")");
	  Serial.print   (", background = 0x");
	  Serial.print   (b);
	  Serial.print   (" (");
	  Serial.print   (background, HEX);
	  Serial.println (")");
	  Serial.print   ("iterations = ");
	  Serial.print   (iterations);
	  Serial.print   (", first pixel = ");
	  Serial.print   (first_pixel);
	  Serial.print   (", flip = ");
	  Serial.print   (((mask & NEOPIXEL_MASK_MIRROR_REVERSE) != 0) ? "true" : "false");
	  Serial.print   (", single cycle = ");
	  Serial.println (single_cycle);
#endif
	  // fall through to doing single pixel

	  // set a single pixel, resetting the previous one
	case CHASE_STATE_SINGLE_PIXEL:
	  pixel = neo_obj->current ();

#ifdef CHASE_DEBUG
	  Serial.print ("Chase single pixel, pixel = ");
	  Serial.print (pixel);
	  Serial.print (", foreground = 0x");
	  Serial.print (foreground, HEX);
	  Serial.print (", background = 0x");
	  Serial.print (background, HEX);
	  Serial.println ("");
#endif

	  neo_obj->set_pixel_color_32bit_mask (pixel, background, mask);
	  pixel = neo_obj->next (true);
	  neo_obj->set_pixel_color_32bit_mask (pixel, foreground, mask);
	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

	  if (pixel == first_pixel)
	    {
	      if (--single_cycle == 0)
		{
		  state = CHASE_STATE_ALL_PIXELS;

#ifdef CHASE_DEBUG
		  Serial.println ("==========");
		  Serial.println ("Go to all pixels state");
#endif
		}
	      else
		{
#ifdef CHASE_DEBUG
		  Serial.println ("==========");
		  Serial.print   ("Single pixel, single_cycle = ");
		  Serial.println (single_cycle);
#endif
		}
	    }

	  break;

	  // Set all pixels for the final iteration
	case CHASE_STATE_ALL_PIXELS:
	  pixel = neo_obj->next ();
	  neo_obj->set_pixel_color_32bit_mask (pixel, foreground, mask);
	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

#ifdef CHASE_DEBUG
	  Serial.print ("Chase all pixels, pixel = ");
	  Serial.print (pixel);
	  Serial.print (", foreground = 0x");
	  Serial.print (foreground, HEX);
	  Serial.println ("");
#endif

	  if (pixel == first_pixel)
	    {
	      state = CHASE_STATE_DONE;
	      timeout.run_delay (delay_large);

#ifdef CHASE_DEBUG
	      Serial.println ("==========");
	      Serial.println ("Go to final state");
#endif
	    }

	  break;

	  // Done, reset things and set up for next iteration
	case CHASE_STATE_DONE:
	  iterations--;
	  state = CHASE_STATE_INITIAL;
	  neo_obj->fill (0);
	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

#ifdef CHASE_DEBUG
	  Serial.println ("==========");
	  Serial.print ("Chase end of cycle, iterations = ");
	  Serial.println (iterations);
#endif

	  if (iterations == 0)
	    {
	      iterations = random (min_iterations, max_iterations+1);
	      ret	 = false;
	    }

	  break;

	  // Unknown, just bail out
	default:
	  ret = false;
#ifdef CHASE_DEBUG
	  Serial.println ("==========");
	  Serial.println ("bad case");
#endif
	  break;
	}
    }

#ifdef CHASE_DEBUG
  if (!ret)
    {
      Serial.println ("==========");
      Serial.println ("Chase returns false");
    }
#endif
      
  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Chase.cpp,v $
// Revision 1.30  2022/08/28 21:43:25  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.29  2022/02/13 01:05:07  michaelmeissner
// Update copyright year.
//
// Revision 1.28  2020/01/28 04:14:52  michaelmeissner
// Update copyright year.
//
// Revision 1.27  2019/01/14 04:39:24  michaelmeissner
// Update copyright.
//
// Revision 1.26  2016/11/10 15:45:46  michaelmeissner
// Switch to using pack_rgb instead of our own version.
//
// Revision 1.25  2016/11/10 15:38:11  michaelmeissner
// Delete printing neopixel_pin, it no longer exists.
//
// Revision 1.24  2016/11/05 12:10:52  michaelmeissner
// SetPixelColor -> set_pixel_color_32bit_mask
//
// Revision 1.23  2016/11/04 15:09:04  michaelmeissner
// Rename neopixel object -> neo_obj.
//
// Revision 1.22  2016/04/03 22:45:11  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.21  2016/03/21 14:01:35  michaelmeissner
// Update copyright year.
//
// Revision 1.20  2015/02/15 17:29:48  michaelmeissner
// Reorder constructor to avoid -Wreorder warnings.
//
// Revision 1.19  2015/02/01 15:09:45  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.18  2015/01/11 14:30:00  michaelmeissner
// Change timeout functions.
//
// Revision 1.17  2015/01/10 22:49:33  michaelmeissner
// Fix spacing.
//
// Revision 1.16  2015/01/10 22:47:33  michaelmeissner
// Fix typo.
//
// Revision 1.15  2015/01/10 22:46:32  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.14  2015/01/10 03:16:24  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.13  2015/01/09 12:48:23  michaelmeissner
// Use num_pixels field.
//
// Revision 1.12  2015/01/09 05:43:12  michaelmeissner
// correct off by one errors in random calls.
//
// Revision 1.11  2015/01/09 04:36:43  michaelmeissner
// Remove show/loop parameter do_blink.
//
// Revision 1.10  2015/01/06 13:41:26  michaelmeissner
// Add mask bit for randomly setting mirror reverse.
//
// Revision 1.9  2015/01/05 04:17:50  michaelmeissner
// Avoid overloading 'start'.
//
// Revision 1.8  2015/01/05 03:04:27  michaelmeissner
// Add types for colors.
//
// Revision 1.7  2015/01/05 02:50:00  michaelmeissner
// Add types for holding iterations and power level
//
// Revision 1.6  2015/01/05 02:35:06  michaelmeissner
// Make delay types unsigned long.
//
// Revision 1.5  2015/01/05 00:45:06  michaelmeissner
// Make Neopixel_chase a superclass of Neopixel_action.
//
// Revision 1.4  2015/01/03 20:34:06  michaelmeissner
// Wait up to 3 seconds for the Serial monitor if debugging.
//
// Revision 1.3  2015/01/03 06:09:44  michaelmeissner
// Implement delay_whole
//
// Revision 1.2  2015/01/02 17:51:16  michaelmeissner
// print 0x in debugging hex values
//
// Revision 1.1  2015/01/02 17:14:49  michaelmeissner
// Initial version.
//
