// -*- arduino -*-
// Simple chase pattern, fill a ring with a background color, find a random
// pixel and set it to the foreground color.  Have the pixel go around the ring
// for a few times, setting the pixel to be the foreground color and reseting
// the previous spot to the background color.  After a few times, no longer
// reset the previous spot until the whole ring is filled with the foreground
// color.
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_NEOPIXEL_CHASE_H
#define MEISSNER_NEOPIXEL_CHASE_H	1

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Delay.h>

// Enable tracing to the Serial device
// #define CHASE_DEBUG	1

// Various tunings and constants
const unsigned long	CHASE_DELAY_INNER	= 125;	// timeout between changing the case pixel
const unsigned long	CHASE_DELAY_WHOLE	= 3000;	// timeout after acheiving a single color
const Neopixel_iter_t	CHASE_MAX_ITERATIONS	= 5;	// max # of iterations before going to the next pattern
const Neopixel_iter_t	CHASE_MIN_ITERATIONS	= 2;	// min # of iterations before going to the next pattern
const Neopixel_iter_t	CHASE_MAX_SINGLE_COUNT	= 8;	// max # of iterations of a single pixel
const Neopixel_iter_t	CHASE_MIN_SINGLE_COUNT	= 3;	// min # of iterations of a single pixel
const Neopixel_power_t	CHASE_BRIGHTNESS	= 10;	// max brightness to use

typedef enum {
  CHASE_STATE_INITIAL,					// initial state
  CHASE_STATE_SINGLE_PIXEL,				// changing a single pixel
  CHASE_STATE_ALL_PIXELS,				// changing all of the pixels
  CHASE_STATE_DONE					// done the whole cycle
} Chase_state_t;

class Neopixel_chase : Neopixel_action {
 private:
  Neopixel_mask_t	mask;				// mask to pass to neopixel_object
  Neopixel_index_t	first_pixel;			// first pixel that was changed
  Chase_state_t		state;				// current state
  Neopixel_rgb_t	background;			// background color
  Neopixel_rgb_t	foreground;			// foreground color
  Neopixel_iter_t	max_single_count;		// max # of iterations of a single pixel
  Neopixel_iter_t	min_single_count;		// min # of iterations of a single pixel
  Neopixel_iter_t	single_cycle;			// # of cycles with the single pixel
  Neopixel_iter_t	current_cycle;			// cycle # for doing the single pixel

 public:
  // constructor, destructor
  Neopixel_chase (Neopixel_iter_t	max_iter	= CHASE_MAX_ITERATIONS,
		  Neopixel_iter_t	min_iter	= CHASE_MIN_ITERATIONS,
		  Neopixel_power_t	bright		= CHASE_BRIGHTNESS,
		  unsigned long		delay_inner	= CHASE_DELAY_INNER,
		  unsigned long		delay_whole	= CHASE_DELAY_WHOLE);

  ~Neopixel_chase (void);

  // Setup for chase
  void begin (Neopixel_object  *neo,
	      Neopixel_power_t	chase_brightness	= CHASE_BRIGHTNESS,
	      Neopixel_mask_t	chase_mask		= NEOPIXEL_MASK_MIRROR_RANDOM,
	      Neopixel_iter_t	chase_min_iterations	= CHASE_MIN_ITERATIONS,
	      Neopixel_iter_t	chase_max_iterations	= CHASE_MAX_ITERATIONS,
	      Neopixel_iter_t	chase_min_single_count	= CHASE_MIN_SINGLE_COUNT,
	      Neopixel_iter_t	chase_max_single_count	= CHASE_MAX_SINGLE_COUNT,
	      unsigned long	chase_delay_inner	= CHASE_DELAY_INNER,
	      unsigned long	chase_delay_whole	= CHASE_DELAY_WHOLE);


  // Function to set up the actions, possibly setting iterations or brightness levels
  void start (Neopixel_iter_t iter = 0, Neopixel_power_t bright = 0);

  // Function called from the main loop function to possibly do the next
  // action.  Return false, when all of the current actions are finished.  If
  // do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.  It is assumed that the user has set
  // *do_show_p to false before the call.
  bool loop (bool *do_show_p = NULL);

  // Return a string giving the name of the action for debugging
  const char *name (void)
  {
    return "Chase";
  }
};

#endif	/* MEISSNER_NEOPIXEL_CHASE_H */


// HISTORY
// $Log: Meissner_Neopixel_Chase.h,v $
// Revision 1.19  2022/02/13 00:59:53  michaelmeissner
// Update copyright year.
//
// Revision 1.18  2020/01/28 04:06:35  michaelmeissner
// Update copyright year.
//
// Revision 1.17  2019/01/14 04:39:04  michaelmeissner
// Update copyright.
//
// Revision 1.16  2016/04/03 22:44:52  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.15  2016/03/21 14:00:58  michaelmeissner
// Update copyright year.
//
// Revision 1.14  2015/02/01 15:11:46  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.13  2015/01/12 05:34:24  michaelmeissner
// Add virtual function to return the action name.
//
// Revision 1.12  2015/01/10 22:44:11  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.11  2015/01/10 03:16:04  michaelmeissner
// Re-add copyright line
//
// Revision 1.10  2015/01/10 03:15:29  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.9  2015/01/09 12:49:47  michaelmeissner
// Delete max_pixels field in favor of num_pixels.
//
// Revision 1.8  2015/01/09 04:35:51  michaelmeissner
// Remove show/loop parameter do_blink.
//
// Revision 1.7  2015/01/06 13:41:43  michaelmeissner
// Add mask bit for randomly setting mirror reverse.
//
// Revision 1.6  2015/01/05 03:01:14  michaelmeissner
// Add types for colors.
//
// Revision 1.5  2015/01/05 02:48:00  michaelmeissner
// Add types for holding iterations and power level
//
// Revision 1.4  2015/01/05 02:34:02  michaelmeissner
// Make delay types unsigned long.
//
// Revision 1.3  2015/01/05 00:44:40  michaelmeissner
// Make Neopixel_chase a superclass of Neopixel_action.
//
// Revision 1.2  2015/01/02 17:15:27  michaelmeissner
// Turn off debugging.
//
// Revision 1.1  2015/01/02 17:14:25  michaelmeissner
// Initial version.
//
