// -*- arduino -*-
// Meissner neopixel simple color rotate library
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Color_Rotate.h>
#include <Meissner_Delay.h>


// Constructor, destructor
Neopixel_color_rotate::Neopixel_color_rotate (Neopixel_power_t	bright,
					      Neopixel_mask_t	mask,
					      Neopixel_iter_t	max_iter,
					      Neopixel_iter_t	min_iter,
					      unsigned long	delay_amount)

      :	Neopixel_action (max_iter, min_iter, bright, delay_amount, 0),
	mask_current	(mask),
	color		(1),
	updown_p	(false),
	direction_up	(true)
{
}

Neopixel_color_rotate::~Neopixel_color_rotate (void)
{
}


// begin function to initialize everything.
// This is expected to be called only once.

void
Neopixel_color_rotate::begin (Neopixel_object  *neo,
			      bool		updown,
			      Neopixel_mask_t	mask,
			      Neopixel_power_t	bright,
			      Neopixel_iter_t	min_iter,
			      Neopixel_iter_t	max_iter,
			      unsigned long	delay_amount)
  {
    color          = 1;
    neo_obj	   = neo;
    updown_p	   = updown;
    direction_up   = true;

    if (mask)
      mask_current = mask;

    if (bright)
      brightness = bright;

    if (min_iter)
      min_iterations = min_iter;

    if (max_iter)
      max_iterations = max_iter;

    if (delay_amount)
      delay_small = delay_amount;

#ifdef COLOR_ROTATE_DEBUG
#ifndef NEOPIXEL_DEBUG
    Serial.begin (9600);
    delay (1000);

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;
#endif

    Serial.println ("==========");
    Serial.println ("Color rotate begin");
    Serial.print   ("min_iterations = ");	Serial.println (min_iterations);
    Serial.print   ("max_iterations = ");	Serial.println (max_iterations);
    Serial.print   ("brightness     = ");	Serial.println (brightness);
    Serial.print   ("delay amount   = ");	Serial.println (delay_amount);
    Serial.print   ("mask_current   = 0x");	Serial.println (mask_current, HEX);
    Serial.print   ("updown_p       = ");	Serial.println (updown_p ? "true" : "false");
    Serial.print   ("neo_obj        = 0x");	Serial.println ((size_t)neo_obj, HEX);
#endif

    if (neo)
      {
	num_pixels = neo_obj->max_pixels ();
	neo_obj->fill (0);
	neo_obj->show ();
      }

    // Setup in case start is not called.
    iterations = random (min_iterations, max_iterations+1);

    // start the timer to go off imediately on the first call to loop
    timeout.run_immediate ();
  }


// Setup for color rotate.  This is expected to be called each time we switch to doing the rotate
// sequence to reset the information, and start the timer.

void
Neopixel_color_rotate::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);

  if (bright)
    brightness = bright;

  // start the timer to go off imediately on the first call to loop
  timeout.run_immediate ();

#ifdef COLOR_ROTATE_DEBUG
    Serial.println ("==========");
    Serial.println ("Color rotate start");
    Serial.print   ("iterations     = ");	Serial.println (iterations);
    Serial.print   ("brightness     = ");	Serial.println (brightness);
#endif
}


// loop function to be called from the main loop, return false when all
// iterations are done If do_show_p is not a NULL pointer, the function will
// not call show, but instead set *do_show_p to true.  It is assumed that the
// user has set *do_show_p to false before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool Neopixel_color_rotate::loop (bool *do_show_p)
{
  Neopixel_power_t	red	= ((color & 0x1) != 0) * brightness;
  Neopixel_power_t	green	= ((color & 0x2) != 0) * brightness;
  Neopixel_power_t	blue	= ((color & 0x4) != 0) * brightness;
  Neopixel_index_t	pixel	= neo_obj->current ();
  bool			ret	= true;

  if (neo_obj && neo_obj->busy ())
    return true;

  if (timeout.delay_done ())
    {
      // set up next timeout
      timeout.run_delay (delay_small);

      neo_obj->set_pixel_color_32bit_mask (pixel, pack_rgb (red, green, blue), mask_current);
      if (do_show_p)
	*do_show_p = true;
      else
	neo_obj->show ();

      // Go to next element based on whether we are ascending or descending
      // If we are at the boundary, switch direction and color
      if (direction_up)
	{
	  pixel = neo_obj->next (true);
	  if (pixel == 0)
	    {
	      color = (color + 1) & 0x7;

	      // If we get back to color 1, decrement the iteration count.
	      // also possibly set the mirror reversal bit randomly
	      if (color == 1)
		{
		  if ((mask_current & NEOPIXEL_MASK_MIRROR_RANDOM) != 0)
		    {
		      if (random (0, 2) == 0)
			mask_current &= ~NEOPIXEL_MASK_MIRROR_REVERSE;
		      else
			mask_current |= NEOPIXEL_MASK_MIRROR_REVERSE;
		    }

		  if (--iterations == 0)
		    {
		      ret = false;
		      iterations = max_iterations;
		    }
		}

	      // If we are alternating  going up/down, start going down
	      if (updown_p)
		{
		  neo_obj->set_current (num_pixels - 1);
		  direction_up = false;
		}

#ifdef COLOR_ROTATE_DEBUG
	      Serial.print ("color = ");
	      Serial.print (color);
	      Serial.print (", iterations = ");
	      Serial.print (iterations);
	      Serial.print (", updown_p = ");
	      Serial.print (updown_p ? "true" : "false");
	      Serial.print (", direction = ");
	      Serial.print (direction_up ? "up" : "down");
	      Serial.print (", ret = ");
	      Serial.println (ret ? "true" : "false");
#endif
	    }
	}
      else
	{
	  pixel = neo_obj->previous (true);
	  if (pixel == num_pixels - 1)
	    {
	      color = (color + 1) & 0x7;
	      neo_obj->set_current (0);
	      direction_up = true;

	      // If we get back to color 1, decrement the iteration count.
	      if (color == 1)
		{
		  if (--iterations == 0)
		    {
		      ret = false;
		      iterations = max_iterations;
		    }
		}

#ifdef COLOR_ROTATE_DEBUG
	      Serial.print ("color = ");
	      Serial.print (color);
	      Serial.print (", iterations = ");
	      Serial.print (iterations);
	      Serial.print (", updown_p = ");
	      Serial.print (updown_p ? "true" : "false");
	      Serial.print (", direction = ");
	      Serial.print (direction_up ? "up" : "down");
	      Serial.print (", ret = ");
	      Serial.println (ret ? "true" : "false");
#endif
	    }
	}
    }

  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Color_Rotate.cpp,v $
// Revision 1.38  2022/08/28 21:45:50  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.37  2022/02/13 01:05:24  michaelmeissner
// Update copyright year.
//
// Revision 1.36  2020/01/28 04:15:00  michaelmeissner
// Update copyright year.
//
// Revision 1.35  2019/01/14 04:40:07  michaelmeissner
// Update copyright.
//
// Revision 1.34  2016/11/05 12:36:30  michaelmeissner
// SetPixelColor -> set_pixel_color_32bit_mask
//
// Revision 1.33  2016/11/04 15:08:52  michaelmeissner
// Rename neopixel object -> neo_obj.
//
// Revision 1.32  2016/04/03 22:45:58  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.31  2016/03/21 14:02:17  michaelmeissner
// Update copyright year.
//
// Revision 1.30  2015/02/15 16:13:53  michaelmeissner
// Reorder constructor so -Wreorder is happy.
//
// Revision 1.29  2015/02/01 15:11:04  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.28  2015/01/11 14:33:10  michaelmeissner
// Change timeout functions.
//
// Revision 1.27  2015/01/10 22:48:48  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.26  2015/01/10 03:22:15  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.25  2015/01/09 12:54:44  michaelmeissner
// Delete num_pixels, superclass now defines it.  Call show with no arguments.
//
// Revision 1.24  2015/01/09 05:42:09  michaelmeissner
// correct off by one errors in random calls.
//
// Revision 1.23  2015/01/09 04:29:39  michaelmeissner
// Remove loop parameter do_blink.
//
// Revision 1.22  2015/01/08 14:02:59  michaelmeissner
// Change order of constructor and add mask.
//
// Revision 1.21  2015/01/06 13:42:25  michaelmeissner
// Add mask bit for randomly setting mirror reverse.
//
// Revision 1.20  2015/01/06 13:23:54  michaelmeissner
// Fix thinko.
//
// Revision 1.19  2015/01/05 04:19:13  michaelmeissner
// Don't call start in begin; Avoid overloading 'start'.
//
// Revision 1.18  2015/01/05 03:05:42  michaelmeissner
// Add types for colors.
//
// Revision 1.17  2015/01/05 02:53:27  michaelmeissner
// Add types for holding iterations and power level
//
// Revision 1.16  2015/01/05 02:32:15  michaelmeissner
// Make delay types unsigned long.
//
// Revision 1.15  2015/01/05 01:09:37  michaelmeissner
// Make Neopixel_color_rotate a superclass of Neopixel_action.
//
// Revision 1.14  2015/01/03 20:49:50  michaelmeissner
// Add serial debug support.
//
// Revision 1.13  2015/01/02 14:22:15  michaelmeissner
// Move timeout structure into Neopixel_color_rotate class.
//
// Revision 1.12  2015/01/02 13:55:27  michaelmeissner
// Clear neopixels in begin().
//
// Revision 1.11  2015/01/01 23:35:50  michaelmeissner
// Use timeout run immediately option.
//
// Revision 1.10  2015/01/01 19:14:32  michaelmeissner
// Update copyright year
//
// Revision 1.9  2014/12/31 14:17:59  michaelmeissner
// Move setPixelColor/show into if delay_done statement
//
// Revision 1.8  2014/12/31 13:54:31  michaelmeissner
// Move constructor/destructor into .cpp file; Use Meissner_Delay.h.
//
// Revision 1.7  2014/12/29 15:26:16  michaelmeissner
// Pass mask to setPixelColor
//
// Revision 1.6  2014/12/29 14:32:28  michaelmeissner
// Fix typo.
//
// Revision 1.5  2014/12/29 14:31:06  michaelmeissner
// Allow blink to be turned on/off
//
// Revision 1.4  2014/12/29 14:23:48  michaelmeissner
// Move blink support into Meissner_Neopixel.
//
// Revision 1.3  2014/12/29 13:59:16  michaelmeissner
// Update type.
//
// Revision 1.2  2014/12/29 07:11:45  michaelmeissner
// Rework going up/down support.
//
// Revision 1.1  2014/12/29 06:31:42  michaelmeissner
// Initial version
//
