// -*- arduino -*-
// Meissner neopixel simple color rotate library
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_NEOPIXEL_COLOR_ROTATE_H
#define MEISSNER_NEOPIXEL_COLOR_ROTATE_H	1

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Delay.h>

// #define COLOR_ROTATE_DEBUG	1			// define to enable debugging

const unsigned long	COLOR_ROTATE_DELAY		= 125UL;
const Neopixel_power_t	COLOR_ROTATE_BRIGHTNESS		= 10;
const Neopixel_iter_t	COLOR_ROTATE_MIN_ITERATIONS	= 2;
const Neopixel_iter_t	COLOR_ROTATE_MAX_ITERATIONS	= 10;

class Neopixel_color_rotate : Neopixel_action {
 private:
  Neopixel_mask_t		mask_current;		// current mask to pass to setPixelColor
  uint8_t			color;			// color to use (0..7)
  bool				updown_p;		// if we are doing up/down processing
  bool				direction_up;		// if we are currently ascending

 public:
  // constructor/destructor
  Neopixel_color_rotate (Neopixel_power_t bright	= COLOR_ROTATE_BRIGHTNESS,
			 Neopixel_mask_t  mask		= NEOPIXEL_MASK_MIRROR_RANDOM,
			 Neopixel_iter_t  max_iter	= COLOR_ROTATE_MAX_ITERATIONS,
			 Neopixel_iter_t  min_iter	= COLOR_ROTATE_MIN_ITERATIONS,
			 unsigned long	  delay_amount	= COLOR_ROTATE_DELAY);

  ~Neopixel_color_rotate (void);

  // begin function to initialize everything
  void begin (Neopixel_object  *neo,
	      bool		updown			= false,
	      Neopixel_mask_t	mask			= 0,
	      Neopixel_power_t	bright			= 0,
	      Neopixel_iter_t	min_iter		= 0,
	      Neopixel_iter_t	max_iter		= 0,
	      unsigned long	delay_amount		= 0);

  // Function to set up the actions, possibly setting iterations or brightness levels
  void start (Neopixel_iter_t iter = 0, Neopixel_power_t bright = 0);

  // Function called from the main loop function to possibly do the next action.
  // Return false, when all of the current actions are finished.
  // If do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.  It is assumed that the user has set
  // *do_show_p  to false before the call.
  bool loop (bool *do_show_p = NULL);

  // Return a string giving the name of the action for debugging
  const char *name (void)
  {
    return "Color_Rotate";
  }
};

#endif	/* MEISSNER_NEOPIXEL_COLOR_ROTATE_H */


// HISTORY
// $Log: Meissner_Neopixel_Color_Rotate.h,v $
// Revision 1.25  2022/02/13 01:00:16  michaelmeissner
// Update copyright year.
//
// Revision 1.24  2020/01/28 04:06:46  michaelmeissner
// Update copyright year.
//
// Revision 1.23  2019/01/14 04:39:46  michaelmeissner
// Update copyright.
//
// Revision 1.22  2016/04/03 22:45:40  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.21  2016/03/21 14:02:00  michaelmeissner
// Update copyright year.
//
// Revision 1.20  2015/02/01 15:10:20  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.19  2015/01/12 05:34:50  michaelmeissner
// Add virtual function to return the action name.
//
// Revision 1.18  2015/01/10 22:47:13  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.17  2015/01/10 03:17:01  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.16  2015/01/09 04:28:55  michaelmeissner
// Remove loop parameter do_blink.
//
// Revision 1.15  2015/01/08 14:02:05  michaelmeissner
// Change order of constructor and add mask.
//
// Revision 1.14  2015/01/08 14:00:11  michaelmeissner
// Set defaults for begin to 0 to not change the constructor.
//
// Revision 1.13  2015/01/06 13:42:05  michaelmeissner
// Add mask bit for randomly setting mirror reverse.
//
// Revision 1.12  2015/01/05 02:52:05  michaelmeissner
// Add types for holding iterations and power level
//
// Revision 1.11  2015/01/05 02:31:22  michaelmeissner
// Make delay types unsigned long.
//
// Revision 1.10  2015/01/05 01:09:54  michaelmeissner
// Make Neopixel_color_rotate a superclass of Neopixel_action.
//
// Revision 1.9  2015/01/03 20:49:23  michaelmeissner
// Add serial debug support.
//
// Revision 1.8  2015/01/02 14:22:32  michaelmeissner
// Move timeout structure into Neopixel_color_rotate class.
//
// Revision 1.7  2015/01/01 19:14:14  michaelmeissner
// Update copyright year
//
// Revision 1.6  2014/12/31 13:54:05  michaelmeissner
// Move constructor/destructor into .cpp file; Use Meissner_Delay.h.
//
// Revision 1.5  2014/12/29 15:26:43  michaelmeissner
// Add mask option.
//
// Revision 1.4  2014/12/29 14:30:31  michaelmeissner
// Allow blink to be turned on/off
//
// Revision 1.3  2014/12/29 14:23:32  michaelmeissner
// Move blink support into Meissner_Neopixel.
//
// Revision 1.2  2014/12/29 07:11:56  michaelmeissner
// Rework going up/down support.
//
// Revision 1.1  2014/12/29 06:31:52  michaelmeissner
// Initial version
//
