// -*- c++ -*-
// Adafruit strandtest:
// https://github.com/adafruit/Adafruit_NeoPixel/blob/master/examples/strandtest/strandtest.ino
// this code was completely rewritten by Michael Meissner (arduino@the-meissners.org)

// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Colorwipe.h>
#include <Meissner_Delay.h>


// Constructor, destructor
Neopixel_colorwipe::Neopixel_colorwipe (Neopixel_power_t bright,
					Neopixel_mask_t	 mask,
					Neopixel_iter_t  min_iter,
					Neopixel_iter_t  max_iter,
					unsigned long    delay_led)

  : Neopixel_action (min_iter, max_iter, bright, 0, delay_led),
    current_mask  (mask),
    current_color (pack_rgb (bright, 0, 0))				// set to red
{
}

Neopixel_colorwipe::~Neopixel_colorwipe (void)
{
}


// Setup for colorwipe.  This is expected to be called only once.

void
Neopixel_colorwipe::begin (Neopixel_object  *neo,
			   Neopixel_power_t  bright,
			   Neopixel_mask_t   mask,
			   Neopixel_iter_t   min_iter,
			   Neopixel_iter_t   max_iter,
			   unsigned long     delay_led)
{
  neopixel = neo;

  if (bright)
    brightness = bright;

  if (min_iter)
    min_iterations = min_iter;

  if (max_iter)
    max_iterations = max_iter;

  if (delay_led)
    delay_large = delay_led;

  iterations = random (min_iterations, max_iterations+1);

  if (neopixel)
    num_pixels = neopixel->max_pixels ();

#ifdef COLORWIPE_DEBUG
#ifndef NEOPIXEL_DEBUG
    Serial.begin (9600);
    delay (1000);

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;
#endif

    Serial.println ("==========");
    Serial.println ("Colorwipe begin");
#endif
}


// Setup for the colorwipe.  This is expected to be called each time we switch
// to doing the colorwipe sequence to reset the information, and start the
// timer.

void
Neopixel_colorwipe::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);

  if (bright)
    brightness = bright;

  current_color = pack_rgb (brightness, 0, 0);		// red color

  // start the timer to go off imediately on the first call to loop
  timeout.run_immediate ();

  // clear LEDs
  if (neopixel)
    {
      neopixel->set_current (0);
      neopixel->fill (0);
    }

#ifdef SOLID_COLOR_DEBUG
    Serial.println ("==========");
    Serial.print   ("Colorwipe, iterations = ");
    Serial.print   (iterations);
    Serial.print   (", brightness = ");
    Serial.print   (brightness);
#endif
}


// Simple colorwipe test If do_show_p is not a NULL pointer, the function will
// not call show, but instead set *do_show_p to true.  It is assumed that the
// user has set *do_show_p to false before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool
Neopixel_colorwipe::loop (bool *do_show_p)
{
  bool ret = true;

  if (!neopixel)
    return false;

  if (neopixel->busy ())
    return true;

  if (timeout.delay_done ())
    {
      Neopixel_index_t pos = neopixel->current ();

      // set up next timeout
      timeout.run_delay (delay_large);

      if (pos == 0)
	neopixel->fill (0);

      // Write the current pixel
      neopixel->set_pixel_color_32bit_mask (pos, current_color, current_mask);

      if (do_show_p)
	*do_show_p = true;
      else
	neopixel->show ();

      // Set up for the next pixel, if we've wrapped around, go to the next pixel
      if (neopixel->next (true) == 0)
	{
	  current_color >>= 8;
	  if (current_color == 0)
	    {
	      current_color = pack_rgb (brightness, 0, 0);
	      if (--iterations == 0)
		{
		  ret = false;
		  iterations = random (min_iterations, max_iterations);
		}
	    }
	}

#ifdef COLORWIPE_DEBUG
      Serial.println ("==========");
      Serial.print   ("Colorwipe, delay = ");
      Serial.print   (delay_large);
      Serial.print   (", current color = 0x");
      Serial.print   (current_color);
      Serial.print   (", iterations = ");
      Serial.print   (iterations);
      Serial.print   (", ret = ");
      Serial.println (ret ? "true" : "false");
#endif
    }
      
  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Colorwipe.cpp,v $
// Revision 1.10  2022/08/28 21:47:11  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.9  2022/02/13 01:05:47  michaelmeissner
// Update copyright year.
//
// Revision 1.8  2020/01/28 04:15:09  michaelmeissner
// Update copyright year.
//
// Revision 1.7  2019/01/14 04:40:43  michaelmeissner
// Update copyright.
//
// Revision 1.6  2016/11/05 12:14:53  michaelmeissner
// Use pack_rgb in Meissner_Neopixel.h
//
// Revision 1.5  2016/11/05 12:13:08  michaelmeissner
// SetPixelColor -> set_pixel_color_32bit_mask
//
// Revision 1.4  2016/03/21 14:03:05  michaelmeissner
// Update copyright year.
//
// Revision 1.3  2015/02/15 17:31:22  michaelmeissner
// Reorder constructor to avoid -Wreorder warnings.
//
// Revision 1.2  2015/02/02 03:21:52  michaelmeissner
// Add meissner copyright and notice about it being rewritten.
//
// Revision 1.1  2015/02/02 03:19:08  michaelmeissner
// Initial version.
//
// Revision 1.2  2015/02/01 15:20:37  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.1  2015/01/16 15:18:46  michaelmeissner
// Initial version.
//
