// -*- arduino -*-
// Run random neopixel patterns for a time, before going on to the next for two
// rings.
//
// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Neopixel_Combined_Ring.h>

// Setup function
void Neopixel_combined_ring::setup (void)
{
  setup_called = true;

#if USE_SERIAL_OUTPUT
  if (use_serial_output ())
    {
      Serial.begin (9600);
      if (millis () < 1000L)
	delay (1000L);

      // wait for the USB initialization, but don't wait forever if we are
      // using a battery.
      while (!Serial && millis () < 3000)
	;

      print_info ();
    }
#endif

  Random_mask_t mask_random, mask_linear, mask_first;

  if (button_action == BUTTON_ACTION_SEQUENTIAL)
    {
      // do all patterns sequentially in linear mode
      mask_random = RANDOM_MASK_ALL;
      mask_linear = RANDOM_MASK_ALL;
      mask_first  = RANDOM_MASK_ALL;
    }
  else if (button_action == BUTTON_ACTION_SOLID)
    {
      // only do the boring solid color pattern when the button is pushed
      mask_random = RANDOM_MASK_ALL & ~0x1;
      mask_linear = 0x1;
      mask_first  = 0;
    }
  else
    {
      // always do random patterns
      mask_random = RANDOM_MASK_ALL;
      mask_linear = 0;
      mask_first  = 0;
    }

  set_random_seed (true);

  // Set up the various sub-data structures
  if (dual_rings)
    {
      pixels_dual.begin (&np, 0, true);
      pixels = &pixels_dual;
    }
  else
    {
      pixels_single.begin (&np, 0, true);
      pixels = &pixels_single;
    }

  pixels_chase.begin         (pixels);
  pixels_color.begin         (pixels);
  pixels_k_eye.begin	     (pixels);
  pixels_roulette.begin      (pixels);
  pixels_solid_color.begin   (pixels);
  pixels_theater_chase.begin (pixels);

  // Set up the random actions
  pixels_random.begin  (pixels,						// neopixel object to use
			0,						// override delay
			pin_button,					// button to change the actions
			mask_random,					// mask of actions to do randomly
			mask_linear,					// mask of actions to do when the button is pressed
			mask_first,					// mask of actions to do in the first iteration
			(Neopixel_action *) &pixels_solid_color,	// first action (must be solid color)
			(Neopixel_action *) &pixels_chase,		// second action
			(Neopixel_action *) &pixels_color,		// third action
			(Neopixel_action *) &pixels_k_eye,		// fourth action
			(Neopixel_action *) &pixels_roulette,		// fifth action
			(Neopixel_action *) &pixels_theater_chase);	// sixth action

#if USE_SERIAL_OUTPUT
  if (use_serial_output ())
    {
      pixels_random.set_print (true);
      Serial.println ("Flash LEDs");
    }
#endif

  pixels->flash_leds (30, 1000);

#if USE_SERIAL_OUTPUT
  if (use_serial_output ())
    Serial.println ("Start random patterns");
#endif

  pixels_random.start ();
}


// Loop function
void Neopixel_combined_ring::loop (void)
{
  pixels_random.loop ();
}


// Print the initial information to the USB terminal
void Neopixel_combined_ring::print_info (void)
{
#if USE_SERIAL_OUTPUT
  if (use_serial_output ())
    {
      const char *action;

      switch (button_action)
	{
	case BUTTON_ACTION_RANDOM:	action = "random";	break;
	case BUTTON_ACTION_SEQUENTIAL:	action = "sequential";	break;
	case BUTTON_ACTION_SOLID:	action = "solid";	break;
	default:			action = "unknown";	break;
	}
      
      Serial.println ("Combined ring initialization.");

#if USE_SERIAL_PRINTF
      Serial.printf ("Number of pixels = %2d\n",  num_pixels);
      Serial.printf ("Neopixel pin     = %2d\n",  pin_neopixel);
      Serial.printf ("Button pin       = %2d\n",  pin_button);
      Serial.printf ("Led pin          = %2d\n",  pin_led);
      Serial.printf ("Neopixel type    = 0x%x\n", neopixel_type);
      Serial.printf ("Button action    = %s\n",   action);
      Serial.printf ("Dual rings       = %s\n",   dual_rings ? "yes" : "no");

#else
      Serial.print ("Number of pixels = ");   Serial.println (num_pixels);
      Serial.print ("Neopixel pin     = ");   Serial.println (pin_neopixel);
      Serial.print ("Button pin       = ");   Serial.println (pin_button);
      Serial.print ("Led pin          = ");   Serial.println (pin_led);
      Serial.print ("Neopixel type    = 0x"); Serial.println (neopixel_type, HEX);
      Serial.print ("Button action    = ");   Serial.println (action);
      Serial.print ("Dual rings       = ");   Serial.println (dual_rings ? "yes" : "no");
#endif

      Serial.println ("");
    }
#endif
}


// HISTORY
// $Log: Meissner_Neopixel_Combined_Ring.cpp,v $
// Revision 1.5  2022/09/01 15:09:58  michaelmeissner
// Use use_serial_output within setup.
//
// Revision 1.4  2022/02/28 02:24:02  michaelmeissner
// Use use_serial_output () function and USE_SERIAL_PRINTF define.
//
// Revision 1.3  2022/02/13 06:29:57  michaelmeissner
// Iterate on fleshing out code #2.
//
// Revision 1.2  2022/02/13 04:04:31  michaelmeissner
// Iterate on fleshing out code.
//
// Revision 1.1  2022/02/13 02:58:50  michaelmeissner
// Initial file.
//
