// -*- arduino -*-
// Run random neopixel patterns for a time, before going on to the next for two
// rings.
//
// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_NEOPIXEL_COMBINED_RING_H
#define MEISSNER_NEOPIXEL_COMBINED_RING_H	1

#include <Meissner_Config.h>
#include <Meissner_Cpu_Info.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Random.h>
#include <Meissner_Neopixel_Chase.h>
#include <Meissner_Neopixel_Color_Rotate.h>
#include <Meissner_Neopixel_Kaleidoscope_Eyes.h>
#include <Meissner_Neopixel_Roulette.h>
#include <Meissner_Neopixel_Solid_Color.h>
#include <Meissner_Neopixel_Theater_Chase.h>
#include <Meissner_Delay.h>
#include <Meissner_Random.h>
#include <Bounce2.h>

// Defaults.
#ifndef DEFAULT_NUM_PIXELS
#define DEFAULT_NUM_PIXELS		16				// default ring size
#endif

#ifndef DEFAULT_PIN_NEOPIXEL
#define DEFAULT_PIN_NEOPIXEL		PIN_EXTERNAL_NEOPIXEL		// default pin neopixels are on
#endif

#ifndef DFEAULT_PIN_BUTTON
#define DEFAULT_PIN_BUTTON		PIN_BUTTON			// default button that changes the action
#endif

#ifndef DFEAULT_PIN_LED
#define DEFAULT_PIN_LED			PIN_LED				// default output pin used to show progress
#endif

#ifndef DEFAULT_DUAL_RINGS
#define DEFAULT_DUAL_RINGS		true
#endif

#ifndef DEFAULT_NEOPIXEL_TYPE
#define DEFAULT_NEOPIXEL_TYPE		(NEO_GRB + NEO_KHZ800)		// default color order and speed of the neopixel
#endif

// Button actions
enum combined_ring_action {
  BUTTON_ACTION_RANDOM,							// button press selects a new random pattern
  BUTTON_ACTION_SEQUENTIAL,						// button press goes through actions sequentially
  BUTTON_ACTION_SOLID							// button press toggles doing only solid color pattern (i.e. 'boring')
};

#ifndef DEFAULT_BUTTON_ACTION
#define DEFAULT_BUTTON_ACTION		BUTTON_ACTION_SOLID		// default button action
#endif

// Class for calling the combined ring
class Neopixel_combined_ring {
 private:
  int				 num_pixels;				// number of pixels in a single ring
  enum combined_ring_action	 button_action;				// what does the button do
  int				 pin_neopixel;				// pin the neopixels are on
  int				 pin_button;				// pin the button to change actions are on
  int				 pin_led;				// output pin to show progress
  bool				 dual_rings;				// whether we have 1 or 2 rings
  int				 neopixel_type;				// default color order and speed
  bool				 setup_called;				// if setup has been called.

  Adafruit_NeoPixel		 np;					// neopixels data structure
  Neopixel_ring			 pixels_single;				// single ring
  Neopixel_ring_mirror		 pixels_dual;				// two rings
  Neopixel_object		*pixels;				// switch between single ring or two rings
  Neopixel_random		 pixels_random;				// switch to random pixels
  Neopixel_chase		 pixels_chase;				// chase action
  Neopixel_color_rotate		 pixels_color;				// color rotate
  Neopixel_k_eye		 pixels_k_eye;				// kaleidoscope eyes
  Neopixel_roulette		 pixels_roulette;			// roulette
  Neopixel_solid_color		 pixels_solid_color;			// solid color
  Neopixel_theater_chase	 pixels_theater_chase;			// theater chase

 public:
  // constructor
  Neopixel_combined_ring (int				_num_pixels	= DEFAULT_NUM_PIXELS,		// number of pixels in a single ring
			  enum combined_ring_action	_button_action	= DEFAULT_BUTTON_ACTION,	// what does the button do
			  int				_pin_neopixel	= DEFAULT_PIN_NEOPIXEL,		// pin the neopixels are on
			  int				_pin_button	= DEFAULT_PIN_BUTTON,		// pin the button to change actions are on
			  int				_pin_led	= DEFAULT_PIN_LED,		// output pin to use a LED to show progress
			  bool				_dual_rings	= DEFAULT_DUAL_RINGS,		// whether we are doing 1 or 2 rings
			  int				_neopixel_type	= DEFAULT_NEOPIXEL_TYPE)	// default color order and speed

    :	num_pixels		(_num_pixels),
	button_action		(_button_action),
	pin_neopixel		(_pin_neopixel),
	pin_button		(_pin_button),
	pin_led			(_pin_led),
	dual_rings		(_dual_rings),
	neopixel_type		(_neopixel_type),
	setup_called		(false),

	np			(Adafruit_NeoPixel (((dual_rings) ? 2 : 1) * num_pixels,
						    pin_neopixel,
						    neopixel_type)),

	pixels_single		(Neopixel_ring (num_pixels,
						pin_neopixel,
						pin_led)),

	pixels_dual		(Neopixel_ring_mirror (num_pixels,
						       pin_neopixel,
						       pin_led)),

	pixels			(NULL),
	pixels_random		(Neopixel_random (pin_button)),
	pixels_chase		(Neopixel_chase ()),
	pixels_color		(Neopixel_color_rotate ()),
	pixels_k_eye		(Neopixel_k_eye ()),
	pixels_roulette		(Neopixel_roulette ()),
	pixels_solid_color	(Neopixel_solid_color ()),
	pixels_theater_chase	(Neopixel_theater_chase ())
  {
  }

  // destructor
  ~Neopixel_combined_ring ()
  {
  }

  // initial setup
  void setup (void);

  // loop iteration
  void loop (void);

  // Print the initial information to the USB terminal
  void print_info (void);

  // show pixels right now
  void show  (void)
  {
    if (setup_called && pixels != NULL)
      pixels->show ();
  }

  // return number of pixels
  int get_num_pixels (void)
  {
    return num_pixels;
  }
};

#endif	/* MEISSNER_NEOPIXEL_COMBINED_RING_H.  */


// HISTORY
// $Log: Meissner_Neopixel_Combined_Ring.h,v $
// Revision 1.7  2022/08/29 14:35:41  michaelmeissner
// Spacing.
//
// Revision 1.6  2022/03/22 05:14:37  michaelmeissner
// PIN_NEOPIXEL -> PIN_EXTERNAL_NEOPIXEL.
//
// Revision 1.5  2022/02/27 04:59:14  michaelmeissner
// Don't define DEFAULT_BRIGHTNESS here, define it in Meissner_Neopixel.h.
//
// Revision 1.4  2022/02/27 04:57:57  michaelmeissner
// Define DEFAULT_BRIGHTNESS.
//
// Revision 1.3  2022/02/13 06:30:16  michaelmeissner
// Iterate on fleshing out code #2.
//
// Revision 1.2  2022/02/13 04:31:33  michaelmeissner
// Iterate on fleshing out code.
//
// Revision 1.1  2022/02/13 02:59:04  michaelmeissner
// Initial file.
//

