// -*- arduino -*-
// Originally from https://learn.adafruit.com/kaleidoscope-eyes-neopixel-led-goggles-trinket-gemma/software
// Completely rewritten by Michael Meissner, using the algorithms in the Adafruit example, but not the code.

// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Kaleidoscope_Eyes.h>
#include <Meissner_Delay.h>


// constructor
Neopixel_k_eye::Neopixel_k_eye (Neopixel_power_t bright,	// brightness to use for the color
				Neopixel_iter_t	 min_iter,	// mininum number of iterations to do
				Neopixel_iter_t	 max_iter,	// maximum number of iterations to do
				unsigned long	 d_random,	// delay between setting random LEDs
				unsigned long	 d_wheel,	// delay between moving the wheel to the next position
				unsigned long	 d_switch)	// delay between switching modes

  :	Neopixel_action (min_iter, max_iter, bright, d_random, d_wheel),
	timeout_switch		(Delay ()),
	delay_switch		(d_switch),
	color			(0),
	random_mode		(true)
{
}

// destructor
Neopixel_k_eye::~Neopixel_k_eye (void)
{
}


// begin function to initialize everything (non-zero arguments, changes the default).  This is expected
// to be called once.

void
Neopixel_k_eye::begin (Neopixel_object *neo,		// Neopixel object to use
		       Neopixel_power_t bright,		// brightness to use for the color
		       Neopixel_iter_t	min_iter,	// mininum number of iterations to do
		       Neopixel_iter_t	max_iter,	// maximum number of iterations to do
		       unsigned long	d_random,	// delay between setting random LEDs
		       unsigned long	d_wheel,	// delay between moving the wheel to the next position
		       unsigned long	d_switch)	// delay between switching modes
{
  neo_obj = neo;

  if (bright)
    brightness = bright;

  if (min_iter)
    min_iterations = min_iter;

  if (max_iter)
    max_iterations = max_iter;

  if (d_random)
    delay_small = d_random;

  if (d_wheel)
    delay_large = d_wheel;

  if (d_switch)
    delay_switch = d_switch;

  random_mode = true;
  iterations  = random (min_iterations, max_iterations+1);
  color	      = ((Neopixel_rgb_t)brightness) << 16;

  // Set timeout to fire immediately
  timeout.run_immediate ();

  // Switch modes every 8 seconds or so
  timeout_switch.run_delay_repeat (delay_switch);

#ifdef K_EYE_DEBUG
#ifndef NEOPIXEL_DEBUG
  Serial.begin (9600);
  delay (1000);

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;
#endif

    Serial.println ("==========");
    Serial.println ("Kaleidoscope Eyes begin");
    Serial.print   ("min_iterations  = ");	Serial.println (min_iterations);
    Serial.print   ("max_iterations  = ");	Serial.println (max_iterations);
    Serial.print   ("iterations      = ");	Serial.println (iterations);
    Serial.print   ("delay random    = ");	Serial.println (delay_small);
    Serial.print   ("delay wheel     = ");	Serial.println (delay_large);
    Serial.print   ("delay switch    = ");	Serial.println (delay_switch);
    Serial.print   ("neo_obj         = 0x");	Serial.println ((size_t)neo_obj, HEX);
#endif

    if (neo)
      {
	num_pixels = neo_obj->max_pixels ();
	neo_obj->fill (0);
	neo_obj->show ();
      }
}


// Function to set up the actions, possibly setting iterations or brightness levels

void
Neopixel_k_eye::start (Neopixel_iter_t  iter,		// override # iterations
		       Neopixel_power_t bright)		// override brightness
{
  if (bright)
    brightness = bright;

  random_mode = true;
  iterations  = (iter) ? iter : random (min_iterations, max_iterations+1);
  color	      = ((Neopixel_rgb_t)brightness) << 16;

  // set timeout to fire immediately
  timeout.run_immediate ();

  // Switch modes every 8 seconds or so
  timeout_switch.run_delay (delay_switch);

#ifdef K_EYE_DEBUG
    Serial.println ("==========");
    Serial.println ("Kaleidoscope Eyes start");
    Serial.print   ("iterations      = ");	Serial.println (iterations);
    Serial.print   ("brightness      = ");	Serial.println (brightness);
#endif
}


// Function called from the main loop function to possibly do the next action.
// Return false, when all of the current actions are finished.  If do_show_p is
// not a NULL pointer, the function will not call show, but instead set
// *do_show_p to true.  It is assumed that the user has set *do_show_p to false
// before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool
Neopixel_k_eye::loop (bool *do_show_p)
{
  bool ret = true;

  if (neo_obj && neo_obj->busy ())
    return true;

  if (neo_obj && timeout.delay_done ())
    {
      neo_obj->fill (0);

      // Is it time to switch modes?
      if (timeout_switch.delay_done ())
	{
	  --iterations;
	  neo_obj->set_current (random (0, num_pixels));

	  if (random_mode)
	    {
	      // Switch to wheel mode
	      random_mode = false;
	    }
	  else
	    {
	      // Switch to random mode, and decrement the iterations
	      random_mode = true;
	      color >>= 8;
	      if (!color)
		{
		  color = ((Neopixel_rgb_t)brightness) << 16;

		  if (iterations == 0)
		    {
		      ret = false;
		      iterations = random (min_iterations, max_iterations+1);
		    }
		}
	    }

#ifdef K_EYE_DEBUG
	  Serial.println ("==========");
	  Serial.println ("Kaleidoscope Eyes switch mode");
	  Serial.print   ("iterations      = ");	Serial.println (iterations);
	  Serial.print   ("current pixel   = ");	Serial.println (neo_obj->current ());
	  Serial.print   ("color           = 0x");	Serial.println (color, HEX);
	  Serial.print   ("mode            = ");	Serial.println (random_mode ? "random" : "wheel");
	  Serial.print   ("ret             = ");	Serial.println (ret ? "true" : "false");
#endif
	}

      if (random_mode)
	{
	  neo_obj->set_pixel_color_32bit_mask (neo_obj->current (), color, NEOPIXEL_MASK_MIRROR_RANDOM);
	  neo_obj->set_current (random (0, num_pixels));
	  timeout.run_delay (delay_small);
	}
      else
	{
	  // implement rotating wheels.  Each wheel has 4 leds set, 2 on each side, and the two
	  // rings rotate in opposite directions.  Each iteration bumps the current pixel, so the
	  // wheels rotate
	  Neopixel_index_t left1  = neo_obj->current ();
	  Neopixel_index_t left2  = neo_obj->next (true);
	  Neopixel_index_t right1 = (left1 + (num_pixels/2)) % num_pixels;
	  Neopixel_index_t right2 = (right1 == num_pixels-1) ? 0 : right1+1;

	  neo_obj->set_pixel_color_32bit_mask (left1,  color, NEOPIXEL_MASK_MIRROR_REVERSE);
	  neo_obj->set_pixel_color_32bit_mask (left2,  color, NEOPIXEL_MASK_MIRROR_REVERSE);
	  neo_obj->set_pixel_color_32bit_mask (right1, color, NEOPIXEL_MASK_MIRROR_REVERSE);
	  neo_obj->set_pixel_color_32bit_mask (right2, color, NEOPIXEL_MASK_MIRROR_REVERSE);
	  timeout.run_delay (delay_large);
	}

      // show off the pixels
      if (do_show_p)
	*do_show_p = true;
      else
	neo_obj->show ();
    }

  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Kaleidoscope_Eyes.cpp,v $
// Revision 1.14  2022/08/28 21:44:45  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.13  2022/02/13 01:06:03  michaelmeissner
// Update copyright year.
//
// Revision 1.12  2020/01/28 04:17:28  michaelmeissner
// Update copyright year.
//
// Revision 1.11  2019/01/14 04:41:21  michaelmeissner
// Update copyright.
//
// Revision 1.10  2016/11/06 13:48:02  michaelmeissner
// Don't assume we have two rings, just set random mask.
//
// Revision 1.9  2016/11/05 12:37:35  michaelmeissner
// SetPixelColor -> set_pixel_color_32bit_mask
//
// Revision 1.8  2016/11/04 15:08:42  michaelmeissner
// Rename neopixel object -> neo_obj.
//
// Revision 1.7  2016/04/03 22:41:55  michaelmeissner
// Update copyright year
//
// Revision 1.6  2015/02/15 17:38:35  michaelmeissner
// Fix typo.
//
// Revision 1.5  2015/02/01 15:13:27  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.4  2015/01/12 02:42:25  michaelmeissner
// If we do not have mirrored rings, make random pixels only on the main ring/stick/etc.
//
// Revision 1.3  2015/01/11 14:36:07  michaelmeissner
// Change timeout functions.
//
// Revision 1.2  2015/01/10 22:51:38  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.1  2015/01/10 22:13:37  michaelmeissner
// Initial version.
//
