// -*- arduino -*-
// Originally from https://learn.adafruit.com/kaleidoscope-eyes-neopixel-led-goggles-trinket-gemma/software
// Completely rewritten by Michael Meissner, using the algorithms in the Adafruit example, but not the code.

// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_NEOPIXEL_KALEIDOSCOPE_EYES_H
#define MEISSNER_NEOPIXEL_KALEIDOSCOPE_EYES_H	1

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Delay.h>

// #define K_EYE_DEBUG	1					// define to enable debugging

const Neopixel_power_t	K_EYE_BRIGHTNESS		= 30;	// max brightness for kaleidoscope eyes
const Neopixel_iter_t	K_EYE_MIN_ITERATIONS		= 4;	// min # of iterations
const Neopixel_iter_t	K_EYE_MAX_ITERATIONS		= 16;	// max # of iterations
const unsigned long	K_EYE_DELAY_RANDOM		= 10;	// timeout between flashing a new pixel
const unsigned long	K_EYE_DELAY_WHEEL		= 50;	// timeout between the 8 element chase sequence
const unsigned long	K_EYE_DELAY_SWITCH		= 8000;	// timeout between switching modes

class Neopixel_k_eye : Neopixel_action {
 private:
  Delay		 timeout_switch;				// when to switch between random leds and wheels
  unsigned long	 delay_switch;					// timeout milliseconds to switch between modes
  Neopixel_rgb_t color;						// current color
  bool		 random_mode;					// whether we are doing random pixels or wheels

 public:
  // constructor/destructor
  Neopixel_k_eye (Neopixel_power_t	bright		= K_EYE_BRIGHTNESS,
		  Neopixel_iter_t	min_iter	= K_EYE_MIN_ITERATIONS,
		  Neopixel_iter_t	max_iter	= K_EYE_MAX_ITERATIONS,
		  unsigned long		d_random	= K_EYE_DELAY_RANDOM,
		  unsigned long		d_wheel		= K_EYE_DELAY_WHEEL,
		  unsigned long		d_switch	= K_EYE_DELAY_SWITCH);

  ~Neopixel_k_eye (void);

  // begin function to initialize everything (non-zero arguments, changes the default)
  void begin (Neopixel_object  *neo,
	      Neopixel_power_t	bright			= 0,
	      Neopixel_iter_t	min_iter		= 0,
	      Neopixel_iter_t	max_iter		= 0,
	      unsigned long	d_random		= 0,
	      unsigned long	d_wheel			= 0,
	      unsigned long	d_switch		= 0);

  // Function to set up the actions, possibly setting iterations or brightness levels
  void start (Neopixel_iter_t  iter			= 0,
	      Neopixel_power_t bright			= 0);

  // Function called from the main loop function to possibly do the next
  // action.  Return false, when all of the current actions are finished.  If
  // do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.  It is assumed that the user has set
  // *do_show_p to false before the call.
  bool loop (bool *do_show_p = NULL);

  // Return a string giving the name of the action for debugging
  const char *name (void)
  {
    return "Kaleidoscope_Eyes";
  }
};

#endif	/* MEISSNER_NEOPIXEL_KALEIDOSCOPE_EYES_H */


// HISTORY
// $Log: Meissner_Neopixel_Kaleidoscope_Eyes.h,v $
// Revision 1.10  2022/02/13 01:00:58  michaelmeissner
// Update copyright year.
//
// Revision 1.9  2020/01/28 04:07:11  michaelmeissner
// Update copyright year.
//
// Revision 1.8  2019/01/14 04:41:06  michaelmeissner
// Update copyright.
//
// Revision 1.7  2016/04/03 22:41:34  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.6  2016/04/03 22:40:58  michaelmeissner
// Update copyright year
//
// Revision 1.5  2015/02/01 15:12:32  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.4  2015/01/12 05:36:32  michaelmeissner
// Add virtual function to return the action name.
//
// Revision 1.3  2015/01/11 01:52:19  michaelmeissner
// Make the default # iterations smaller.
//
// Revision 1.2  2015/01/10 22:50:46  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.1  2015/01/10 22:14:26  michaelmeissner
// Initial version.
//

