// -*- c++ -*-
// Run random neopixel patterns for a time, before going on to the next.
//
// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Random.h>
#include <Meissner_Delay.h>
#include <Meissner_Random.h>
#include <Bounce2.h>

// constructor

Neopixel_random::Neopixel_random (Neopixel_pin_t  b,			// button to use to change modes
				  unsigned long   d,			// default delay
				  bool		  r,			// whether we are doing random patterns
				  Neopixel_iter_t min_iter,		// minimum iterations
				  Neopixel_iter_t max_iter,		// maximum iterations
				  bool		  print)		// print the pattern when it is changed to the Serial monitor

  :	Neopixel_action (min_iter, max_iter, 0, 0UL, d),

#if USE_BOUNCE2
	button_bounce	(Bounce ()),
#endif

	num_elements	(0),
	current		(0),
	enable_random	(RANDOM_MASK_ALL),
	enable_linear	(RANDOM_MASK_ALL),
	enable_first	(RANDOM_MASK_ALL),
	sub_iterations	(0),
	sub_bright	(0),
	button		(b),
	change_wakeup	((Delay *)0),

#if !USE_BOUNCE2
	prev_button	(-1),
#endif

	do_random	(r),
	do_first	(true),
	do_print	(print),
	initialized	(false),
	max_name_len	(0)
{
  for (size_t i = 0; i < RANDOM_NUM_ELEMENTS; i++)
    elements[i] = NULL;
}

// destructor

Neopixel_random::~Neopixel_random (void)
{
}


// register patterns

void
Neopixel_random::begin (Neopixel_object *neo,				// Neopixel object to use
			unsigned long    d,				// override delay
			Neopixel_pin_t   b,				// override button to change patterns immediately
			Random_mask_t	 mask_random,			// mask for random actions
			Random_mask_t	 mask_linear,			// mask for linear actions
			Random_mask_t	 mask_first,			// mask for actions to do on the first pass
			Neopixel_action *first,				// 1-8 patterns
			Neopixel_action *second,
			Neopixel_action *third,
			Neopixel_action *fourth,
			Neopixel_action *fifth,
			Neopixel_action *sixth,
			Neopixel_action *seventh,
			Neopixel_action *eighth)
{
  Random_mask_t mask_possible;

  neo_obj = neo;

  // override defaults set up in the constructor
  if (d)
    delay_large = d;

  if (b >= 0)
    button = b;

  if (button >= 0)
    {
      pinMode (button, INPUT_PULLUP);

#if USE_BOUNCE2
      button_bounce.attach (button);
      button_bounce.interval (RANDOM_DEBOUNCE);
#else
      prev_button = digitalRead (button);
#endif
    }

  num_elements = 0;
  if (first)
    elements[num_elements++] = first;

  if (second)
    elements[num_elements++] = second;

  if (third)
    elements[num_elements++] = third;

  if (fourth)
    elements[num_elements++] = fourth;

  if (fifth)
    elements[num_elements++] = fifth;

  if (sixth)
    elements[num_elements++] = sixth;

  if (seventh)
    elements[num_elements++] = seventh;

  if (eighth)
    elements[num_elements++] = eighth;

  // select the possible actions to do in random and linear modes
  mask_possible = (((Random_mask_t)1) << num_elements) - 1;
  enable_random = mask_random & mask_possible;
  enable_linear = mask_linear & mask_possible;
  enable_first  = mask_first  & mask_possible;
  if ((enable_random == 0) && (enable_linear == 0))
    enable_random = enable_linear = mask_possible;

  // setup the first action
  do_first = (enable_first != 0);
  select_next (NULL, false, true);

  timeout.run_delay (delay_large);

  if (use_serial_output ())
    {
      Serial.println ("==========");
      Serial.println ("Random begin");
      Serial.print   ("min_iterations  = ");	Serial.println (min_iterations);
      Serial.print   ("max_iterations  = ");	Serial.println (max_iterations);
      Serial.print   ("iterations      = ");	Serial.println (iterations);
      Serial.print   ("sub_iterations  = ");	Serial.println (sub_iterations);
      Serial.print   ("delay switch    = ");	Serial.println (delay_large);
      Serial.print   ("num_elements    = ");	Serial.println (num_elements);
      Serial.print   ("current         = ");	Serial.println (current);
      Serial.print   ("button          = ");	Serial.println (button);

#if USE_BOUNCE2
      Serial.print   ("debounce        = ");	Serial.println (RANDOM_DEBOUNCE);
      Serial.println ("use bounce2     = true");
#else
      Serial.println ("use bounce2     = false");
#endif

      Serial.print   ("do_random       = ");	Serial.println (do_random ? "true" : "false");
      Serial.print   ("enable random   = 0x");	Serial.println (enable_random, HEX);
      Serial.print   ("enable linear   = 0x");	Serial.println (enable_linear, HEX);
      Serial.print   ("neo_obj         = 0x");	Serial.println ((size_t)neo_obj, HEX);

      for (size_t i = 0; i < num_elements; i++)
	{
	  Serial.print   ("elements[");
	  Serial.print   (i);
	  Serial.print   ("]     = 0x");
	  Serial.print   ((size_t)elements[i], HEX);
	  Serial.print   (", name = ");

	  const char *name = elements[i]->name ();
	  size_t len = strlen (name);
	  if (max_name_len < len)
	    max_name_len = len;

	  Serial.println (name);
	}

      Serial.print   ("max_name_len    = ");	Serial.println (max_name_len);
    }

    if (neo)
      {
	num_pixels = neo_obj->max_pixels ();
	neo_obj->fill (0);
	neo_obj->show ();
      }

    initialized = true;
    return;
}


// Function to set up the actions, possibly setting iterations or brightness levels

void
Neopixel_random::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  // remember arguments to pass to sub-actions, when we start them
  sub_iterations = iter;
  sub_bright	 = bright;

  // setup the timouts
  timeout.run_delay (delay_large);

  // setup the first action
  select_next (NULL, false, true);

#if RANDOM_DEUBG && USE_SERIAL_OUTPUT
  if (initialized && use_serial_output ())
    {
      Serial.println ("==========");
      Serial.println ("Random start");
      Serial.print   ("current         = ");	Serial.println (current);
      Serial.print   ("action          = ");	Serial.println (elements[current]->name ());
      Serial.print   ("iterations      = ");	Serial.println (iterations);
      Serial.print   ("sub_iterations  = ");	Serial.println (sub_iterations);
      Serial.print   ("sub_bright      = ");	Serial.println (sub_bright);
      Serial.print   ("do_random       = ");	Serial.println (do_random ? "true" : "false");
    }
#endif

  elements[current]->start (iterations, sub_bright);
}


// Select the next pattern to run

void
Neopixel_random::select_next (bool *do_show_p, bool button_changed, bool init_p)
{
  Random_mask_t enable_current;

#if USE_SERIAL_OUTPUT
  const char *action;

#define ACTION(STR) action = STR;
#else
#define ACTION(STR)
#endif

  iterations = (sub_iterations) ? sub_iterations : random (min_iterations, max_iterations+1);

  // Pick the next action to run.  If a button was pressed, switch from random
  // mode to linear mode, and iterate over the actions in a linear fashion.
  // This allows having one pattern that is selected by pressing a button, but
  // that pattern might be skipped in random mode (useful if you are wearing
  // steampunk goggles and want to go to a sedate pattern when talking to
  // somebody)

  if (do_first && enable_first == 0)
    do_first = false;

  if (do_random && enable_random == 0)
    do_random = false;

  if (button_changed)
    {
      do_first = false;
      if ((enable_random != 0) && (enable_linear != 0))
	{
	  if (do_random)
	    {
	      do_random = false;
	      init_p = true;
	    }
	  else
	    {
	      do_random = true;
	      init_p = false;
	    }
	}
    }

  do
    {
      ACTION ("Unknown");
      if (do_random && !do_first)
	{
	  current = random_unique (0, num_elements, current, RANDOM_MAX_RETRIES);
	  ACTION ("Random,");
	  enable_current = enable_random;
	}

      else if (init_p)
	{
	  current = 0;
	  init_p  = false;
	  if (do_first)
	    {
	      enable_current = enable_first;
	      ACTION ("First, ");
	    }
	  else
	    {
	      enable_current = enable_linear;
	      ACTION ("Linear,");
	    }
	}

      else if (current >= num_elements-1)
	{
	  if (do_random || (do_first && (enable_random != 0)))
	    {
	      current = random_unique (0, num_elements, current, RANDOM_MAX_RETRIES);
	      ACTION ("Random,");
	      enable_current = enable_random;
	    }
	  else
	    {
	      current = 0;
	      ACTION ("Linear,");
	      enable_current = enable_linear;
	    }
	  do_first = false;
	}

      else if (do_first)
	{
	  current++;
	  ACTION ("First, ");
	  enable_current = enable_first;
	  enable_first &= ~(((Random_mask_t)1) << current);
	}

      else
	{
	  current++;
	  ACTION ("Linear,");
	  enable_current = enable_linear;
	}
    }
  while (((((Random_mask_t)1) << current) & enable_current) == 0);

  // tell the action to start
  elements[current]->start (iterations, sub_bright);

  // reset the timeout
  timeout.run_delay (delay_large);

  // clear the neopixels
  if (neo_obj)
    {
      neo_obj->fill (0);
      if (do_show_p)
	*do_show_p = true;
      else
	neo_obj->show ();
    }

  /* If the user wanted a delay to be cancelled when we changed patterns, do it now.  */
  if (change_wakeup && initialized)
    change_wakeup->run_immediate (true);

#if RANDOM_DEBUG && USE_SERIAL_OUTPUT
  if (initialized && use_serial_output ())
    {
      const char	*pattern	= current_pattern ();
      char		*name_print	= (char *) alloca (max_name_len + 5);
      char		*p		= name_print;
      size_t		 len		= strlen (pattern);

      *p++ = '"';
      strcpy (p, pattern);
      p += len;
      *p++ = '"';
      *p++ = ',';
      for (size_t i = max_name_len - len; i > 0; i--)
	*p++ = ' ';

      *p = '\0';

#if USE_SERIAL_PRINTF
      Serial.printf ("Random select_next, %s pattern = %s # = %d, iterations = %4d, sub_iterations = %4d, button = %s random = %s\n",
		     action,
		     name_print,
		     current,
		     iterations,
		     sub_iterations,
		     button_changed	? "true, " : "false,",
		     do_random		? "true"   : "false");
#else
      Serial.print ("Random select_next, ");
      Serial.print (action);
      Serial.print (" pattern = ");
      Serial.print (name_print);
      Serial.print (" # = ");
      Serial.print (current);
      Serial.print (", iterations = ");
      Serial.print (iterations);
      Serial.print (", sub_iterations = ");
      Serial.print (sub_iterations);
      Serial.print (", button = ");
      Serial.print (button_changed ? "true, " : "false,");
      Serial.print (" random = ");
      Serial.println (do_random ? "true" : "false");
#endif
    }
#endif
}


// Function called from the main loop function to possibly do the next action.
// Return false, when all of the current actions are finished.
// If do_show_p is not a NULL pointer, the function will not call show, but
// instead set *do_show_p to true.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool
Neopixel_random::loop (bool *do_show_p)
{
  bool ret = true;
  bool modified;

  if (do_show_p)
    *do_show_p = false;

  if (neo_obj)
    {
      if (neo_obj->busy ())
	return true;

      bool button_changed;

      if (button < 0)
	button_changed = false;
      else
	{
#if USE_BOUNCE2
	  button_bounce.update ();
	  button_changed = button_bounce.fell ();
#else
	  int cur_button = digitalRead (button);
	  button_changed = (cur_button == LOW && prev_button == HIGH);
	  prev_button = cur_button;
#endif
	}

#if RANDOM_DEBUG && USE_SERIAL_OUTPUT
      if (do_print && button_changed && initialized && use_serial_output ())
	{
	  Serial.print ("Random: button ");
	  Serial.print (button);
	  Serial.println (" pushed");
	}
#endif

      // is it time to switch sub-actions either via timeout or by the button
      // being pushed?
      if (timeout.delay_done () || button_changed)
	{
	  if (--iterations == 0)
	    ret = false;

	  select_next (do_show_p, button_changed, false);
	}

      // Run the current action.  If the action indicates it is done,
      // select a new action.  Just in case, make sure *do_show_p
      // status is preserved.
      modified = (do_show_p && *do_show_p);
      if (!elements[current]->loop (do_show_p))
	{
	  modified |= (do_show_p && *do_show_p);
	  select_next (do_show_p, false, false);
	  if (do_show_p)
	    *do_show_p |= modified;
	}
    }

  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Random.cpp,v $
// Revision 1.45  2022/08/28 21:41:23  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.44  2022/02/28 05:00:17  michaelmeissner
// Align iterations debug output.
//
// Revision 1.43  2022/02/28 02:14:24  michaelmeissner
// Line up debug output.
//
// Revision 1.42  2022/02/28 02:02:01  michaelmeissner
// Align name in debug printouts.
//
// Revision 1.41  2022/02/28 01:34:05  michaelmeissner
// Reorganize debug output.
//
// Revision 1.40  2022/02/28 01:14:25  michaelmeissner
// Switch to using printf #2.
//
// Revision 1.39  2022/02/28 01:04:15  michaelmeissner
// Switch to using printf.
//
// Revision 1.38  2022/02/13 01:06:47  michaelmeissner
// Update copyright year.
//
// Revision 1.37  2020/01/28 04:17:49  michaelmeissner
// Update copyright year.
//
// Revision 1.36  2019/01/14 04:41:56  michaelmeissner
// Update copyright.
//
// Revision 1.35  2018/12/28 00:30:03  michaelmeissner
// Use USE_SERIAL_OUTPUT instead of !NO_SERIAL
//
// Revision 1.34  2018/03/18 21:30:50  michaelmeissner
// Add ability to reset a wake up delay if we changed patterns.
//
// Revision 1.33  2017/10/06 05:52:33  michaelmeissner
// Update copyright year.
//
// Revision 1.32  2016/11/05 15:46:48  michaelmeissner
// Make whether to use bounce2 configurable via #if.
//
// Revision 1.31  2016/11/04 15:09:18  michaelmeissner
// Rename neopixel object -> neo_obj.
//
// Revision 1.30  2016/11/02 04:43:00  michaelmeissner
// Disable using Bounce2 for now.
//
// Revision 1.29  2016/10/31 14:04:30  michaelmeissner
// Explicitly include Bounce2.h; Set initialized variable true after begin runs; Don't print stuff until initialized.
//
// Revision 1.28  2016/10/31 04:12:33  michaelmeissner
// Print debounce time.
//
// Revision 1.27  2016/09/28 05:19:52  michaelmeissner
// If both linear and random, make button press switch between linear and random modes.
//
// Revision 1.26  2016/09/28 05:01:27  michaelmeissner
// Print whether actions are first, linear or random; Rework code setting actions somewhat.
//
// Revision 1.25  2016/09/28 03:26:50  michaelmeissner
// Switch to use Bounce2 library.
//
// Revision 1.24  2016/09/27 13:58:04  michaelmeissner
// Disable bounce support for now
//
// Revision 1.23  2016/08/29 12:25:22  michaelmeissner
// Check for neopixels before printing trace information.
//
// Revision 1.22  2016/08/11 04:02:51  michaelmeissner
// Reset iterations each time we select a new pattern (unless we were passed a specific iteration count)
//
// Revision 1.21  2016/08/10 15:09:01  michaelmeissner
// Print iterations when printing choice.
//
// Revision 1.20  2016/08/10 14:55:49  michaelmeissner
// Include Meissner_Random.h
//
// Revision 1.19  2016/08/10 14:54:59  michaelmeissner
// Don't repeat the current pattern when choosing a new random pattern.
//
// Revision 1.18  2016/08/09 14:27:27  michaelmeissner
// Add option to print pattern name as it is changed and default it to on.
//
// Revision 1.17  2016/04/03 22:47:46  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.16  2016/03/21 14:03:50  michaelmeissner
// Update copyright year.
//
// Revision 1.15  2015/02/15 17:34:14  michaelmeissner
// Reorder constructor to avoid -Wreorder warnings.
//
// Revision 1.14  2015/02/02 19:36:37  michaelmeissner
// Add support for running patterns sequentially on first pass.
//
// Revision 1.13  2015/01/12 14:37:13  michaelmeissner
// Make sure any change to *do_show_p is propigated to the caller.
//
// Revision 1.12  2015/01/12 14:30:46  michaelmeissner
// Fix off-by-1 error
//
// Revision 1.11  2015/01/12 14:26:30  michaelmeissner
// Fix spacing.
//
// Revision 1.10  2015/01/12 14:24:45  michaelmeissner
// Reset timer when selecting the next pattern.
//
// Revision 1.9  2015/01/12 14:22:50  michaelmeissner
// Update a comment #2.
//
// Revision 1.8  2015/01/12 14:21:24  michaelmeissner
// Update a comment.
//
// Revision 1.7  2015/01/12 14:20:51  michaelmeissner
// Add a comment.
//
// Revision 1.6  2015/01/12 14:12:55  michaelmeissner
// Add options to have different actions enabled in random mode or linear mode, and pressing the button switches to linear mode.
//
// Revision 1.5  2015/01/12 13:24:12  michaelmeissner
// Add button debouncing.
//
// Revision 1.4  2015/01/12 05:47:30  michaelmeissner
// Read button state in start function; Add debugging to start.
//
// Revision 1.3  2015/01/12 05:40:01  michaelmeissner
// If debugging, print the sub-action name.
//
// Revision 1.2  2015/01/12 05:29:37  michaelmeissner
// Don't use Bounce library for the moment.
//
// Revision 1.1  2015/01/12 05:15:20  michaelmeissner
// Initial version.
//
