// -*- c++ -*-
// Run random neopixel patterns for a time, before going on to the next.
//
// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_NEOPIXEL_RANDOM_H
#define MEISSNER_NEOPIXEL_RANDOM_H	1

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Delay.h>
#include <Bounce2.h>

// #define USE_BOUNCE2 1

// Enable tracing to the Serial device
#if USE_SERIAL_OUTPUT && !defined(RANDOM_DEBUG)
#define RANDOM_DEBUG	1
#endif

typedef uint16_t Random_mask_t;						// mask type for enablement actions

const Neopixel_iter_t	RANDOM_MIN_ITERATIONS	= 16;			// minimum iterations
const Neopixel_iter_t	RANDOM_MAX_ITERATIONS	= 1024;			// maximum iterations
const unsigned long	RANDOM_DELAY_SWITCH	= (120UL * 1000UL);	// default time (2 minutes) between mode switches
const unsigned long	RANDOM_DEBOUNCE		= 500;			// debounce button for 500 milliseconds
const size_t		RANDOM_NUM_ELEMENTS	= 8;			// max # of actions to iterate between
const bool		RANDOM_RANDOM		= true;			// whether to do sub-actions randomly
const Random_mask_t	RANDOM_MASK_ALL		= 0xff;			// mask to note which of 8 actions are enabled

class Neopixel_random : Neopixel_action {
 private:
  Neopixel_action	*elements[RANDOM_NUM_ELEMENTS];			// actions to implement neopixel patterns

#if USE_BOUNCE2
  Bounce		 button_bounce;					// debounce the button
#endif

  size_t		 num_elements;					// number of elements active
  size_t		 current;					// current action
  Random_mask_t		 enable_random;					// mask for actions to do in random mode
  Random_mask_t		 enable_linear;					// mask for actions to do in linear mode
  Random_mask_t		 enable_first;					// mask for actions to do on the first pass
  Neopixel_iter_t	 sub_iterations;				// iterations to pass when starting sub-actions
  Neopixel_power_t	 sub_bright;					// brightness to pass when starting sub-actions
  Neopixel_pin_t	 button;					// record pin used for the button
  Delay			*change_wakeup;					// enable timeout of timer if we changed the pattern

#if !USE_BOUNCE2
  uint8_t		 prev_button;					// state of the previous button
#endif

  bool			 do_random;					// whether we are processing actions in random order
  bool			 do_first;					// do first patterns
  bool			 do_print;					// print the pattern when it is changed to the Serial monitor
  bool			 initialized;					// whether all initialization has been done
  size_t		 max_name_len;					// length of the maximum name

  // select next pattern to run
  void select_next (bool *, bool, bool);

public:
  // constructor/destructor
  Neopixel_random (Neopixel_pin_t  b	    = -1,			// button to use to change modes (default no button)
		   unsigned long   delay    = RANDOM_DELAY_SWITCH,	// default delay
		   bool		   r	    = RANDOM_RANDOM,		// whether we start out doing random patterns
		   Neopixel_iter_t min_iter = RANDOM_MIN_ITERATIONS,	// minimum # of iterations
		   Neopixel_iter_t max_iter = RANDOM_MAX_ITERATIONS,	// maximum # of iterations
		   bool		   print    = true);			// print the pattern when it is changed to the Serial monitor

  ~Neopixel_random (void);

  // register patterns
  void begin (Neopixel_object *neo,					// Neopixel object to use
	      unsigned long    d,					// override delay
	      Neopixel_pin_t   b,					// override button to change patterns immediately
	      Random_mask_t    mask_random,				// actions to do in random mode
	      Random_mask_t    mask_linear,				// actions to do in linear mode
	      Random_mask_t    mask_first,				// actions to do on the first pass
	      Neopixel_action *first,					// 1-8 patterns
	      Neopixel_action *second	    = NULL,
	      Neopixel_action *third	    = NULL,
	      Neopixel_action *fourth	    = NULL,
	      Neopixel_action *fifth	    = NULL,
	      Neopixel_action *sixth	    = NULL,
	      Neopixel_action *seventh	    = NULL,
	      Neopixel_action *eighth	    = NULL);

  // Function to set up the actions, possibly setting iterations or brightness levels
  void start (Neopixel_iter_t  iter	    = 0,
	      Neopixel_power_t bright	    = 0);

  // Function called from the main loop function to possibly do the next
  // action.  Return false, when all of the current actions are finished.  If
  // do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.  It is assumed that the user has set
  // *do_show_p to false before the call.
  bool loop (bool *do_show_p		    = NULL);

  // Return a string giving the name of the action for debugging
  const char *name (void)
  {
    return "Random";
  }

  // Return a string giving the name of the current pattern
  const char *current_pattern (void)
  {
    return (elements[current]->name ());
  }

  // Set the flag for printing the patterns on the serial monitor
  void set_print (bool print = true)
  {
    do_print = print;
#if !defined(NO_SERIAL)
    if (do_print)
      {
	Serial.println ("Start printing random pattern name");
      }
#endif
  }

  // Set a timer to fire immediately if we change the pattern
  void wake_on_change (Delay *d)
  {
    change_wakeup = d;
  }
};

#endif	/* MEISSNER_NEOPIXEL_RANDOM_H */

// HISTORY
// $Log: Meissner_Neopixel_Random.h,v $
// Revision 1.28  2022/02/28 01:38:59  michaelmeissner
// Add max_name_len.
//
// Revision 1.27  2022/02/13 01:01:39  michaelmeissner
// Update copyright year.
//
// Revision 1.26  2020/01/28 04:08:42  michaelmeissner
// Update copyright year.
//
// Revision 1.25  2019/01/14 04:41:42  michaelmeissner
// Update copyright.
//
// Revision 1.24  2018/12/28 00:26:51  michaelmeissner
// Turn on random tracing by default.
//
// Revision 1.23  2018/03/18 21:28:07  michaelmeissner
// Add wake_on_change function to signal when the pattern changed.
//
// Revision 1.22  2017/10/06 05:52:17  michaelmeissner
// Update copyright year.
//
// Revision 1.21  2016/11/05 15:46:35  michaelmeissner
// Make whether to use bounce2 configurable via #if.
//
// Revision 1.20  2016/11/02 04:43:15  michaelmeissner
// Disable using Bounce2 for now.
//
// Revision 1.19  2016/10/31 14:03:07  michaelmeissner
// Add initialized boolean.
//
// Revision 1.18  2016/10/31 13:44:20  michaelmeissner
// Bump up debounce time to 500 ms
//
// Revision 1.17  2016/09/28 03:27:13  michaelmeissner
// Switch to use Bounce2 library.
//
// Revision 1.16  2016/09/27 13:57:47  michaelmeissner
// Disable bounce support for now
//
// Revision 1.15  2016/08/09 14:27:50  michaelmeissner
// Add option to print pattern name as it is changed and default it to on.
//
// Revision 1.14  2016/04/03 22:48:05  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.13  2016/03/21 14:03:33  michaelmeissner
// Update copyright year.
//
// Revision 1.12  2015/02/02 19:16:04  michaelmeissner
// Add do_first field.
//
// Revision 1.11  2015/02/02 19:10:58  michaelmeissner
// Add argument to do actions sequentially on the first pass.
//
// Revision 1.10  2015/02/01 15:14:31  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.9  2015/01/12 14:14:21  michaelmeissner
// Reorder fields slightly.
//
// Revision 1.8  2015/01/12 14:13:19  michaelmeissner
// Add options to have different actions enabled in random mode or linear mode, and pressing the button switches to linear mode.
//
// Revision 1.7  2015/01/12 13:23:51  michaelmeissner
// Add button debouncing.
//
// Revision 1.6  2015/01/12 06:57:52  michaelmeissner
// Turn off debugging
//
// Revision 1.5  2015/01/12 05:36:55  michaelmeissner
// Add virtual function to return the action name.
//
// Revision 1.4  2015/01/12 05:31:25  michaelmeissner
// Remove Bounce.h for the moment.
//
// Revision 1.3  2015/01/12 05:24:02  michaelmeissner
// Default constructor button to PIN_BUTTON.
//
// Revision 1.2  2015/01/12 05:22:56  michaelmeissner
// Don't use Bounce library for the moment.
//
// Revision 1.1  2015/01/12 05:15:34  michaelmeissner
// Initial version.
//
