// -*- arduino -*-
// Meissner neopixel shift between red and green (Christmas ornament)
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Red_Green.h>
#include <Meissner_Delay.h>


// Constructor, destructor
Neopixel_red_green::Neopixel_red_green (Neopixel_power_t bright,
					Neopixel_power_t step,
					Neopixel_iter_t  max_iter,
					Neopixel_iter_t  min_iter,
					unsigned long	 delay_inner,
					unsigned long	 delay_whole)

      :	Neopixel_action (max_iter, min_iter, bright, delay_inner, delay_whole),
	red	     (bright),
	green	     (0),
	bright_step  (step),
	do_red	     (true)
{
}

Neopixel_red_green::~Neopixel_red_green (void)
{
}


// begin function to initialize everything.
// This is expected to be called only once.

void
Neopixel_red_green::begin (Neopixel_object  *neo,
			   Neopixel_power_t  bright,
			   Neopixel_power_t  step,
			   Neopixel_iter_t   min_iter,
			   Neopixel_iter_t   max_iter,
			   unsigned long     delay_fade,
			   unsigned long     delay_whole)
{
  neo_obj = neo;

  if (bright)
    brightness = bright;

  if (step)
    bright_step = step;

  if (min_iter)
    min_iterations = min_iter;

  if (max_iter)
    max_iterations = max_iter;

  if (delay_fade)
    delay_small = delay_fade;

  if (delay_whole)
    delay_large = delay_whole;

#ifdef RED_GREEN_DEBUG
#ifndef NEOPIXEL_DEBUG
  Serial.begin (9600);
  delay (1000);

  // wait up to 3 seconds for the Serial device to become available
  long unsigned debug_start = millis ();
  while (!Serial && ((millis () - debug_start) <= 3000))
    ;
#endif

  Serial.println ("==========");
  Serial.println ("Red/green begin");
  Serial.print   ("min_iterations = ");	Serial.println (min_iterations);
  Serial.print   ("max_iterations = ");	Serial.println (max_iterations);
  Serial.print   ("brightness     = ");	Serial.println (brightness);
  Serial.print   ("bright step    = ");	Serial.println (bright_step);
  Serial.print   ("delay fade     = ");	Serial.println (delay_small);
  Serial.print   ("delay whole    = ");	Serial.println (delay_large);
#endif

  if (neo)
    {
      neo_obj->fill (0);
      neo_obj->show (false);
    }
}


// Setup for red/green shift.  This is expected to be called each time we switch to doing the shift
// sequence to reset the information, and start the timer.

void
Neopixel_red_green::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  red	     = brightness;
  green	     = 0;
  do_red     = true;
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);

  if (bright)
    brightness = bright;

  // start the timer to go off imediately on the first call to loop
  timeout.run_immediate ();

#ifdef RED_GREEN_DEBUG
    Serial.println ("==========");
    Serial.println ("Red/green start");
    Serial.print   ("iterations     = ");	Serial.println (iterations);
    Serial.print   ("brightness     = ");	Serial.println (brightness);
#endif
}


// loop function to be called from the main loop, return false when all
// iterations are done If do_show_p is not a NULL pointer, the function will
// not call show, but instead set *do_show_p to true.  It is assumed that the
// user has set *do_show_p to false before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool
Neopixel_red_green::loop (bool *do_show_p)
{
  bool ret = true;

  if (neo_obj && neo_obj->busy ())
    return true;

  if (timeout.delay_done ())
    {
      // set up next timeout
      timeout.run_delay (delay_small);

      neo_obj->fill_rgb ((Neopixel_power_t)red,
			 (Neopixel_power_t)green,
			 (Neopixel_power_t)0);

      if (do_show_p)
	*do_show_p = true;
      else
	neo_obj->show ();

      // Go to the next color
      if (do_red)
	{
	  red   += bright_step;
	  green -= bright_step;

	  if (red > brightness || green < 0)
	    {
	      // If we are at max brightness, delay for a bit longer, and consider this the end of an iteration
	      timeout.run_delay (delay_large);

	      if (--iterations == 0)
		{
		  ret	     = false;
		  red	     = brightness;
		  green	     = 0;
		  do_red     = true;
		  iterations = random (min_iterations, max_iterations+1);
		}

#ifdef RED_GREEN_DEBUG
	      Serial.print ("Red/green, iterations = ");
	      Serial.print (iterations);
	      Serial.println (", fade green to red");
#endif

	      red    = brightness - bright_step;
	      green  = bright_step;
	      do_red = false;
	    }
	}
      else
	{
	  red   -= bright_step;
	  green += bright_step;

	  if (green > brightness || red < 0)
	    {
	      // If we are at max brightness, delay for a bit longer
	      timeout.run_delay (delay_large);

#ifdef RED_GREEN_DEBUG
	      Serial.print ("Red/green, iterations = ");
	      Serial.print (iterations);
	      Serial.println (", fade red to green");
#endif

	      red    = bright_step;
	      green  = brightness - bright_step;
	      do_red = true;
	    }
	}
    }

  // If we are returning end of sequence, setup for the next run, just in case
  // start is not called.

  if (!ret)
    {
      red	 = brightness;
      green	 = 0;
      do_red     = true;
      iterations = random (min_iterations, max_iterations+1);

#ifdef RED_GREEN_DEBUG
      Serial.println ("==========");
      Serial.println ("Red/green end of sequence");
#endif
    }

  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Red_Green.cpp,v $
// Revision 1.20  2022/08/28 21:48:57  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.19  2022/02/13 01:07:03  michaelmeissner
// Update copyright year.
//
// Revision 1.18  2020/01/28 04:17:56  michaelmeissner
// Update copyright year.
//
// Revision 1.17  2019/01/14 04:42:27  michaelmeissner
// Update copyright.
//
// Revision 1.16  2016/11/05 12:45:33  michaelmeissner
// neopixel -> neo_obj #2
//
// Revision 1.15  2016/11/05 12:39:28  michaelmeissner
// Rename neopixel -> neo_obj
//
// Revision 1.14  2016/04/03 22:48:48  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.13  2016/03/21 14:04:32  michaelmeissner
// Update copyright year.
//
// Revision 1.12  2015/02/15 17:35:05  michaelmeissner
// Reorder constructor to avoid -Wreorder warnings.
//
// Revision 1.11  2015/02/01 15:17:30  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.10  2015/01/11 14:41:22  michaelmeissner
// Change timeout functions; Redo sequence ending.
//
// Revision 1.9  2015/01/10 22:50:05  michaelmeissner
// Add comment.
//
// Revision 1.8  2015/01/10 22:43:37  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.7  2015/01/10 03:20:12  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.6  2015/01/09 05:43:58  michaelmeissner
// correct off by one errors in random calls.
//
// Revision 1.5  2015/01/09 04:39:50  michaelmeissner
// Remove show/loop parameter do_blink.
//
// Revision 1.4  2015/01/08 13:50:00  michaelmeissner
// Fix indentation.
//
// Revision 1.3  2015/01/05 04:16:24  michaelmeissner
// Set up for next sequence and end of sequence.
//
// Revision 1.2  2015/01/05 04:13:53  michaelmeissner
// Don't call start in begin; Avoid overloading 'start'.
//
// Revision 1.1  2015/01/05 04:12:36  michaelmeissner
// Initial version.
//
