// -*- arduino -*-
// Meissner neopixel shift between red and green (Christmas ornament)
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#ifndef MEISSNER_NEOPIXEL_RED_GREEN_H
#define MEISSNER_NEOPIXEL_RED_GREEN_H	1

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Delay.h>

// #define RED_GREEN_DEBUG	1				// define to enable debugging

const unsigned long	RED_GREEN_DELAY_FADE	 = 125;		// timeout between fading
const unsigned long	RED_GREEN_DELAY_WHOLE	 = 3000;	// timeout after acheiving a single color
const Neopixel_power_t	RED_GREEN_BRIGHTNESS	 = 30;		// max brightness for red/green
const Neopixel_power_t	RED_GREEN_BRIGHT_STEP	 = 1;		// step between fades
const Neopixel_iter_t	RED_GREEN_MIN_ITERATIONS = 4;		// min # of iterations
const Neopixel_iter_t	RED_GREEN_MAX_ITERATIONS = 16;		// max # of iterations

class Neopixel_red_green : Neopixel_action {
 private:
  // Normally we would use Neopixel_power_t for red/green, but we want a larger type
  // that is signed to make some tests simpler.
  int			red;					// current red brightness
  int			green;					// current green brightness
  Neopixel_power_t	bright_step;				// step in changing the brightness
  bool			do_red;					// whether we are increasing red or green

 public:
  // constructor/destructor
  Neopixel_red_green (Neopixel_power_t	bright		= RED_GREEN_BRIGHTNESS,
		      Neopixel_power_t	step		= RED_GREEN_BRIGHT_STEP,
		      Neopixel_iter_t	min_iter	= RED_GREEN_MIN_ITERATIONS,
		      Neopixel_iter_t	max_iter	= RED_GREEN_MAX_ITERATIONS,
		      unsigned long	delay_fade	= RED_GREEN_DELAY_FADE,
		      unsigned long	delay_whole	= RED_GREEN_DELAY_WHOLE);

  ~Neopixel_red_green (void);

  // begin function to initialize everything (non-zero changes the default from the constructor)
  void begin (Neopixel_object  *neo,
	      Neopixel_power_t	bright			= 0,
	      Neopixel_power_t	step			= 0,
	      Neopixel_iter_t	min_iter		= 0,
	      Neopixel_iter_t	max_iter		= 0,
	      unsigned long	delay_fade		= 0,
	      unsigned long	delay_whole		= 0);

  // Function to set up the actions, possibly setting iterations or brightness levels
  void start (Neopixel_iter_t  iter   = 0,
	      Neopixel_power_t bright = 0);

  // Function called from the main loop function to possibly do the next
  // action.  Return false, when all of the current actions are finished.  If
  // do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.  It is assumed that the user has set
  // *do_show_p to false before the call.
  bool loop (bool *do_show_p = NULL);

  // Return a string giving the name of the action for debugging
  const char *name (void)
  {
    return "Red_Green";
  }
};

#endif	/* MEISSNER_NEOPIXEL_RED_GREEN_H */


// HISTORY
// $Log: Meissner_Neopixel_Red_Green.h,v $
// Revision 1.14  2022/02/13 01:01:56  michaelmeissner
// Update copyright year.
//
// Revision 1.13  2020/01/28 04:08:50  michaelmeissner
// Update copyright year.
//
// Revision 1.12  2019/01/14 04:42:13  michaelmeissner
// Update copyright.
//
// Revision 1.11  2016/04/03 22:48:34  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.10  2016/03/21 14:04:15  michaelmeissner
// Update copyright year.
//
// Revision 1.9  2015/02/01 15:16:34  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.8  2015/01/12 05:35:48  michaelmeissner
// Add virtual function to return the action name.
//
// Revision 1.7  2015/01/11 14:42:01  michaelmeissner
// Delete sequence_end, which is now unused.
//
// Revision 1.6  2015/01/10 22:42:03  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.5  2015/01/10 03:19:38  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.4  2015/01/09 04:39:18  michaelmeissner
// Remove show/loop parameter do_blink.
//
// Revision 1.3  2015/01/08 13:51:14  michaelmeissner
// Set defaults for begin to 0 to not change the constructor.
//
// Revision 1.2  2015/01/05 04:19:52  michaelmeissner
// Turn off debugging.
//
// Revision 1.1  2015/01/05 04:12:17  michaelmeissner
// Initial version.
//
