// -*- arduino -*-
// Initial code from: http://damage-designs.com/
// but heavily modified by Michael Meissner (arduino@the-meissners.org)
//
// Quote from original author:
// Neopixel Ring Roulette wheel sketch.
// "Spins" a roulette wheel and lands on a random value.
// I am not 100% satisfied with the aesthetic, but its as close as I can get.
// I would prefer a more realistic spinning to a slow effect with the final numbers slowly dropping onto a value
// similar to the price is right money wheel.
// I tried several approaches, but I can not seem to nail it.
// A linear increment of a delay does not look right
//
// If anyone can improve upon it,  I would love to see what results they come up with.
// If you make it improve it, let me know at tom[at]damage[dot]cc
// It would be rad to build a full 36+2 position American style roulette wheel out of neopixels!
//
// Code is based on the adafruit neopixel 'strandtest' code.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Roulette.h>
#include <Meissner_Delay.h>


// constructor/destructor

Neopixel_roulette::Neopixel_roulette (Neopixel_power_t	bright,
				      Neopixel_mask_t	m,
				      Neopixel_iter_t	min_iter,
				      Neopixel_iter_t	max_iter,
				      Neopixel_iter_t	min_spins,
				      Neopixel_iter_t	max_spins,
				      unsigned long	delay_advance,
				      unsigned long	delay_done)

  :	Neopixel_action (min_iter, max_iter, bright, delay_advance, delay_done),
	mask		(m),
	state		(ROULETTE_STATE_INITIAL),
	spin_min	(min_spins),
	spin_max	(max_spins),
	num_advance	(0),
	max_advance	(0),
	clockwise	(false)
{
}

Neopixel_roulette::~Neopixel_roulette (void)
{
}



// begin function to initialize everything.  This is expected to be called only once.

void
Neopixel_roulette::begin (Neopixel_object      *neo,
			  Neopixel_mask_t	m,
			  Neopixel_power_t	bright,
			  Neopixel_iter_t	min_iter,
			  Neopixel_iter_t	max_iter,
			  Neopixel_iter_t	min_spins,
			  Neopixel_iter_t	max_spins,
			  unsigned long		delay_advance,
			  unsigned long		delay_done)
{
  neo_obj = neo;
  state    = ROULETTE_STATE_INITIAL;

  if (m)
    mask = m;

  if (bright)
    brightness = bright;

  if (min_iter)
    min_iterations = min_iter;

  if (max_iter)
    max_iterations = max_iter;

  if (min_spins)
    spin_min = min_spins;

  if (max_spins)
    spin_max = max_spins;

  if (delay_advance)
    delay_small = delay_advance;

  if (delay_done)
    delay_large = delay_done;

  iterations = random (min_iterations, max_iterations+1);

#ifdef ROULETTE_DEBUG
#ifndef NEOPIXEL_DEBUG
  Serial.begin (9600);
  delay (1000);

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;
#endif

    Serial.println ("==========");
    Serial.println ("Roulette begin");
    Serial.print   ("min_iterations  = ");	Serial.println (min_iterations);
    Serial.print   ("max_iterations  = ");	Serial.println (max_iterations);
    Serial.print   ("iterations      = ");	Serial.println (iterations);
    Serial.print   ("spin_iterations = ");	Serial.println (min_iterations);
    Serial.print   ("spin_iterations = ");	Serial.println (max_iterations);
    Serial.print   ("brightness      = ");	Serial.println (brightness);
    Serial.print   ("mask            = 0x");	Serial.println (mask, HEX);
    Serial.print   ("delay advance   = ");	Serial.println (delay_small);
    Serial.print   ("delay done      = ");	Serial.println (delay_large);
#endif

    if (neo)
      {
	num_pixels = neo_obj->max_pixels ();
	neo_obj->fill (0);
	neo_obj->show ();
      }
}


// Function to set up the actions, possibly setting iterations or brightness levels
// This is expected to be called each time we begin the roulette pattern

void
Neopixel_roulette::start (Neopixel_iter_t  iter, Neopixel_power_t bright)
{
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);
  state      = ROULETTE_STATE_INITIAL;

  if (bright)
    brightness = bright;

  if (neo_obj)
    {
      neo_obj->fill (0);
      neo_obj->show ();
      neo_obj->set_current (0);
    }

  // set timeout to fire immediately
  timeout.run_immediate ();

#ifdef ROULETTE_DEBUG
    Serial.println ("==========");
    Serial.println ("Roulette start");
    Serial.print   ("iterations      = ");	Serial.println (iterations);
    Serial.print   ("brightness      = ");	Serial.println (brightness);
    Serial.print   ("mask            = 0x");	Serial.println (mask, HEX);
#endif
}


// Advance One Position Function
// every time this function is called, the led is advanced 1 position on the wheel

void
Neopixel_roulette::advance_one (bool *do_show_p)
{
  Neopixel_index_t pos   = neo_obj->current ();
  Neopixel_power_t red	 = 0;
  Neopixel_power_t green = 0;
  Neopixel_power_t blue	 = 0;

  // clear the display
  neo_obj->fill (0);

  // first pixel is always green, then alternate red/blue
  if (pos == 0)
    green = brightness;
  else if ((pos & 1) != 0)
    red = brightness;
  else
    blue = brightness;

  neo_obj->set_pixel_color_32bit_mask (pos, pack_rgb (red, green, blue), mask);
  neo_obj->advance (clockwise);

  if (do_show_p)
    *do_show_p = true;
  else
    neo_obj->show ();
}


// Function called from the main loop function to possibly do the next action.
// Return false, when all of the current actions are finished.  If do_show_p is
// not a NULL pointer, the function will not call show, but instead set
// *do_show_p to true.  It is assumed that the user has set *do_show_p to false
// before the call.

bool
Neopixel_roulette::loop (bool *do_show_p)
{
  bool ret = true;

  if (neo_obj && timeout.delay_done ())
    {
      switch (state)
	{
	// initial state
	case ROULETTE_STATE_INITIAL:
	  state = ROULETTE_STATE_SPINNING;
	  num_advance = 0;
	  max_advance = random (spin_min, spin_max+1) * num_pixels;

	  if ((mask & NEOPIXEL_MASK_MIRROR_RANDOM) != 0)
	    {
	      if (random (0, 2) == 0)
		mask &= ~NEOPIXEL_MASK_MIRROR_REVERSE;
	      else
		mask |= NEOPIXEL_MASK_MIRROR_REVERSE;
	    }

#ifdef ROULETTE_DEBUG
	  Serial.println ("==========");
	  Serial.println ("Roulette initialize");
	  Serial.print   ("iterations      = ");	Serial.println (iterations);
	  Serial.print   ("max advance     = ");	Serial.println (max_advance);
	  Serial.print   ("brightness      = ");	Serial.println (brightness);
	  Serial.print   ("mask            = 0x");	Serial.println (mask, HEX);
#endif
	  // fall through

	// spinning at full speed
	case ROULETTE_STATE_SPINNING:
	  advance_one (do_show_p);

	  // set timeout for next iteration
	  timeout.run_delay (delay_small);

	  // Is it time to settle down?
	  if (++num_advance >= max_advance)
	    {
	      state	  = ROULETTE_STATE_SETTLE;
	      num_advance = 0;
	      max_advance = (num_pixels * 6) + random (0, num_pixels);

#ifdef ROULETTE_DEBUG
	      Serial.println ("==========");
	      Serial.println ("Roulette settle");
	      Serial.print   ("iterations      = ");	Serial.println (iterations);
	      Serial.print   ("max advance     = ");	Serial.println (max_advance);
	      Serial.print   ("brightness      = ");	Serial.println (brightness);
	      Serial.print   ("mask            = 0x");	Serial.println (mask, HEX);
#endif
	    }
	  break;

	// slow down and 'settle' on final number
	case ROULETTE_STATE_SETTLE:
	  advance_one (do_show_p);

	  // set timeout to be progressively slower
	  timeout.run_delay (((num_advance + delay_small - 1) / delay_small) + delay_small);

	  if (++num_advance >= max_advance)
	    {
	      state = ROULETTE_STATE_INITIAL;
	      if (--iterations == 0)
		ret = false;

	      // set timeout to let the user see the last position
	      timeout.run_delay (delay_large);

#ifdef ROULETTE_DEBUG
	      Serial.println ("==========");
	      Serial.println ("Roulette done");
	      Serial.print   ("iterations      = ");	Serial.println (iterations);
	      Serial.print   ("ret             = ");	Serial.println (ret ? "true" : "false");
#endif
	    }
	}
    }

  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Roulette.cpp,v $
// Revision 1.10  2016/11/05 12:16:07  michaelmeissner
// SetPixelColor -> set_pixel_color_32bit_mask
//
// Revision 1.9  2016/11/04 15:11:28  michaelmeissner
// Rename neopixel object -> neo_obj.
//
// Revision 1.8  2015/02/15 17:35:57  michaelmeissner
// Reorder constructor to avoid -Wreorder warnings.
//
// Revision 1.7  2015/02/01 15:19:02  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.6  2015/01/11 14:45:40  michaelmeissner
// Change timeout functions.
//
// Revision 1.5  2015/01/10 22:54:43  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.4  2015/01/09 12:52:33  michaelmeissner
// Delete num_pixels, superclass now defines it.  Call show with no arguments.
//
// Revision 1.3  2015/01/09 05:41:37  michaelmeissner
// correct off by one errors in random calls.
//
// Revision 1.2  2015/01/09 05:39:01  michaelmeissner
// Add debug support.
//
// Revision 1.1  2015/01/09 05:31:08  michaelmeissner
// Initial version.
//

