// -*- arduino -*-
// Initial code from: http://damage-designs.com/
// but heavily modified by Michael Meissner (arduino@the-meissners.org)
//
// Quote from original author:
// Neopixel Ring Roulette wheel sketch.
// "Spins" a roulette wheel and lands on a random value.
// I am not 100% satisfied with the aesthetic, but its as close as I can get.
// I would prefer a more realistic spinning to a slow effect with the final numbers slowly dropping onto a value
// similar to the price is right money wheel.
// I tried several approaches, but I can not seem to nail it.
// A linear increment of a delay does not look right
//
// If anyone can improve upon it,  I would love to see what results they come up with.
// If you make it improve it, let me know at tom[at]damage[dot]cc
// It would be rad to build a full 36+2 position American style roulette wheel out of neopixels!
//
// Code is based on the adafruit neopixel 'strandtest' code.

#ifndef MEISSNER_NEOPIXEL_ROULETTE_H
#define MEISSNER_NEOPIXEL_ROULETTE_H	1

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Delay.h>

// #define ROULETTE_DEBUG	1				// define to enable debugging

const unsigned long	ROULETTE_DELAY_ADVANCE	= 4;		// timeout between pixel advances
const unsigned long	ROULETTE_DELAY_DONE	= 2000;		// timeout after doing the whole iteration
const Neopixel_power_t	ROULETTE_BRIGHTNESS	= 30;		// max brightness for red/green
const Neopixel_iter_t	ROULETTE_MIN_ITERATIONS	= 4;		// min # of iterations
const Neopixel_iter_t	ROULETTE_MAX_ITERATIONS = 16;		// max # of iterations
const Neopixel_iter_t	ROULETTE_MIN_SPINS	= 2;		// min # of fast spins to use
const Neopixel_iter_t	ROULETTE_MAX_SPINS	= 20;		// max # of fast spins to use

// state machine for the roulette wheel
typedef enum {
  ROULETTE_STATE_INITIAL,					// initial state
  ROULETTE_STATE_SPINNING,					// spinning at full speed
  ROULETTE_STATE_SETTLE						// slow down and 'settle' on final number
} Roulette_state_t;

class Neopixel_roulette : Neopixel_action {
 private:
  Neopixel_mask_t	mask;					// mask to pass to setPixelColor
  Roulette_state_t	state;					// state the roulette wheel is in
  Neopixel_iter_t	spin_min;				// min # of fast spins to use
  Neopixel_iter_t	spin_max;				// max # of fast spins to use
  Neopixel_iter_t	num_advance;				// how many times we have called advance_one in the current cycle
  Neopixel_iter_t	max_advance;				// how many times to call advance_one before going to the next state
  bool			clockwise;				// whether to go clockwise or counter-clockwise

  // advance one position
  void advance_one (bool *);

 public:
  // constructor/destructor
  Neopixel_roulette (Neopixel_power_t	bright		= ROULETTE_BRIGHTNESS,
		     Neopixel_mask_t	mask		= NEOPIXEL_MASK_MIRROR_RANDOM,
		     Neopixel_iter_t	min_iter	= ROULETTE_MIN_ITERATIONS,
		     Neopixel_iter_t	max_iter	= ROULETTE_MAX_ITERATIONS,
		     Neopixel_iter_t	min_spins	= ROULETTE_MIN_SPINS,
		     Neopixel_iter_t	max_spins	= ROULETTE_MAX_SPINS,
		     unsigned long	delay_advance	= ROULETTE_DELAY_ADVANCE,
		     unsigned long	delay_done	= ROULETTE_DELAY_DONE);

  ~Neopixel_roulette (void);

  // begin function to initialize everything
  void begin (Neopixel_object  *neo,
	      Neopixel_mask_t	mask			= NEOPIXEL_MASK_NONE,
	      Neopixel_power_t	bright			= 0,
	      Neopixel_iter_t	min_iter		= 0,
	      Neopixel_iter_t	max_iter		= 0,
	      Neopixel_iter_t	min_spins		= 0,
	      Neopixel_iter_t	max_spins		= 0,
	      unsigned long	delay_advance		= 0,
	      unsigned long	delay_done		= 0);

  // Function to set up the actions, possibly setting iterations or brightness levels
  void start (Neopixel_iter_t  iter   = 0,
	      Neopixel_power_t bright = 0);

  // Function called from the main loop function to possibly do the next action.
  // Return false, when all of the current actions are finished.
  // If do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.  It is assumed that the user has set
  // *do_show_p  to false before the call.
  bool loop (bool *do_show_p = NULL);

  // Return a string giving the name of the action for debugging
  const char *name (void)
  {
    return "Roulette";
  }
};

#endif	/* MEISSNER_NEOPIXEL_ROULETTE_H */


// HISTORY
// $Log: Meissner_Neopixel_Roulette.h,v $
// Revision 1.7  2015/02/01 15:18:02  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.6  2015/01/12 05:35:19  michaelmeissner
// Add virtual function to return the action name.
//
// Revision 1.5  2015/01/10 22:53:29  michaelmeissner
// Change advance_one signature
//
// Revision 1.4  2015/01/10 22:52:15  michaelmeissner
// Add optional argument to loop to return if show should be called.
//
// Revision 1.3  2015/01/09 12:51:08  michaelmeissner
// Delete num_pixels, superclass now defines it.
//
// Revision 1.2  2015/01/09 05:33:19  michaelmeissner
// Make begin not change defaults, unless the user passed new values.
//
// Revision 1.1  2015/01/09 05:31:25  michaelmeissner
// Initial version.
//
