// -*- arduino -*-
// Simple pattern, to fill a ring with a solid color, and after awhile go on to
// the next color.  The idea is to have a simple pattern that is not
// distracting to people like patterns that chase or choose random pixels do
//
// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Solid_Color.h>
#include <Meissner_Delay.h>

//#define SOLID_COLOR_DEBUG 1

// Possible color choices
const size_t NUM_COLORS	= 7;

static Neopixel_rgb_t colors[NUM_COLORS];


// Constructor, destructor
Neopixel_solid_color::Neopixel_solid_color (Neopixel_iter_t  max_iter,
					    Neopixel_iter_t  min_iter,
					    Neopixel_power_t bright,
					    unsigned long    delay_color)

  : Neopixel_action (max_iter, min_iter, bright, 0, delay_color),
    current_color (0)
{
}

Neopixel_solid_color::~Neopixel_solid_color (void)
{
}


// Setup for solid colors.  This is expected to be called only once.

void
Neopixel_solid_color::begin (Neopixel_object  *neo,
			     Neopixel_power_t  bright,
			     Neopixel_iter_t   min_iter,
			     Neopixel_iter_t   max_iter,
			     unsigned long     delay_color)
{
  neo_obj = neo;

  if (bright)
    brightness = bright;

  if (min_iter)
    min_iterations = min_iter;

  if (max_iter)
    max_iterations = max_iter;

  if (delay_color)
    delay_large = delay_color;

  current_color = random (0, NUM_COLORS);
  iterations	= random (min_iterations, max_iterations+1);

  if (neo_obj)
    num_pixels = neo_obj->max_pixels ();

#ifdef SOLID_COLOR_DEBUG
#ifndef NEOPIXEL_DEBUG
    Serial.begin (9600);
    delay (1000);
#endif

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;

    Serial.println ("==========");
    Serial.print   ("Solid color begin, num pixels = ");
    Serial.print   (num_pixels);
    Serial.print   (", brightness = ");
    Serial.print   (brightness);
    Serial.print   (", min iterations = ");
    Serial.print   (min_iterations);
    Serial.print   (", max iterations = ");
    Serial.print   (max_iterations);
    Serial.print   (", delay large = ");
    Serial.print   (delay_large);
    Serial.print   (", current color = ");
    Serial.print   (current_color);
    Serial.print   (", iterations = ");
    Serial.println (iterations);
#endif
}


// Convert red, green, blue into a 32-bit value

static Neopixel_rgb_t
rgb (Neopixel_power_t red,
     Neopixel_power_t green,
     Neopixel_power_t blue)
{
  return (  (((Neopixel_rgb_t)(red))   << 16)
	  | (((Neopixel_rgb_t)(green)) <<  8)
	  | (((Neopixel_rgb_t)(blue))  <<  0));
}


// Setup for the solid colors.  This is expected to be called each time we
// switch to doing the solid color sequence to reset the information, and start
// the timer.

void
Neopixel_solid_color::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);

  if (bright)
    brightness = bright;

  // setup the colors now
  colors[0] = rgb (brightness,	 0,		0);			// all red
  colors[1] = rgb (0,		 brightness,	0);			// all green
  colors[2] = rgb (0,		 0,		brightness);		// all blue
  colors[3] = rgb (brightness/2, brightness/2,	0);			// red+green
  colors[4] = rgb (brightness/2, 0,		brightness/2);		// red+blue
  colors[5] = rgb (0,		 brightness/2,	brightness/2);		// green+blue
  colors[6] = rgb (brightness/3, brightness/3,	brightness/3);		// white

  // start the timer to go off imediately on the first call to loop
  timeout.run_immediate ();

#ifdef SOLID_COLOR_DEBUG
    Serial.print   ("Solid color start, iterations = ");
    Serial.print   (iterations);
    Serial.print   (", brightness = ");
    Serial.println (brightness);
#endif
}


// Simple solid color test If do_show_p is not a NULL pointer, the function
// will not call show, but instead set *do_show_p to true.  It is assumed that
// the user has set *do_show_p to false before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool
Neopixel_solid_color::loop (bool *do_show_p)
{
  size_t next_color;
  bool ret = true;

  if (!neo_obj)
    return false;

  if (neo_obj->busy ())
    return true;

  if (timeout.delay_done ())
    {
      // set up next timeout
      timeout.run_delay (delay_large);

      do
	{
	  next_color = random (0, NUM_COLORS);
	}
      while (next_color == current_color);

      current_color = next_color;
      neo_obj->fill (colors[current_color]);

      if (do_show_p)
	*do_show_p = true;
      else
	neo_obj->show ();

      if (--iterations == 0)
	{
	  ret = false;
	  iterations = random (min_iterations, max_iterations);

#ifdef SOLID_COLOR_DEBUG
	  Serial.print   ("Solid color loop, delay = ");
	  Serial.print   (delay_large);
	  Serial.print   (", current color = 0x");
	  Serial.print   (colors[current_color], HEX);
	  Serial.print   (", iterations = ");
	  Serial.print   (iterations);
	  Serial.print   (", ret = ");
	  Serial.println (ret ? "true" : "false");
#endif
	}
    }
      
  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Solid_Color.cpp,v $
// Revision 1.10  2022/08/28 21:49:38  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.9  2022/02/27 05:05:52  michaelmeissner
// Update debug print statements.
//
// Revision 1.8  2022/02/13 01:07:18  michaelmeissner
// Update copyright year.
//
// Revision 1.7  2020/02/04 13:01:45  michaelmeissner
// Update copyright
//
// Revision 1.6  2019/01/14 04:43:23  michaelmeissner
// Update copyright.
//
// Revision 1.5  2016/11/04 15:12:32  michaelmeissner
// Rename neopixel object -> neo_obj.
//
// Revision 1.4  2016/04/03 22:49:41  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.3  2016/03/21 14:05:33  michaelmeissner
// Update copyright year.
//
// Revision 1.2  2015/02/01 15:20:37  michaelmeissner
// Assume do_show_p is set to false before the call.
//
// Revision 1.1  2015/01/16 15:18:46  michaelmeissner
// Initial version.
//
