// -*- c++ -*-
// Theaterchase from Adafruit strandtest:
// https://github.com/adafruit/Adafruit_NeoPixel/blob/master/examples/strandtest/strandtest.ino
// this code was completely rewritten by Michael Meissner (arduino@the-meissners.org)

// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Theater_Chase.h>
#include <Meissner_Delay.h>


// Convert red, green, blue into a 32-bit value

static Neopixel_rgb_t
rgb (Neopixel_power_t red,
     Neopixel_power_t green,
     Neopixel_power_t blue)
{
  return (  (((Neopixel_rgb_t)(red))   << 16)
	  | (((Neopixel_rgb_t)(green)) <<  8)
	  | (((Neopixel_rgb_t)(blue))  <<  0));
}


// Constructor, destructor
Neopixel_theater_chase::Neopixel_theater_chase (Neopixel_power_t bright,
						Neopixel_mask_t  mask,
						Neopixel_iter_t  min_iter,
						Neopixel_iter_t  max_iter,
						unsigned long    delay_led)

  : Neopixel_action (min_iter, max_iter, bright, 0, delay_led),
    current_mask  (mask),
    current_color (0),
    state	  (THEATER_CHASE_WHITE),
    offset	  (0)
{
}

Neopixel_theater_chase::~Neopixel_theater_chase (void)
{
}


// Setup for theater_chase.  This is expected to be called only once.

void
Neopixel_theater_chase::begin (Neopixel_object  *neo,
			       Neopixel_power_t  bright,
			       Neopixel_mask_t   mask,
			       Neopixel_iter_t   min_iter,
			       Neopixel_iter_t   max_iter,
			       unsigned long     delay_led)
{
  neo_obj = neo;
  state   = THEATER_CHASE_WHITE;
  offset  = 0;

  if (bright)
    brightness = bright;

  if (min_iter)
    min_iterations = min_iter;

  if (max_iter)
    max_iterations = max_iter;

  if (delay_led)
    delay_large = delay_led;

  iterations = random (min_iterations, max_iterations+1);
  current_color = rgb (brightness/3, brightness/3, brightness/3);

  if (neo_obj)
    num_pixels = neo_obj->max_pixels ();

#ifdef THEATER_CHASE_DEBUG
#ifndef NEOPIXEL_DEBUG
    Serial.begin (9600);
    delay (1000);

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;
#endif

    Serial.println ("==========");
    Serial.println ("Theater chase begin");
#endif
}


// Setup for the theater_chase.  This is expected to be called each time we switch
// to doing the theater_chase sequence to reset the information, and start the
// timer.

void
Neopixel_theater_chase::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);

  if (bright)
    brightness = bright;

  state = THEATER_CHASE_WHITE;
  offset = 0;
  current_color = rgb (brightness/3, brightness/3, brightness/3);	// white color

  // start the timer to go off imediately on the first call to loop
  timeout.run_immediate ();

  // clear LEDs, get # pixels
  if (neo_obj)
    {
      num_pixels = neo_obj->max_pixels ();
      neo_obj->fill (0);
    }

#ifdef SOLID_COLOR_DEBUG
    Serial.println ("==========");
    Serial.print   ("Theater chase, iterations = ");
    Serial.print   (iterations);
    Serial.print   (", brightness = ");
    Serial.print   (brightness);
#endif
}


// Simple theater chase test. If do_show_p is not a NULL pointer, the function
// will not call show, but instead set *do_show_p to true.  It is assumed that
// the user has set *do_show_p to false before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool
Neopixel_theater_chase::loop (bool *do_show_p)
{
  bool ret = true;

  if (!neo_obj)
    return false;

  if (neo_obj->busy ())
    return true;

  if (timeout.delay_done ())
    {
      // set up next timeout
      timeout.run_delay (delay_large);

      // Write pixels every 3rd led
      neo_obj->fill (0);
      for (Neopixel_index_t i = offset; i < num_pixels; i += THEATER_CHASE_MAX_OFFSET)
	neo_obj->set_pixel_color_32bit_mask (i, current_color, current_mask);

      if (do_show_p)
	*do_show_p = true;
      else
	neo_obj->show ();

      // Set up for the next offset, if we've wrapped around, go to the next color
      if (++offset >= THEATER_CHASE_MAX_OFFSET)
	{
	  offset = 0;

	  if (state == THEATER_CHASE_WHITE)
	    {
	      state = THEATER_CHASE_RED;
	      current_color = rgb (brightness, 0, 0);
	    }

	  else if (state == THEATER_CHASE_RED)
	    {
	      state = THEATER_CHASE_BLUE;
	      current_color = rgb (0, 0, brightness);
	    }

	  else
	    {
	      state = THEATER_CHASE_WHITE;
	      current_color = rgb (brightness/3, brightness/3, brightness/3);

	      if (--iterations == 0)
		{
		  ret = false;
		  iterations = random (min_iterations, max_iterations);
		}
	    }
	}

#ifdef THEATER_CHASE_DEBUG
      Serial.println ("==========");
      Serial.print   ("Theater Chase, delay = ");
      Serial.print   (delay_large);
      Serial.print   (", current color = 0x");
      Serial.print   (current_color);
      Serial.print   (", offset = ");
      Serial.print   (offset);
      Serial.print   (", iterations = ");
      Serial.print   (iterations);
      Serial.print   (", ret = ");
      Serial.println (ret ? "true" : "false");
#endif
    }
      
  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Theater_Chase.cpp,v $
// Revision 1.10  2022/08/28 21:50:18  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.9  2022/02/13 01:07:42  michaelmeissner
// Update copyright year.
//
// Revision 1.8  2020/01/28 04:52:36  michaelmeissner
// Update copyright year.
//
// Revision 1.7  2019/01/14 04:43:57  michaelmeissner
// Update copyright.
//
// Revision 1.6  2016/11/05 12:35:04  michaelmeissner
// SetPixelColor -> set_pixel_color_32bit_mask
//
// Revision 1.5  2016/11/04 15:13:53  michaelmeissner
// Rename neopixel object -> neo_obj.
//
// Revision 1.4  2016/03/21 14:06:15  michaelmeissner
// Update copyright year.
//
// Revision 1.3  2015/02/15 17:37:07  michaelmeissner
// Reorder constructor to avoid -Wreorder warnings.
//
// Revision 1.2  2015/02/02 04:09:19  michaelmeissner
// Initial version.
//
