// -*- arduino -*-
// Meissner neopixel triple ring support
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// The triple ring has 3 rings of neopixels:
//	Pixels  0 - 23: Outer ring
//	Pixesl 24 - 39: Middle ring
//	Pixels 40 - 43: Inner ring

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Neopixel_Triple_Ring.h>
#include <Meissner_Delay.h>

// Possible color choices
const size_t NUM_COLORS	= 7;

static Neopixel_rgb_t colors[NUM_COLORS];


// Constructor, destructor
Neopixel_triple_ring::Neopixel_triple_ring (Neopixel_power_t bright,
					    Neopixel_mask_t  mask,
					    Neopixel_iter_t  min_iter,
					    Neopixel_iter_t  max_iter,
					    Neopixel_iter_t  min_single,
					    Neopixel_iter_t  max_single,
					    unsigned long    delay_inner,
					    unsigned long    delay_whole)

      :	Neopixel_action  (min_iter, max_iter, bright, delay_inner, delay_whole),
	mask_current	 (mask),
	updown_p	 (false),
	direction_up	 (true),
	state		 (TRIPLE_RING_INITIAL),
	start_pin	 (0),
	background	 (0),
	foreground	 (0),
	min_single_count (min_single),
	max_single_count (max_single),
	single_cycle	 (0),
	current_cycle	 (0)
{
}

Neopixel_triple_ring::~Neopixel_triple_ring (void)
{
}


// begin function to initialize everything.
// This is expected to be called only once.

void
Neopixel_triple_ring::begin (Neopixel_object *neo,
			     Neopixel_power_t	triple_ring_brightness,
			     Neopixel_mask_t	triple_ring_mask,
			     Neopixel_iter_t	triple_ring_min_iterations,
			     Neopixel_iter_t	triple_ring_max_iterations,
			     Neopixel_iter_t	triple_ring_min_single_count,
			     Neopixel_iter_t	triple_ring_max_single_count,
			     unsigned long	triple_ring_delay_inner,
			     unsigned long	triple_ring_delay_whole)
{
  neo_obj = neo;
  mask_current = triple_ring_mask;

  if (triple_ring_brightness)
    brightness = triple_ring_brightness;

  if (triple_ring_min_iterations)
    min_iterations = triple_ring_min_iterations;

  if (triple_ring_max_iterations)
    max_iterations = triple_ring_max_iterations;

  if (triple_ring_min_single_count)
    min_single_count = triple_ring_min_single_count;

  if (triple_ring_max_single_count)
    max_single_count = triple_ring_max_single_count;

  if (triple_ring_delay_inner)
    delay_small = triple_ring_delay_inner;

  if (triple_ring_delay_whole)
    delay_large = triple_ring_delay_whole;

  state = TRIPLE_RING_INITIAL;

  if (neo_obj)
    num_pixels = neo_obj->max_pixels ();

#ifdef TRIPLE_RING_DEBUG
#ifndef NEOPIXEL_DEBUG
    Serial.begin (9600);
    delay (1000);

    // wait up to 3 seconds for the Serial device to become available
    long unsigned debug_start = millis ();
    while (!Serial && ((millis () - debug_start) <= 3000))
      ;
#endif

    Serial.println ("==========");
    Serial.println ("Triple_Ring begin");
#endif
}


// Setup for triple ring.  This is expected to be called each time we switch to doing the rotate
// sequence to reset the information, and start the timer.

void
Neopixel_triple_ring::start (Neopixel_iter_t iter, Neopixel_power_t bright)
{
  state	     = TRIPLE_RING_INITIAL;
  iterations = (iter) ? iter : random (min_iterations, max_iterations+1);

  if (bright)
    brightness = bright;

  // setup the colors now
  colors[0] = pack_rgb (brightness,	0,		0);			// all red
  colors[1] = pack_rgb (0,		brightness,	0);			// all green
  colors[2] = pack_rgb (0,		0,		brightness);		// all blue
  colors[3] = pack_rgb (brightness/2,	brightness/2,	0);			// red+green
  colors[4] = pack_rgb (brightness/2,	0,		brightness/2);		// red+blue
  colors[5] = pack_rgb (0,		brightness/2,	brightness/2);		// green+blue
  colors[6] = pack_rgb (brightness/3,	brightness/3,	brightness/3);		// white

  // start the timer to go off imediately on the first call to loop
  timeout.run_immediate ();

#ifdef TRIPLE_RING_DEBUG
    Serial.println ("==========");
    Serial.print   ("Triple_Ring start, iterations = ");
    Serial.print   (iterations);
    Serial.print   (", brightness = ");
    Serial.println (brightness);
#endif
}


// loop function to be called from the main loop, return false when all
// iterations are done If do_show_p is not a NULL pointer, the function will
// not call show, but instead set *do_show_p to true.  It is assumed that the
// user has set *do_show_p to false before the call.
//
// If we are using WS2812Serial, skip this loop if the previous action is still
// completing intiializing the colors.

bool Neopixel_triple_ring::loop (bool *do_show_p)
{
  size_t f, b;
  Neopixel_index_t pixel;
  bool ret = true;

  if (!neo_obj)
    return false;

  if (neo_obj->busy ())
    return true;

  if (timeout.delay_done ())
    {
      // set up next timeout
      timeout.run_delay (delay_small);

      switch (state)
	{
	  // Initial state, set up defaults for chasing the outer ring
	case TRIPLE_RING_INITIAL:
	  current_cycle = 0;
	  single_cycle	= random (min_single_count, max_single_count+1);
	  start_pin	= random (TRIPLE_RING_OUTER_START, TRIPLE_RING_OUTER_END + 1);
	  state		= TRIPLE_RING_OUTER_SINGLE_PIXEL;

	  do
	    {
	      f = random (0, NUM_COLORS);
	      b = random (0, NUM_COLORS);
	    }
	  while (f == b);

	  foreground = colors[f];
	  background = colors[b];

	  neo_obj->set_current (start_pin);
	  neo_obj->fill (background);
	  neo_obj->set_pixel_color_32bit_mask (start_pin, foreground, mask_current);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

	  if ((mask_current & NEOPIXEL_MASK_MIRROR_RANDOM) != 0)
	    {
	      if (random (0, 2) == 0)
		mask_current &= ~NEOPIXEL_MASK_MIRROR_REVERSE;
	      else
		mask_current |= NEOPIXEL_MASK_MIRROR_REVERSE;
	    }

#ifdef TRIPLE_RING_DEBUG
	  Serial.println ("==========");
	  Serial.print   ("Triple_Ring setup outer ring, pixels = ");
	  Serial.println (num_pixels);
	  Serial.print   ("delay small = ");
	  Serial.print   (delay_small);
	  Serial.print   (", delay large = ");
	  Serial.print   (delay_large);
	  Serial.print   (", led pin = ");
	  Serial.println (neo_obj->led_pin ());
	  Serial.print   ("foreground = 0x");
	  Serial.print   (f);
	  Serial.print   (" (");
	  Serial.print   (foreground, HEX);
	  Serial.print   (")");
	  Serial.print   (", background = 0x");
	  Serial.print   (b);
	  Serial.print   (" (");
	  Serial.print   (background, HEX);
	  Serial.println (")");
	  Serial.print   ("iterations = ");
	  Serial.print   (iterations);
	  Serial.print   (", first pixel = ");
	  Serial.print   (start_pin);
	  Serial.print   (", flip = ");
	  Serial.print   (((mask_current & NEOPIXEL_MASK_MIRROR_REVERSE) != 0) ? "true" : "false");
	  Serial.print   (", single cycle = ");
	  Serial.println (single_cycle);
#endif
	  // fall through to doing single pixel on the outer ring

	  // set a single pixel on the outer ring, resetting the previous one
	case TRIPLE_RING_OUTER_SINGLE_PIXEL:
	  pixel = neo_obj->current ();

#ifdef TRIPLE_RING_DEBUG
	  Serial.printf ("Triple Ring single pixel, outer ring, pixel = %2d, foreground = 0x%.6lx, background = 0x%.6lx, single_cycle = %d\n",
			 pixel,
			 (unsigned long) foreground,
			 (unsigned long) background,
			 single_cycle);
#endif

	  neo_obj->set_pixel_color_32bit_mask (pixel, background, mask_current);
	  pixel = neo_obj->current () + 1;
	  if (pixel > TRIPLE_RING_OUTER_END)
	    pixel = TRIPLE_RING_OUTER_START;

	  neo_obj->set_current (pixel);
	  neo_obj->set_pixel_color_32bit_mask (pixel, foreground, mask_current);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

	  if (pixel == start_pin)
	    {
	      if (--single_cycle == 0)
		{
		  state = TRIPLE_RING_OUTER_ALL_PIXELS;

#ifdef TRIPLE_RING_DEBUG
		  Serial.println ("==========");
		  Serial.println ("Go to all pixels state for outer ring");
#endif
		}
	      else
		{
#ifdef TRIPLE_RING_DEBUG
		  Serial.println ("==========");
		  Serial.print   ("Single pixel, single_cycle for outer ring = ");
		  Serial.println (single_cycle);
#endif
		}
	    }

	  break;

	  // Set all pixels for the final iteration
	case TRIPLE_RING_OUTER_ALL_PIXELS:
	  pixel = neo_obj->current () + 1;
	  if (pixel > TRIPLE_RING_OUTER_END)
	    pixel = TRIPLE_RING_OUTER_START;

	  neo_obj->set_current (pixel);
	  neo_obj->set_pixel_color_32bit_mask (pixel, foreground, mask_current);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

#ifdef TRIPLE_RING_DEBUG
	  Serial.printf ("Triple_Ring all pixels outer ring, pixel = %2d, foreground = 0x%.6lx\n",
			 pixel,
			 (unsigned long) foreground);
#endif

	  if (pixel == start_pin)
	    {
	      current_cycle = 0;
	      single_cycle  = random (min_single_count, max_single_count+1);
	      state         = TRIPLE_RING_MIDDLE_SINGLE_PIXEL;
	      start_pin     = random (TRIPLE_RING_MIDDLE_START, TRIPLE_RING_MIDDLE_END + 1);
	      neo_obj->set_current (start_pin == TRIPLE_RING_MIDDLE_END ? TRIPLE_RING_MIDDLE_START : start_pin + 1);
	      timeout.run_delay (delay_small);

#ifdef TRIPLE_RING_DEBUG
	      Serial.println ("==========");
	      Serial.println ("Go to middle ring state");
#endif
	    }

	  break;


	  // set a single pixel on the middle ring, resetting the previous one
	case TRIPLE_RING_MIDDLE_SINGLE_PIXEL:
	  pixel = neo_obj->current ();

#ifdef TRIPLE_RING_DEBUG
	  Serial.printf ("Triple Ring single pixel, middle ring, pixel = %2d, foreground = 0x%.6lx, background = 0x%.6lx, single_cycle = %d\n",
			 pixel,
			 (unsigned long) foreground,
			 (unsigned long) background,
			 single_cycle);
#endif

	  neo_obj->set_pixel_color_32bit_mask (pixel, background, mask_current);
	  pixel = neo_obj->current () + 1;
	  if (pixel > TRIPLE_RING_MIDDLE_END)
	    pixel = TRIPLE_RING_MIDDLE_START;

	  neo_obj->set_current (pixel);
	  neo_obj->set_pixel_color_32bit_mask (pixel, foreground, mask_current);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

	  if (pixel == start_pin)
	    {
	      if (--single_cycle == 0)
		{
		  state = TRIPLE_RING_MIDDLE_ALL_PIXELS;

#ifdef TRIPLE_RING_DEBUG
		  Serial.println ("==========");
		  Serial.println ("Go to all pixels state for middle ring");
#endif
		}
	      else
		{
#ifdef TRIPLE_RING_DEBUG
		  Serial.println ("==========");
		  Serial.print   ("Single pixel, single_cycle for middle ring = ");
		  Serial.println (single_cycle);
#endif
		}
	    }

	  break;

	  // Set all pixels for the final iteration of the middle ring
	case TRIPLE_RING_MIDDLE_ALL_PIXELS:
	  pixel = neo_obj->current () + 1;
	  if (pixel > TRIPLE_RING_MIDDLE_END)
	    pixel = TRIPLE_RING_MIDDLE_START;

	  neo_obj->set_current (pixel);
	  neo_obj->set_pixel_color_32bit_mask (pixel, foreground, mask_current);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

#ifdef TRIPLE_RING_DEBUG
	  Serial.printf ("Triple_Ring all pixels middle ring, pixel = %2d, foreground = 0x%.6lx\n",
			 pixel,
			 (unsigned long) foreground);
#endif

	  if (pixel == start_pin)
	    {
	      state = TRIPLE_RING_INNER_ALL_PIXELS;
	      neo_obj->set_current (TRIPLE_RING_INNER_START - 1);
	      timeout.run_delay (delay_small);

#ifdef TRIPLE_RING_DEBUG
	      Serial.println ("==========");
	      Serial.println ("Go to middle ring state");
#endif
	    }

	  break;

	  // Set all pixels for the final iteration of the inner 4 pixels
	case TRIPLE_RING_INNER_ALL_PIXELS:
	  pixel = neo_obj->next (true);
	  neo_obj->set_pixel_color_32bit_mask (pixel, foreground, mask_current);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

#ifdef TRIPLE_RING_DEBUG
	  Serial.printf ("Triple_Ring all pixels inner square, pixel = %2d, foreground = 0x%.6lx\n",
			 pixel,
			 (unsigned long) foreground);
#endif

	  if (pixel == TRIPLE_RING_INNER_END)
	    {
	      state = TRIPLE_RING_DONE;
	      timeout.run_delay (delay_large);

#ifdef TRIPLE_RING_DEBUG
	      Serial.println ("==========");
	      Serial.println ("Go to triple ring done state");
#endif
	    }

	  break;

	  // Done, reset things and set up for next iteration
	case TRIPLE_RING_DONE:
	  iterations--;
	  state = TRIPLE_RING_INITIAL;
	  neo_obj->fill (0);

	  if (do_show_p)
	    *do_show_p = true;
	  else
	    neo_obj->show ();

#ifdef TRIPLE_RING_DEBUG
	  Serial.println ("==========");
	  Serial.print ("Triple_Ring end of cycle, iterations = ");
	  Serial.println (iterations);
#endif

	  if (iterations == 0)
	    {
	      iterations = random (min_iterations, max_iterations+1);
	      ret	 = false;
	    }

	  break;

	  // Unknown, just bail out
	default:
	  ret = false;
#ifdef TRIPLE_RING_DEBUG
	  Serial.println ("==========");
	  Serial.println ("bad case");
#endif
	  break;
	}
    }

#ifdef TRIPLE_RING_DEBUG
  if (!ret)
    {
      Serial.println ("==========");
      Serial.println ("Triple_Ring returns false");
    }
#endif
      
  return ret;
}


// HISTORY
// $Log: Meissner_Neopixel_Triple_Ring.cpp,v $
// Revision 1.5  2022/08/28 21:51:08  michaelmeissner
// If we are using WS2812Serial objects, wait until the current action completes.
//
// Revision 1.4  2022/02/13 01:07:59  michaelmeissner
// Update copyright year.
//
// Revision 1.3  2020/01/28 04:52:44  michaelmeissner
// Update copyright year.
//
// Revision 1.2  2019/07/15 15:43:16  michaelmeissner
// Rewrite, adding chase on the outer, middle rings.
//
// Revision 1.1  2019/07/12 04:51:51  michaelmeissner
// Initial version.
//
