// -*- arduino -*-
// Meissner neopixel support for the Adafruit triple ring neopixel board.
// This is a modified chase ring.  First it does a chase on the outer ring,
// then on the middle ring, and finally on the inner square.
//
// Copyright (C) 2013-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// The triple ring has 3 rings of neopixels:
//	Pixels  0 - 23: Outer ring
//	Pixesl 24 - 39: Middle ring
//	Pixels 40 - 43: Inner square

#ifndef MEISSNER_NEOPIXEL_TRIPLE_RING_H
#define MEISSNER_NEOPIXEL_TRIPLE_RING_H	1

#include <Meissner_Config.h>
#include <Adafruit_NeoPixel.h>
#include <Meissner_Neopixel.h>
#include <Meissner_Delay.h>

// #define TRIPLE_RING_DEBUG	1			// define to enable debugging

const unsigned long	TRIPLE_RING_DELAY_INNER		= 50;		// timeout between changing the case pixel
const unsigned long	TRIPLE_RING_DELAY_WHOLE		= 3000;		// timeout after acheiving a single color
const Neopixel_iter_t	TRIPLE_RING_MAX_ITERATIONS	= 5;		// max # of iterations before going to the next pattern
const Neopixel_iter_t	TRIPLE_RING_MIN_ITERATIONS	= 2;		// min # of iterations before going to the next pattern
const Neopixel_iter_t	TRIPLE_RING_MAX_SINGLE_COUNT	= 8;		// max # of iterations of a single pixel
const Neopixel_iter_t	TRIPLE_RING_MIN_SINGLE_COUNT	= 3;		// min # of iterations of a single pixel
const Neopixel_power_t	TRIPLE_RING_BRIGHTNESS		= 10;		// max brightness to use

// Positions of the three rings
const Neopixel_pin_t	TRIPLE_RING_OUTER_NUM		= 24;		// # pixels in the outer ring
const Neopixel_pin_t	TRIPLE_RING_MIDDLE_NUM		= 16;		// # pixels in the middle ring
const Neopixel_pin_t	TRIPLE_RING_INNER_NUM		=  4;		// # pixels in the inner square
const Neopixel_pin_t	TRIPLE_RING_NUM			= 44;		// total # pixels in the triple ring
const Neopixel_pin_t	TRIPLE_RING_OUTER_START		=  0;
const Neopixel_pin_t	TRIPLE_RING_OUTER_END		= TRIPLE_RING_OUTER_START  + TRIPLE_RING_OUTER_NUM - 1;
const Neopixel_pin_t	TRIPLE_RING_MIDDLE_START	= TRIPLE_RING_OUTER_END    + 1;
const Neopixel_pin_t	TRIPLE_RING_MIDDLE_END		= TRIPLE_RING_MIDDLE_START + TRIPLE_RING_MIDDLE_NUM - 1;
const Neopixel_pin_t	TRIPLE_RING_INNER_START		= TRIPLE_RING_MIDDLE_END   + 1;
const Neopixel_pin_t	TRIPLE_RING_INNER_END		= TRIPLE_RING_INNER_START  + TRIPLE_RING_INNER_NUM - 1;

typedef enum {
  TRIPLE_RING_INITIAL,					// initial state
  TRIPLE_RING_OUTER_SINGLE_PIXEL,			// changing a single pixel
  TRIPLE_RING_OUTER_ALL_PIXELS,				// changing all of the pixels
  TRIPLE_RING_MIDDLE_SINGLE_PIXEL,			// changing a single pixel
  TRIPLE_RING_MIDDLE_ALL_PIXELS,			// changing all of the pixels
  TRIPLE_RING_INNER_ALL_PIXELS,				// changing all of the pixels
  TRIPLE_RING_DONE					// done the whole cycle
} ring_state_t;

class Neopixel_triple_ring : Neopixel_action {
 private:
  Neopixel_mask_t		mask_current;		// current mask to pass to setPixelColor
  bool				updown_p;		// if we are doing up/down processing
  bool				direction_up;		// if we are currently ascending
  ring_state_t			state;			// current state
  Neopixel_pin_t		start_pin;		// start pin on the current ring
  Neopixel_rgb_t		background;		// background color
  Neopixel_rgb_t		foreground;		// foreground color
  Neopixel_iter_t		min_single_count;	// min # of iterations of a single pixel
  Neopixel_iter_t		max_single_count;	// max # of iterations of a single pixel
  Neopixel_iter_t		single_cycle;		// # of cycles with the single pixel
  Neopixel_iter_t		current_cycle;		// cycle # for doing the single pixel

 public:
  // constructor/destructor
  Neopixel_triple_ring (Neopixel_power_t bright			= TRIPLE_RING_BRIGHTNESS,
			Neopixel_mask_t  mask			= NEOPIXEL_MASK_MIRROR_RANDOM,
			Neopixel_iter_t	 min_iter		= TRIPLE_RING_MIN_ITERATIONS,
			Neopixel_iter_t	 max_iter		= TRIPLE_RING_MAX_ITERATIONS,
			Neopixel_iter_t  min_single_count	= TRIPLE_RING_MIN_SINGLE_COUNT,
			Neopixel_iter_t	 max_single_count	= TRIPLE_RING_MAX_SINGLE_COUNT,
			unsigned long	 delay_inner		= TRIPLE_RING_DELAY_INNER,
			unsigned long	 delay_whole		= TRIPLE_RING_DELAY_WHOLE);

  ~Neopixel_triple_ring (void);

  // begin function to initialize everything
  void begin (Neopixel_object  *neo,
	      Neopixel_power_t	triple_ring_brightness		= TRIPLE_RING_BRIGHTNESS,
	      Neopixel_mask_t	triple_ring_mask		= NEOPIXEL_MASK_MIRROR_RANDOM,
	      Neopixel_iter_t	triple_ring_min_iterations	= TRIPLE_RING_MIN_ITERATIONS,
	      Neopixel_iter_t	triple_ring_max_iterations	= TRIPLE_RING_MAX_ITERATIONS,
	      Neopixel_iter_t	triple_ring_min_single_count	= TRIPLE_RING_MIN_SINGLE_COUNT,
	      Neopixel_iter_t	triple_ring_max_single_count	= TRIPLE_RING_MAX_SINGLE_COUNT,
	      unsigned long	triple_ring_delay_inner		= TRIPLE_RING_DELAY_INNER,
	      unsigned long	triple_ring_delay_whole		= TRIPLE_RING_DELAY_WHOLE);

  // Function to set up the actions, possibly setting iterations or brightness levels
  void start (Neopixel_iter_t iter = 0, Neopixel_power_t bright = 0);

  // Function called from the main loop function to possibly do the next action.
  // Return false, when all of the current actions are finished.
  // If do_show_p is not a NULL pointer, the function will not call show, but
  // instead set *do_show_p to true.  It is assumed that the user has set
  // *do_show_p  to false before the call.
  bool loop (bool *do_show_p = NULL);

  // Return a string giving the name of the action for debugging
  const char *name (void)
  {
    return "Triple_Ring";
  }
};

#endif	/* MEISSNER_NEOPIXEL_TRIPLE_RING_H */


// HISTORY
// $Log: Meissner_Neopixel_Triple_Ring.h,v $
// Revision 1.6  2022/02/13 01:03:06  michaelmeissner
// Update copyright year.
//
// Revision 1.5  2020/01/28 04:14:44  michaelmeissner
// Update copyright year.
//
// Revision 1.4  2019/07/15 15:46:43  michaelmeissner
// Update comment
//
// Revision 1.3  2019/07/15 15:45:26  michaelmeissner
// Speed up chase
//
// Revision 1.2  2019/07/15 15:42:58  michaelmeissner
// Rewrite, adding chase on the outer, middle rings.
//
// Revision 1.1  2019/07/12 04:51:36  michaelmeissner
// Initial version.
//
