// -*- arduino -*-
// Meissner random support file
//
// Copyright (C) 2015-2022, Michael Meissner (arduino@the-meissners.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#include <Meissner_Config.h>
#include <Meissner_Random.h>

#if defined (PROCESSOR_TEENSY_4_X)
#include <Meissner_Cpu_Info.h>
#endif

#if USE_EEPROM_LIB
#include <Meissner_Eeprom.h>
#endif

#if USE_EEPROM
#include <EEPROM.h>
#endif

//#define RANDOM_DEBUG	1					// define to enable debugging

#if defined(PROCESSOR_ATTINY85)
#include <avr/power.h>						// enable 16Khz on Trinket
#endif

const int		RANDOM_NUMBER_READS	= 5;		// read a couple of times from the floating pin
const unsigned long	RANDOM_NUMBER_DELAY	= 3;		// # ms to delay between reads

typedef unsigned int random_type;

// On squarewear, PIN_RANDOM is the same as the temp sensor, so avoid doing reads
// from PIN_RANDOM in that case.

#if defined(PROCESSOR_SQUAREWEAR)
#if PIN_RANDOM == PIN_TEMP_SENSOR
#define DO_ANALOG_READ		0
#else
#define DO_ANALOG_READ		1
#endif	/* PIN_RANDOM == PIN_TEMP_SENSOR */

#elif PIN_RANDOM >= 0	/* not squarewear */
#define DO_ANALOG_READ		1

#else
#define DO_ANALOG_READ		0
#endif	/* PROCESSOR_SQUAREWEAR */


// Combine two values in iterating to generate the next value for the seed argument

static uint32_t
combine_random (uint32_t a, uint32_t b)
{
  uint32_t rotate = (a << 3) | (a >> (32 - 3));
  uint32_t result = (rotate ^ b) & 0x7fffffffUL;

#ifdef RANDOM_DEBUG
  Serial.print ("Random combine (0x");
  Serial.print (a, HEX);
  Serial.print (", 0x");
  Serial.print (b, HEX);
  Serial.print (") = 0x");
  Serial.println (result, HEX);
#endif

  return result;
}


// Read internal temperature sensor if the processor has one

static uint32_t
read_internal_temp (void)
{
  uint32_t result;

#if defined(PROCESSOR_SQUAREWEAR)
  uint32_t light;

  pinMode (PIN_LIGHT_SENSOR, INPUT);
  pinMode (PIN_TEMP_SENSOR,  INPUT);

  result = analogRead (PIN_TEMP_SENSOR);
  light  = analogRead (PIN_LIGHT_SENSOR);
  if (light)
    result = combine_random (result, light);

#elif defined(PROCESSOR_ATMEGA328P) || defined(PROCESSOR_ATMEGA32U4)
  // from http://playground.arduino.cc/Main/InternalTemperatureSensor

  // Read temperature sensor against 1.1V reference
  ADMUX = _BV(REFS1) | _BV(REFS0) | _BV(MUX3);

  delay (RANDOM_NUMBER_DELAY);					// Wait for Vref to settle
  ADCSRA |= _BV(ADSC);						// Convert
  while (bit_is_set(ADCSRA,ADSC))
    ;

  result = ADCL;
  result |= ADCH<<8;
  result = (result - 125) * 1075;

#elif defined (PROCESSOR_TEENSY_3_X) && (PIN_TEMP_SENSOR > 0)
  // from http://dangerousprototypes.com/forum/viewtopic.php?t=4606
  result = analogRead (PIN_TEMP_SENSOR);

#elif defined (PROCESSOR_TEENSY_4_X)
  result = (uint32_t) (GetTemp_C () * 1000000.0f);

#else
  result = 0;
#endif

#ifdef RANDOM_DEBUG
  Serial.print ("Random internal temp: ");
  Serial.print (result, DEC);
  Serial.print (" (0x");
  Serial.print (result, HEX);
  Serial.println (")");
#endif

  return result;
}


// Set the random seed

void
set_random_seed (bool use_eeprom)
{
  uint32_t seed	= 0L;
  uint32_t tmp;

#ifdef RANDOM_DEBUG
  Serial.begin (9600);
  delay (1000UL);

  // wait up to 3 seconds for the Serial device to become available
  long unsigned debug_start = millis ();
  while (!Serial && ((millis () - debug_start) <= 3000))
    ;

  Serial.print ("Random debug, ");
#if USE_EEPROM
  Serial.print ("use_eeprom = ");
  Serial.print (use_eeprom ? "true" : "false");

#else	/* !USE_EEPROM */
  Serial.print ("eeprom not used");
#endif	/* !USE_EEPROM */

  Serial.print (", pin_random = ");
  Serial.print (PIN_RANDOM);
  Serial.print (", EEPROM version = ");
  Serial.print (EEPROM_VERSION);
  Serial.print (", EEPROM = ");
  Serial.print (EEPROM_FIRST_RESERVED);
  Serial.print (":");
  Serial.print (EEPROM_LAST_RESERVED);
  Serial.print (" (");
  Serial.print (EEPROM_LENGTH);
  Serial.println (") bytes");
#endif	/* RANDOM_DEBUG */

#if defined(PROCESSOR_ATTINY85)
  // Make sure Trinket 16Khz is set up
  if (F_CPU == 16000000)
    clock_prescale_set(clock_div_1);
#endif	/* 16 Mhz ATtiny 85 */

#if USE_EEPROM
  if (use_eeprom)
    {
#if USE_EEPROM_LIB
      seed = MRM_eeprom.get_uint32 (EEPROM_RANDOM);

#else	/* !USE_EEPROM_LIB */
      seed = (  (((uint32_t)(uint8_t)EEPROM.read (EEPROM_RANDOM + 0)) <<  0)
	      | (((uint32_t)(uint8_t)EEPROM.read (EEPROM_RANDOM + 1)) <<  8),
	      | (((uint32_t)(uint8_t)EEPROM.read (EEPROM_RANDOM + 2)) << 16),
	      | (((uint32_t)(uint8_t)EEPROM.read (EEPROM_RANDOM + 3)) << 24));

#endif	/* !USE_EEPROM_LIB */

      if (seed == 0xffffffffUL || seed == 0UL)
	seed = 1;
      else
	seed &= 0x7fffffffUL;

#ifdef RANDOM_DEBUG
      Serial.print ("Read random seed, seed: 0x");
      Serial.println (seed, HEX);
#endif	/* RANDOM_DEBUG */
    }

#endif	/* USE_EEPROM */

  // Read from a pin that is not connected (floating), which can return random results
#if DO_ANALOG_READ && (PIN_RANDOM >= 0)
  if (ANALOG_INPUT_OK (PIN_RANDOM))
    {
      uint32_t previous = 0;

#ifdef RANDOM_DEBUG
      Serial.print ("Read from pin PIN_RANDOM (");
      Serial.print (PIN_RANDOM);
      Serial.println (")");
#endif	/* RANDOM_DEBUG */

      for (int i = 0; i < RANDOM_NUMBER_READS; i++)
	{
	  tmp = (uint32_t) analogRead (PIN_RANDOM);
	  if (tmp != previous)
	    {
	      previous = tmp;
	      seed = combine_random (seed, tmp);
	    }

	  delay (RANDOM_NUMBER_DELAY);
	}
    }
#endif

  // Read from the internal temperture sensor
  tmp = read_internal_temp ();
  if (tmp)
    seed = combine_random (seed, tmp);

#ifdef RANDOM_DEBUG
  Serial.print ("Random final seed = 0x");
  Serial.println (seed, HEX);
#endif	/* RANDOM_DEBUG */

  randomSeed (seed);

  // If we are using EEPROM, update for the next time
  // call random once to make sure we don't repeat the seed
#if USE_EEPROM
  if (use_eeprom)
    {
      uint32_t r = random (0, 0x7fffffffUL);
      seed = combine_random (seed, r);

#if USE_EEPROM_LIB
      MRM_eeprom.set_uint32 (EEPROM_RANDOM, seed);

#else	/* !USE_EEPROM_LIB */
      EEPROM.write (EEPROM_RANDOM + 0, (uint8_t)(seed >>  0));
      EEPROM.write (EEPROM_RANDOM + 1, (uint8_t)(seed >>  8));
      EEPROM.write (EEPROM_RANDOM + 2, (uint8_t)(seed >> 16));
      EEPROM.write (EEPROM_RANDOM + 3, (uint8_t)(seed >> 24));

#endif	/* !USE_EEPROM_LIB */
      
#ifdef RANDOM_DEBUG
      Serial.print ("Save random seed, seed: 0x");
      Serial.println (seed, HEX);
#endif	/* RANDOM_DEBUG */
    }

#endif	/* USE_EEPROM */
}


// Choose a unique random number, trying to avoid a previous value

int32_t random_unique (int32_t low, int32_t high, int32_t current, ssize_t retries)
{
  int32_t new_random;

  if ((high - low) <= 1)
    retries = 0;

  do
    {
      new_random = random (low, high);
    }
  while (new_random == current && retries-- > 0);

  return new_random;
}


// HISTORY
// $Log: Meissner_Random.cpp,v $
// Revision 1.31  2022/02/13 01:08:25  michaelmeissner
// Update copyright year.
//
// Revision 1.30  2019/11/26 15:16:04  michaelmeissner
// On Teensy 4.0, use the internal temp. sensor.
//
// Revision 1.29  2019/09/03 05:51:47  michaelmeissner
// Add extra debug for printing the random pin.
//
// Revision 1.28  2019/09/03 05:24:32  michaelmeissner
// Add #if test for being able to do analog read and having a random pin.
//
// Revision 1.27  2019/01/14 02:19:37  michaelmeissner
// Update copyright.
//
// Revision 1.26  2016/10/08 04:42:02  michaelmeissner
// Mark an #endif
//
// Revision 1.25  2016/10/03 15:01:04  michaelmeissner
// Fix get random seed if not using mrm EEPROM library; do a random call to make a new seed before saving it to eeprom
//
// Revision 1.24  2016/10/03 14:52:46  michaelmeissner
// Use rotate in combining two seeds; Use 31 bits for seed generation; Don't allow seed of 0/-1 from EEPROM.
//
// Revision 1.23  2016/10/03 03:22:44  michaelmeissner
// Make seed 32-bits; Print EEPROM information.
//
// Revision 1.22  2016/10/03 02:47:16  michaelmeissner
// Print EEPROM saved location.
//
// Revision 1.21  2016/08/29 03:46:29  michaelmeissner
// Use PIN_TEMP_SENSOR for Teensy, like squarewear
//
// Revision 1.20  2016/08/29 03:38:36  michaelmeissner
// Use PIN_TEMP instead of hardwiring analog pin 38.
//
// Revision 1.19  2016/08/10 14:53:53  michaelmeissner
// If only one result possible, reset retries.
//
// Revision 1.18  2016/08/10 14:44:10  michaelmeissner
// Add random_unique function.
//
// Revision 1.17  2016/04/03 22:47:15  michaelmeissner
// Make (C) in copyright be upper case.
//
// Revision 1.16  2016/03/21 14:07:11  michaelmeissner
// Update copyright year.
//
// Revision 1.15  2015/11/25 12:53:45  michaelmeissner
// Move setting USE_EEPROM, USE_EEPROM_LIB to Meissner_Config.h.
//
// Revision 1.14  2015/03/06 23:19:06  michaelmeissner
// Suppress doing analog reads on squarewear
//
// Revision 1.13  2015/02/26 13:21:50  michaelmeissner
// On squarewear, use the temp. and light sensors to start the random number generator.
//
// Revision 1.12  2015/02/26 13:12:02  michaelmeissner
// Move 16Khz ATtiny 85 setup to set_random_seed from read_internal_temp; Comment #else/#endif
//
// Revision 1.11  2015/02/26 06:30:46  michaelmeissner
// Turn debugging back off
//
// Revision 1.10  2015/02/26 06:29:36  michaelmeissner
// Turn off EEPROM all together on the LC #2
//
// Revision 1.9  2015/02/26 06:27:03  michaelmeissner
// Turn off EEPROM all together on the LC
//
// Revision 1.8  2015/02/26 06:18:18  michaelmeissner
// Turn off using EEprom on the Teensy-LC
//
// Revision 1.7  2015/02/15 19:14:21  michaelmeissner
// Initial version
//
// Revision 1.6  2015/02/02 19:09:18  michaelmeissner
// Add support to put trinkets into 16Khz mode.
//
// Revision 1.5  2015/02/01 17:00:26  michaelmeissner
// On ATtiny85, don't use the EEprom library to save space.
//
// Revision 1.4  2015/02/01 16:57:47  michaelmeissner
// Use a consistant random type; Add ability not to use the EEprom library; Cleanups.
//
// Revision 1.3  2015/01/10 03:21:13  michaelmeissner
// Relicense under the MIT copyright so that we can link in neopixel patterns written by others.
//
// Revision 1.2  2015/01/06 12:46:41  michaelmeissner
// Add debug support; Don't use pin_random stored in EEPROM.
//
// Revision 1.1  2015/01/05 14:58:23  michaelmeissner
// Initial version.
//
